/*
 *                      Yeppp! library build framework
 *
 * This file is part of Yeppp! library infrastructure and licensed under 2-clause BSD license.
 * See library/LICENSE.txt for details.
 *
 */

package info.yeppp.Build;

import info.yeppp.Build.internals.FileSystem;
import org.apache.commons.io.FilenameUtils;
import org.apache.commons.lang3.SystemUtils;

import java.util.*;

public abstract class AbstractCppCompiler implements CppCompiler {
	public String getObjectPath(String sourcePath) {
		if (sourceDirectory == null) {
			throw new IllegalStateException("Source directory was not assigned any value");
		}
		if (objectDirectory == null) {
			throw new IllegalStateException("Object directory was not assigned any value");
		}
		final String relativeSourcePath = FileSystem.getRelativePath(sourcePath, sourceDirectory);
		if (relativeSourcePath == null) {
			throw new IllegalArgumentException("Source path is not within source directory");
		}
		return FilenameUtils.concat(objectDirectory,
			FilenameUtils.removeExtension(relativeSourcePath) + (SystemUtils.IS_OS_WINDOWS ? ".obj" : ".o"));
	}

	public void setSourceDirectory(String sourceDirectory) {
		this.sourceDirectory = sourceDirectory;
	}

	public String getSourceDirectory() {
		return this.sourceDirectory;
	}

	public void setObjectDirectory(String objectDirectory) {
		this.objectDirectory = objectDirectory;
	}

	public String getObjectDirectory() {
		return this.objectDirectory;
	}

	public void setAssemblerDirectory(String assemblerDirectory) {
		this.assemblerDirectory = assemblerDirectory;
	}

	public String getAssemblerDirectory() {
		return this.assemblerDirectory;
	}

	public void setOptimization(Optimization optimization) {
		this.optimization = optimization;
	}

	public Optimization getOptimization() {
		return this.optimization;
	}

	public void setDebugEnabled(boolean debugEnabled) {
		this.debugEnabled = debugEnabled;
	}

	public boolean getDebugEnabled() {
		return this.debugEnabled;
	}

	public void setRuntimeLibrary(RuntimeLibrary runtimeLibrary) {
		this.runtimeLibrary = runtimeLibrary;
	}

	public RuntimeLibrary getRuntimeLibrary() {
		return this.runtimeLibrary;
	}

	@Override
	public boolean getRttiEnabled() {
		return this.rttiEnabled;
	}

	@Override
	public void setExceptionsSupport(Exceptions exceptions) {
		this.exceptions = exceptions;
	}

	@Override
	public Exceptions getExceptionsEnabled() {
		return this.exceptions;
	}

	@Override
	public void setRttiEnabled(boolean rttiEnabled) {
		this.rttiEnabled = rttiEnabled;
	}

	public boolean addMacro(String name) {
		return this.addMacro(name, "1");
	}

	public boolean addMacro(String name, String value) {
		if (macros.containsKey(name)) {
			return false;
		} else {
			macros.put(name, value);
			return true;
		}
	}

	public boolean removeMacro(String name) {
		return (macros.remove(name) != null);
	}

	public void clearMacros() {
		macros.clear();
	}

	public boolean addGlobalIncludeDirectory(String includePath) {
		return globalIncludes.add(FilenameUtils.normalizeNoEndSeparator(includePath));
	}

	public boolean removeGlobalIncludeDirectory(String includePath) {
		return globalIncludes.remove(includePath);
	}

	public void clearGlobalIncludeDirectories() {
		globalIncludes.clear();
	}

	public boolean addIncludeDirectory(String includePath) {
		return includes.add(FilenameUtils.normalizeNoEndSeparator(includePath));
	}

	public boolean removeIncludeDirectory(String includePath) {
		return includes.remove(FilenameUtils.normalizeNoEndSeparator(includePath));
	}

	public void clearIncludeDirectories() {
		includes.clear();
	}

	public void clearAll() {
		this.clearGlobalIncludeDirectories();
		this.clearIncludeDirectories();
		this.clearMacros();
	}

	protected String sourceDirectory = null;
	protected String objectDirectory = null;
	protected String assemblerDirectory = null;
	protected Map<String, String> macros = new HashMap<String, String>();
	protected Set<String> includes = new HashSet<String>();
	protected Set<String> globalIncludes = new HashSet<String>();
	protected Optimization optimization = Optimization.NoOptimization;
	protected boolean debugEnabled = false;
	protected Exceptions exceptions = Exceptions.SyncronousExceptions;
	protected boolean rttiEnabled = true;
	protected RuntimeLibrary runtimeLibrary = RuntimeLibrary.DynamicRuntimeLibrary;
}
