/*
 *                      Yeppp! library build framework
 *
 * This file is part of Yeppp! library infrastructure and licensed under 2-clause BSD license.
 * See library/LICENSE.txt for details.
 *
 */

package info.yeppp.Build;

import info.yeppp.Build.sdk.JavaSdk;
import org.apache.commons.io.FileUtils;
import org.apache.commons.io.FilenameUtils;
import org.apache.commons.lang3.SystemUtils;

import java.io.File;
import java.util.*;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;

public class LibraryBuilder {
	public static class Toolchain {
		public Toolchain(Architecture architecture, String target) {
			this.architecture = architecture;
			this.target = target;
		}

		final public Architecture architecture;

		@Override
		public String toString() {
			return target;
		}

		final public String target;
		public CppCompiler compiler = null;
		public JavaCompiler javaCompiler = null;
		public JavaSdk javaSdk = null;
		public Linker linker = null;
		public Librarian librarian = null;
		public NetwideAssembler nasm = null;
		public GnuAssembler gas = null;
		public ResourceCompiler resourceCompiler = null;
	}

	public LibraryBuilder(String projectDirectory, Toolchain toolchain) {
		this.projectDirectory = projectDirectory;
		this.target = toolchain.target;
		this.architecture = toolchain.architecture;
		this.cppCompiler = toolchain.compiler;
		this.javaCompiler = toolchain.javaCompiler;
		this.javaSdk = toolchain.javaSdk;
		this.linker = toolchain.linker;
		this.librarian = toolchain.librarian;
		this.nasm = toolchain.nasm;
		this.gas = toolchain.gas;
		this.resourceCompiler = toolchain.resourceCompiler;
	}

	public BuildMessages compileLibrary() throws Exception {
		final String sourceDirectory = FilenameUtils.concat(projectDirectory, "library\\sources");
		final String headerDirectory = FilenameUtils.concat(projectDirectory, "library\\headers");
		final String objectDirectory = FilenameUtils.concat(projectDirectory, "library\\binaries\\" + target);
		final String resourceDirectory = FilenameUtils.concat(projectDirectory, "library\\resources\\");
		final String libraryPath = FilenameUtils.concat(objectDirectory, "yeppp");
		final String specificationPath = FilenameUtils.concat(sourceDirectory, "yeppp");

		cppCompiler.clearAll();
		cppCompiler.addMacro("YEP_BUILD_LIBRARY");
		cppCompiler.setExceptionsSupport(CppCompiler.Exceptions.NoExceptions);
		cppCompiler.setRttiEnabled(false);
		cppCompiler.setRuntimeLibrary(CppCompiler.RuntimeLibrary.NoRuntimeLibrary);
		if (cppCompiler instanceof GnuCppCompiler) {
			GnuCppCompiler gnuCppCompiler = (GnuCppCompiler) cppCompiler;
			gnuCppCompiler.setPositionIndependentCodeGeneration(GnuCppCompiler.PositionIndependentCodeGeneration.UnlimitedLibraryPIC);
		}
		cppCompiler.setSourceDirectory(sourceDirectory);
		cppCompiler.setObjectDirectory(objectDirectory);
		cppCompiler.addIncludeDirectory(headerDirectory);
		cppCompiler.addIncludeDirectory(sourceDirectory);
		cppCompiler.addDefaultGlobalIncludeDirectories();
		cppCompiler.setOptimization(CppCompiler.Optimization.MaxSpeedOptimization);

//		cppCompiler.setRuntimeLibrary(CppCompiler.RuntimeLibrary.DynamicRuntimeLibrary);
//		cppCompiler.setDebugEnabled(true);

		if (nasm != null) {
			nasm.setSourceDirectory(sourceDirectory);
			nasm.setObjectDirectory(objectDirectory);
		}
		if (gas != null) {
			gas.setSourceDirectory(sourceDirectory);
			gas.setObjectDirectory(objectDirectory);
		}

		final ExecutorService executorService = Executors.newFixedThreadPool(Runtime.getRuntime().availableProcessors() * 2);
//		final ExecutorService executorService = Executors.newFixedThreadPool(1);
		final Collection<File> sources = FileUtils.listFiles(new File(sourceDirectory), null, true);
		final Collection<File> resources = FileUtils.listFiles(new File(resourceDirectory), new String[] { "rc" }, true);
		final List<String> objectPaths = new ArrayList<String>(sources.size());
		final BuildMessages buildMessages = new BuildMessages();
		String targetAsmExtension = null;
		if (architecture.equals(Architecture.X86)) {
			targetAsmExtension = ".x86.asm";
		} else if (architecture.equals(Architecture.X64)) {
			if (SystemUtils.IS_OS_WINDOWS) {
				targetAsmExtension = ".x64-ms.asm";
			} else {
				targetAsmExtension = ".x64-sysv.asm";
			}
		} else if (architecture.equals(Architecture.ARM)) {
			targetAsmExtension = ".arm.asm";
		} else if (architecture.equals(Architecture.IA64)) {
			targetAsmExtension = ".ia64.asm";
		}
		for (File source : sources) {
			if (source.getName().endsWith(".cpp")) {
				final String sourcePath = source.getAbsolutePath();
				executorService.submit(new Runnable() {
					public void run() {
						buildMessages.add(cppCompiler.compile(sourcePath));
					}
				});
				objectPaths.add(cppCompiler.getObjectPath(source.getAbsolutePath()));
			} else if (source.getName().endsWith(targetAsmExtension)) {
				if (nasm != null) {
					buildMessages.add(nasm.assemble(source.getAbsolutePath(), architecture.equals(Architecture.X64)));
					objectPaths.add(nasm.getObjectPath(source.getAbsolutePath()));
				} else {
					buildMessages.add(gas.assemble(source.getAbsolutePath()));
					objectPaths.add(gas.getObjectPath(source.getAbsolutePath()));
				}
			}
		}
		if (resourceCompiler != null) {
			resourceCompiler.clearAll();
			resourceCompiler.setScriptDirectory(resourceDirectory);
			resourceCompiler.setResourceDirectory(objectDirectory);
			resourceCompiler.addIncludeDirectory(headerDirectory);
			resourceCompiler.addDefaultGlobalIncludeDirectories();
			for (File resource : resources) {
				resourceCompiler.compileResourceScript(resource.getAbsolutePath());
				objectPaths.add(resourceCompiler.getResourcePath(resource.getAbsolutePath()));
			}
		}
		executorService.shutdown();
		executorService.awaitTermination(24, TimeUnit.HOURS);

		linker.clearAll();
		linker.setRuntimeLibraryUse(false);
		linker.setObjectDirectory(objectDirectory);
		linker.setBinariesDirectory(objectDirectory);
		linker.addDefaultGlobalLibraryDirectories();
		if (linker instanceof MicrosoftLinker) {
			linker.addDynamicLibraryDependence("kernel32");
		}
		buildMessages.add(linker.linkDynamicLibrary(libraryPath, objectPaths, specificationPath));
		return buildMessages;
	}

	public BuildMessages buildPerformanceTests() throws Exception {
		final String testSourceDirectory = FilenameUtils.concat(projectDirectory, "perf-tests\\sources");
		final String testObjectDirectory = FilenameUtils.concat(projectDirectory, "perf-tests\\binaries\\" + target);
		final String libraryHeaderDirectory = FilenameUtils.concat(projectDirectory, "library\\headers");
		final String librarySourceDirectory = FilenameUtils.concat(projectDirectory, "library\\sources");
		final String libraryBinaryDirectory = FilenameUtils.concat(projectDirectory, "library\\binaries\\" + target);
		final String libraryName = "yeppp";

		cppCompiler.clearAll();
//		cppCompiler.addMacro("YEP_STATIC_LIBRARY");
		cppCompiler.setExceptionsSupport(CppCompiler.Exceptions.NoExceptions);
		cppCompiler.setRttiEnabled(false);
		cppCompiler.setRuntimeLibrary(CppCompiler.RuntimeLibrary.DynamicRuntimeLibrary);
		cppCompiler.setSourceDirectory(testSourceDirectory);
		cppCompiler.setObjectDirectory(testObjectDirectory);
		cppCompiler.addIncludeDirectory(libraryHeaderDirectory);
		cppCompiler.addIncludeDirectory(librarySourceDirectory);
		cppCompiler.addDefaultGlobalIncludeDirectories();

		linker.clearAll();
		linker.addDynamicLibraryDependence(libraryName);
		linker.addLibraryDirectory(libraryBinaryDirectory);
		linker.setBinariesDirectory(testObjectDirectory);
		linker.addDefaultGlobalLibraryDirectories();

		final Collection<File> sources = FileUtils.listFiles(new File(testSourceDirectory), new String[] { "cpp" }, true);
		final BuildMessages buildMessages = new BuildMessages();
		for (File source : sources) {
			if (source.getName().endsWith(".cpp")) {
				final String sourcePath = source.getAbsolutePath();
				buildMessages.add(cppCompiler.compile(source.getAbsolutePath()));
				final String objectPath = cppCompiler.getObjectPath(sourcePath);
				final List<String> objectPathList = new LinkedList<String>();
				objectPathList.add(objectPath);
				buildMessages.add(linker.linkExecutable(FilenameUtils.removeExtension(objectPath), objectPathList));
			}
		}
		return buildMessages;
	}

	public BuildMessages buildExamples() throws Exception {
		final String examplesSourceDirectory = FilenameUtils.concat(projectDirectory, "examples\\sources");
		final String examplesObjectDirectory = FilenameUtils.concat(projectDirectory, "examples\\binaries\\" + target);
		final String libraryHeaderDirectory = FilenameUtils.concat(projectDirectory, "library\\headers");
		final String libraryBinaryDirectory = FilenameUtils.concat(projectDirectory, "library\\binaries\\" + target);
		final String libraryName = "yeppp";

		cppCompiler.clearAll();
		cppCompiler.setExceptionsSupport(CppCompiler.Exceptions.NoExceptions);
		cppCompiler.setRttiEnabled(false);
		cppCompiler.setRuntimeLibrary(CppCompiler.RuntimeLibrary.DynamicRuntimeLibrary);
		cppCompiler.setSourceDirectory(examplesSourceDirectory);
		cppCompiler.setObjectDirectory(examplesObjectDirectory);
		cppCompiler.addIncludeDirectory(libraryHeaderDirectory);
		cppCompiler.addDefaultGlobalIncludeDirectories();

		linker.clearAll();
		if (linker instanceof MicrosoftLinker) {
			linker.addDynamicLibraryDependence("kernel32");
		}
		linker.addDynamicLibraryDependence(libraryName);
		linker.addLibraryDirectory(libraryBinaryDirectory);
		linker.setBinariesDirectory(examplesObjectDirectory);
		linker.addDefaultGlobalLibraryDirectories();

		final Collection<File> sources = FileUtils.listFiles(new File(examplesSourceDirectory), new String[] { "cpp" }, true);
		final BuildMessages buildMessages = new BuildMessages();
		for (File source : sources) {
            if (source.getName().equals("IKN.cpp") && !SystemUtils.IS_OS_WINDOWS)
                continue;

			if (source.getName().endsWith(".cpp")) {
				final String sourcePath = source.getAbsolutePath();
				buildMessages.add(cppCompiler.compile(source.getAbsolutePath()));
				final String objectPath = cppCompiler.getObjectPath(sourcePath);
				final List<String> objectPathList = new LinkedList<String>();
				objectPathList.add(objectPath);
				buildMessages.add(linker.linkExecutable(FilenameUtils.removeExtension(objectPath), objectPathList));
			}
		}
		return buildMessages;
	}

	public BuildMessages buildJniClasses() throws Exception {
		final String jniSourceDirectory = FilenameUtils.concat(projectDirectory, "bindings\\java\\sources-java");
		final String jniClassDirectory = FilenameUtils.concat(projectDirectory, "bindings\\java\\binaries\\java");

		javaCompiler.setSourceDirectory(jniSourceDirectory);
		javaCompiler.setClassDirectory(jniClassDirectory);

		final Collection<File> sources = FileUtils.listFiles(new File(jniSourceDirectory), new String[] { "java" }, true);
		final BuildMessages buildMessages = new BuildMessages();
		for (File source : sources) {
			if (source.getName().endsWith(".java")) {
				final String sourcePath = source.getAbsolutePath();
				buildMessages.add(javaCompiler.compile(source.getAbsolutePath()));
//				final String objectPath = javaCompiler.getObjectPath(sourcePath);
//				final List<String> objectPathList = new LinkedList<String>();
//				objectPathList.add(objectPath);
//				buildMessages.addAll(linker.linkExecutable(FilenameUtils.removeExtension(objectPath), objectPathList));
			}
		}
		return buildMessages;
	}

	public BuildMessages buildJniLibrary() throws Exception {
		final String jniSourceDirectory = FilenameUtils.concat(projectDirectory, "bindings\\java\\sources-jni");
		final String libraryHeaderDirectory = FilenameUtils.concat(projectDirectory, "library\\headers");
		final String libraryBinaryDirectory = FilenameUtils.concat(projectDirectory, "library\\binaries\\" + target);
		final String objectDirectory = FilenameUtils.concat(projectDirectory, "bindings\\java\\binaries\\" + target);
		final String libraryPath = FilenameUtils.concat(objectDirectory, "yeppp-jni");

		cppCompiler.clearAll();
		cppCompiler.setSourceDirectory(jniSourceDirectory);
		cppCompiler.setObjectDirectory(objectDirectory);
		cppCompiler.addIncludeDirectory(libraryHeaderDirectory);
		cppCompiler.addIncludeDirectory(javaSdk.getJniIncludePath());
		cppCompiler.addIncludeDirectory(javaSdk.getJniPlatformIncludePath());
		cppCompiler.addDefaultGlobalIncludeDirectories();

		cppCompiler.setExceptionsSupport(CppCompiler.Exceptions.NoExceptions);
		cppCompiler.setRttiEnabled(false);
		cppCompiler.setRuntimeLibrary(CppCompiler.RuntimeLibrary.NoRuntimeLibrary);
		cppCompiler.setOptimization(CppCompiler.Optimization.MinSizeOptimization);

		final Collection<File> sources = FileUtils.listFiles(new File(jniSourceDirectory), new String[] { "c" }, true);
		final List<String> objectPaths = new ArrayList<String>(sources.size());
		final BuildMessages buildMessages = new BuildMessages();
		for (File source : sources) {
			final String sourcePath = source.getAbsolutePath();
			buildMessages.add(cppCompiler.compile(source.getAbsolutePath()));
			final String objectPath = cppCompiler.getObjectPath(sourcePath);
			objectPaths.add(objectPath);
		}

		linker.clearAll();
		linker.setObjectDirectory(objectDirectory);
		linker.setBinariesDirectory(objectDirectory);
		linker.addLibraryDirectory(libraryBinaryDirectory);
		linker.addDynamicLibraryDependence("yeppp");
		linker.addDefaultGlobalLibraryDirectories();
		buildMessages.add(linker.linkDynamicLibrary(libraryPath, objectPaths, null));
		return buildMessages;
	}


	protected final Architecture architecture;
	protected final String projectDirectory;
	protected String target;
	protected CppCompiler cppCompiler = null;
	protected JavaCompiler javaCompiler = null;
	protected JavaSdk javaSdk = null;
	protected Linker linker = null;
	protected Librarian librarian = null;
	protected NetwideAssembler nasm = null;
	protected GnuAssembler gas = null;
	protected ResourceCompiler resourceCompiler = null;
}
