/*
 *                      Yeppp! library build framework
 *
 * This file is part of Yeppp! library infrastructure and licensed under 2-clause BSD license.
 * See library/LICENSE.txt for details.
 *
 */

package info.yeppp.Build;

import info.yeppp.Build.internals.*;
import info.yeppp.Build.internals.Process;
import info.yeppp.Build.sdk.Windows8Sdk;
import info.yeppp.Build.sdk.WindowsSdk;
import org.apache.commons.io.FileUtils;
import org.apache.commons.io.FilenameUtils;

import java.io.File;
import java.util.HashSet;
import java.util.Set;
import java.util.regex.Pattern;

public class ResourceCompiler{
	public ResourceCompiler(WindowsSdk windowsSdk, VisualStudio visualStudio) {
		this.windowsSdk = windowsSdk;
		this.visualStudio = visualStudio;
		this.resourceCompilerPath = FilenameUtils.concat(windowsSdk.getToolsPath(), "rc.exe");
	}

	public String getResourcePath(String scriptPath) {
		if (scriptDirectory == null) {
			throw new IllegalStateException("Script directory was not assigned any value");
		}
		if (resourceDirectory == null) {
			throw new IllegalStateException("Resource directory was not assigned any value");
		}
		final String relativeScriptPath = FileSystem.getRelativePath(scriptPath, scriptDirectory);
		if (relativeScriptPath == null) {
			throw new IllegalArgumentException("Script path is not within source directory");
		}
		final String resourcePath = FilenameUtils.concat(resourceDirectory,
				FilenameUtils.removeExtension(relativeScriptPath) + ".res");
		return resourcePath;
	}

	public BuildMessages compileResourceScript(final String scriptPath) {
		final String resourcePath = this.getResourcePath(scriptPath);

		final BuildMessages buildMessages = new BuildMessages();

		FileUtils.deleteQuietly(new File(resourcePath));
		createDirectories(resourcePath);

		final info.yeppp.Build.internals.Process.Parameters parameters = prepareParameters(scriptPath, resourcePath);
		final info.yeppp.Build.internals.Process.Environment environment = prepareEnvironment();
		final Process.EventHandler eventHandler = new Process.EventHandler() {
			@Override
			public void onStdOutLineWrite(String line) {
				if (!line.isEmpty()) {
					if (!Pattern.matches("Microsoft \\(R\\) Windows \\(R\\) Resource Compiler Version (\\d+).(\\d+).(\\d+).(\\d+)", line)) {
						if (!Pattern.matches("Copyright \\(C\\) Microsoft Corporation\\.  All rights reserved\\.", line)) {
							buildMessages.add(new BuildMessage(BuildMessage.Type.Error, scriptPath, null, null, line));
						}
					}
				}
			}

			@Override
			public void onStdErrLineWrite(String line) {
				buildMessages.add(new BuildMessage(BuildMessage.Type.Error, scriptPath, null, null, line));
			}

			@Override
			public void onError(String description) {
				buildMessages.add(new BuildMessage(BuildMessage.Type.Exception, scriptPath, null, null, description));
			}

			@Override
			public void onCompletion(Integer status) {
			}
		};
//		System.out.println(parameters.toString());
		final Process resourceCompiler = new Process(resourceCompilerPath, parameters, environment, eventHandler);
		resourceCompiler.join();
		return buildMessages;
	}

	private void createDirectories(String path) {
		File file = new File(path);
		if (!file.isDirectory()) {
			file = file.getParentFile();
		}
		file.mkdirs();
	}

	private Process.Parameters prepareParameters(String scriptPath, String resourcePath) {
		final info.yeppp.Build.internals.Process.Parameters parameters = new Process.Parameters();
		for (String include : includes) {
			parameters.add("/I", include);
		}
		parameters.add("/Fo", resourcePath);
		parameters.add(scriptPath);
		return parameters;
	}

	private Process.Environment prepareEnvironment() {
		final Process.Environment environment =  new Process.Environment();
		environment.prependPathVariable("PATH", FilenameUtils.concat(windowsSdk.getToolsPath(), "bin"));
		for (String globalInclude : globalIncludes) {
			environment.prependPathVariable("INCLUDE", globalInclude);
		}
		return environment;
	}

	public void setResourceDirectory(String resourceDirectory) {
		this.resourceDirectory = resourceDirectory;
	}

	public String getResourceDirectory() {
		return this.resourceDirectory;
	}

	public void setScriptDirectory(String scriptDirectory) {
		this.scriptDirectory = scriptDirectory;
	}

	public String getScriptDirectory() {
		return this.scriptDirectory;
	}

	public boolean addIncludeDirectory(String includePath) {
		return includes.add(FilenameUtils.normalizeNoEndSeparator(includePath));
	}

	public boolean removeIncludeDirectory(String includePath) {
		return includes.remove(FilenameUtils.normalizeNoEndSeparator(includePath));
	}

	public void clearIncludeDirectories() {
		includes.clear();
	}

	public boolean addGlobalIncludeDirectory(String includePath) {
		return globalIncludes.add(FilenameUtils.normalizeNoEndSeparator(includePath));
	}

	public boolean removeGlobalIncludeDirectory(String includePath) {
		return globalIncludes.remove(includePath);
	}

	public void clearGlobalIncludeDirectories() {
		globalIncludes.clear();
	}

	public void addDefaultGlobalIncludeDirectories() {
		this.addGlobalIncludeDirectory(windowsSdk.getIncludePath());
		if (windowsSdk instanceof Windows8Sdk) {
			this.addGlobalIncludeDirectory(((Windows8Sdk) windowsSdk).getSharedIncludePath());
		}
		this.addGlobalIncludeDirectory(visualStudio.getCppIncludePath());
	}

	public void clearAll() {
		this.clearIncludeDirectories();
		this.clearGlobalIncludeDirectories();
	}

	private String scriptDirectory;
	private String resourceDirectory;
	private final WindowsSdk windowsSdk;
	private final VisualStudio visualStudio;
	private final Set<String> globalIncludes = new HashSet<String>();
	private final Set<String> includes = new HashSet<String>();
	private final String resourceCompilerPath;
}
