#
#       Portable Efficient Assembly Code-generator in Higher-level Python (PEACH-Py)
#
# This file is part of Yeppp! library infrastructure and licensed under 2-clause BSD license.
# See codegen/LICENSE.txt for details.
#

__author__ = 'Marat'

import re

class ABI:
	def __init__(self, abi):
		if abi == 'x86':
			import x86
			self.name = 'x86'
			self.pointer_size = 4
			self.general_purpose_register_size = 4
			self.int_size = 4
			self.long_size = 4
			self.long_long_size = 8
			self.stack_alignment = 4
			self.soft_float = False
			self.function_prefix = "_"
			self.callee_save_registers = [x86.ebx, x86.esi, x86.edi, x86.ebp]
			self.argument_registers = []
			self.volatile_registers = [x86.eax, x86.ecx, x86.edx,
			                           x86.mm0, x86.mm1, x86.mm2, x86.mm3, x86.mm4, x86.mm5, x86.mm6, x86.mm7,
			                           x86.xmm0, x86.xmm1, x86.xmm2, x86.xmm3, x86.xmm4, x86.xmm5, x86.xmm6, x86.xmm7]
		elif abi == 'x32':
			import x64
			self.name = 'x32'
			self.pointer_size = 4
			self.general_purpose_register_size = 8
			self.int_size = 4
			self.long_size = 8
			self.long_long_size = 8
			self.stack_alignment = 16
			self.soft_float = False
			self.function_prefix = ""
			self.callee_save_registers = [x64.rbx, x64.rbp, x64.r12, x64.r13, x64.r14, x64.r15]
		elif abi == 'x64-sysv':
			import x64
			self.name = 'x64-sysv'
			self.pointer_size = 8
			self.general_purpose_register_size = 8
			self.int_size = 4
			self.long_size = 8
			self.long_long_size = 8
			self.stack_alignment = 16
			self.soft_float = False
			self.function_prefix = ""
			self.callee_save_registers = [x64.rbx, x64.rbp, x64.r12, x64.r13, x64.r14, x64.r15]
			self.argument_registers = [x64.rdi, x64.rsi, x64.rdx, x64.rcx, x64.r8, x64.r9,
			                           x64.xmm0, x64.xmm1, x64.xmm2, x64.xmm3, x64.xmm4, x64.xmm5, x64.xmm6, x64.xmm7]
			self.volatile_registers = [x64.rax, x64.r10, x64.r11,
			                           x64.mm0, x64.mm1, x64.mm2, x64.mm3, x64.mm4, x64.mm5, x64.mm6, x64.mm7,
			                           x64.xmm0, x64.xmm1, x64.xmm2, x64.xmm3, x64.xmm4, x64.xmm5, x64.xmm6, x64.xmm7,
			                           x64.xmm8, x64.xmm9, x64.xmm10, x64.xmm11, x64.xmm12, x64.xmm13, x64.xmm14, x64.xmm15]
		elif abi == 'x64-ms':
			import x64
			self.name = 'x64-ms'
			self.pointer_size = 8
			self.general_purpose_register_size = 8
			self.int_size = 4
			self.long_size = 4
			self.long_long_size = 8
			self.stack_alignment = 16
			self.soft_float = False
			self.function_prefix = ""
			self.callee_save_registers = [x64.rbx, x64.rsi, x64.rdi, x64.rbp, x64.r12, x64.r13, x64.r14, x64.r15,
			                              x64.xmm6, x64.xmm7, x64.xmm8, x64.xmm9, x64.xmm10, x64.xmm11,
			                              x64.xmm12, x64.xmm13, x64.xmm14, x64.xmm15]
			self.argument_registers = [x64.rcx, x64.rdx, x64.r8, x64.r9,
			                           x64.xmm0, x64.xmm1, x64.xmm2, x64.xmm3]
			self.volatile_registers = [x64.rax, x64.r10, x64.r11,
			                           x64.mm0, x64.mm1, x64.mm2, x64.mm3, x64.mm4, x64.mm5, x64.mm6, x64.mm7,
			                           x64.xmm4, x64.xmm5]
		elif abi == 'arm-eabi-sf':
			self.name = 'arm-eabi-sf'
			self.pointer_size = 4
			self.general_purpose_register_size = 4
			self.int_size = 4
			self.long_size = 4
			self.long_long_size = 8
			self.stack_alignment = 8
			self.soft_float = True
			self.function_prefix = ""
		elif abi == 'arm-eabi-hf':
			self.name = 'arm-eabi-hf'
			self.pointer_size = 4
			self.general_purpose_register_size = 4
			self.int_size = 4
			self.long_size = 4
			self.long_long_size = 8
			self.stack_alignment = 8
			self.soft_float = False
			self.function_prefix = ""
		elif abi == 'mips-o32-hf':
			self.name = 'mips-o32-hf'
			self.pointer_size = 4
			self.general_purpose_register_size = 4
			self.int_size = 4
			self.long_size = 4
			self.long_long_size = 8
			self.stack_alignment = 8
			self.soft_float = False
			self.function_prefix = ""

	def __str__(self):
		return self.name

	def __eq__(self, other):
		if isinstance(other, ABI):
			return self.name == other.name
		else:
			return False

	def __hash__(self):
		return hash(self.name)

	def get_name(self):
		return self.name

class Type(object):
	unsigned_integer_types = ['uint8_t', 'uint16_t', 'uint32_t', 'uint64_t', 'size_t']
	signed_integer_types = ['int8_t', 'int16_t', 'int32_t', 'int64_t', 'ptrdiff_t']
	integer_type = unsigned_integer_types + signed_integer_types
	floating_point_types = ['half', 'float', 'double']

	def __init__(self, type):
		super(Type, self).__init__()
		type_components = re.match("(?:(const)?\\s+)?((?:uint|int)(?:8|16|32|64)_t|size_t|ptrdiff_t|half|float|double)(?:\\s*(\\*?))?", type)
		if type_components is None:
			raise ValueError('The type {0} is not recognized as a C type'.format(type))
		self.is_const_type = bool(type_components.group(1))
		self.primitive_type = type_components.group(2)
		self.is_pointer_type = bool(type_components.group(3))

	def __str__(self):
		text = self.primitive_type
		if self.is_const_type:
			text = "const " + text
		if self.is_pointer_type:
			text = text + "*"
		return text

	def is_size(self):
		if self.is_pointer:
			return False
		else:
			return self.primitive_type == 'size_t'

	def is_unsigned_integer(self):
		if self.is_pointer():
			return False
		else:
			return self.primitive_type in Type.unsigned_integer_types

	def is_signed_integer(self):
		if self.is_pointer():
			return False
		else:
			return self.primitive_type in Type.signed_integer_types

	def is_integer(self):
		if self.is_pointer:
			return False
		else:
			return self.primitive_type in Type.integer_types

	def is_pointer(self):
		return self.is_pointer_type

	def is_floating_point(self):
		if self.is_pointer():
			return False
		else:
			return self.primitive_type in Type.floating_point_types

	def get_primitive_type(self):
		return Type(self.primitive_type)

	def get_size(self, abi):
		if isinstance(abi, ABI):
			if self.is_pointer() or self.primitive_type == 'size_t' or self.primitive_type == 'ptrdiff_t':
				return abi.pointer_size
			elif self.primitive_type in ['uint8_t', 'int8_t']:
				return 1
			elif self.primitive_type in ['uint16_t', 'int16_t', 'half']:
				return 2
			elif self.primitive_type in ['uint32_t', 'int32_t', 'float']:
				return 4
			elif self.primitive_type in ['uint64_t', 'int64_t', 'double']:
				return 8
			else:
				raise ValueError('Unknown type {0}'.format(self.primitive_type))
		else:
			raise TypeError('Wrong type of ABI object')

class Parameter(object):
	def __init__(self, name, ctype):
		super(Parameter, self).__init__()
		if isinstance(name, str):
			self.name = name
		else:
			raise TypeError('Parameter name must be a string')
		if isinstance(ctype, Type):
			self.type = ctype
		else:
			raise TypeError('Parameter tyep must be an instance of Type object')

	def __eq__(self, other):
		return self.name == other.name

	def __hash__(self):
		return hash(self.name)

	def __str__(self):
		return self.name

	def get_name(self):
		return self.name

	def get_type(self):
		return self.type

	def get_size(self, abi):
		return self.type.get_size(abi)


