/*
 *                      Yeppp! library implementation
 *
 * This file is part of Yeppp! library and licensed under 2-clause BSD license.
 * See library/LICENSE.txt for details.
 *
 */

#include <yepPredefines.h>
#include <yepTypes.h>
#include <yepPrivate.hpp>
#include <yepLibrary.h>
#include <library/functions.h>
#include <yepIntrinsics.h>

#if defined(YEP_LINUX_OS)
	#include <yepSyscalls.hpp>

	#include <unistd.h>
	#include <fcntl.h>
	#include <sys/syscall.h>
	#include <linux/unistd.h>
	#if defined(__ANDROID__)
		#include <signal.h>
		#include <sys/ioctl.h>
		#if defined(YEP_ARM_CPU)
			// The last parts of the context structures are not needed and leaved out
			struct sigcontext {
				unsigned long trap_no;
				unsigned long error_code;
				unsigned long oldmask;
				unsigned long arm_r0;
				unsigned long arm_r1;
				unsigned long arm_r2;
				unsigned long arm_r3;
				unsigned long arm_r4;
				unsigned long arm_r5;
				unsigned long arm_r6;
				unsigned long arm_r7;
				unsigned long arm_r8;
				unsigned long arm_r9;
				unsigned long arm_r10;
				unsigned long arm_fp;
				unsigned long arm_ip;
				unsigned long arm_sp;
				unsigned long arm_lr;
				unsigned long arm_pc;
				unsigned long arm_cpsr;
				unsigned long fault_address;
			};
		#endif
		struct ucontext {
			unsigned long	  uc_flags;
			struct ucontext  *uc_link;
			stack_t		  uc_stack;
			struct sigcontext uc_mcontext;
		};

		typedef struct ucontext ucontext_t;
		#if defined(YEP_ARM_CPU) || defined(YEP_MIPS_CPU)
			#define PERF_EVENT_IOC_ENABLE  _IO ('$', 0)
			#define PERF_EVENT_IOC_DISABLE _IO ('$', 1)
			#define PERF_EVENT_IOC_REFRESH _IO ('$', 2)
			#define PERF_EVENT_IOC_RESET   _IO ('$', 3)

			enum perf_type_id {
				PERF_TYPE_HARDWARE   = 0,
				PERF_TYPE_SOFTWARE   = 1,
				PERF_TYPE_TRACEPOINT = 2,
				PERF_TYPE_HW_CACHE   = 3,
				PERF_TYPE_RAW        = 4,
				PERF_TYPE_BREAKPOINT = 5
			};

			enum perf_hw_id {
				PERF_COUNT_HW_CPU_CYCLES              = 0,
				PERF_COUNT_HW_INSTRUCTIONS            = 1,
				PERF_COUNT_HW_CACHE_REFERENCES        = 2,
				PERF_COUNT_HW_CACHE_MISSES            = 3,
				PERF_COUNT_HW_BRANCH_INSTRUCTIONS     = 4,
				PERF_COUNT_HW_BRANCH_MISSES           = 5,
				PERF_COUNT_HW_BUS_CYCLES              = 6,
				PERF_COUNT_HW_STALLED_CYCLES_FRONTEND = 7,
				PERF_COUNT_HW_STALLED_CYCLES_BACKEND  = 8
			};

			enum perf_hw_cache_id {
				PERF_COUNT_HW_CACHE_L1D  = 0,
				PERF_COUNT_HW_CACHE_L1I  = 1,
				PERF_COUNT_HW_CACHE_LL   = 2,
				PERF_COUNT_HW_CACHE_DTLB = 3,
				PERF_COUNT_HW_CACHE_ITLB = 4,
				PERF_COUNT_HW_CACHE_BPU  = 5,
				PERF_COUNT_HW_CACHE_NODE = 6
			};

			enum perf_hw_cache_op_id {
				PERF_COUNT_HW_CACHE_OP_READ     = 0,
				PERF_COUNT_HW_CACHE_OP_WRITE    = 1,
				PERF_COUNT_HW_CACHE_OP_PREFETCH = 2
			};

			enum perf_hw_cache_op_result_id {
				PERF_COUNT_HW_CACHE_RESULT_ACCESS = 0,
				PERF_COUNT_HW_CACHE_RESULT_MISS   = 1
			};

			enum perf_sw_ids {
				PERF_COUNT_SW_CPU_CLOCK        = 0,
				PERF_COUNT_SW_TASK_CLOCK       = 1,
				PERF_COUNT_SW_PAGE_FAULTS      = 2,
				PERF_COUNT_SW_CONTEXT_SWITCHES = 3,
				PERF_COUNT_SW_CPU_MIGRATIONS   = 4,
				PERF_COUNT_SW_PAGE_FAULTS_MIN  = 5,
				PERF_COUNT_SW_PAGE_FAULTS_MAJ  = 6,
				PERF_COUNT_SW_ALIGNMENT_FAULTS = 7,
				PERF_COUNT_SW_EMULATION_FAULTS = 8
			};

			struct perf_event_attr {
				Yep32u type;
				Yep32u size;
				Yep64u config;
				union {
					Yep64u sample_period;
					Yep64u sample_freq;
				};
				Yep64u sample_type;
				Yep64u read_format;
				Yep64u disabled       : 1,
				       inherit        : 1,
				       pinned         : 1,
				       exclusive      : 1,
				       exclude_user   : 1,
				       exclude_kernel : 1,
				       exclude_hv     : 1,
				       exclude_idle   : 1,
				       mmap           : 1,
				       comm           : 1,
				       freq           : 1,
				       inherit_stat   : 1,
				       enable_on_exec : 1,
				       task           : 1,
				       watermark      : 1,
				       precise_ip     : 2,
				       mmap_data      : 1,
				       sample_id_all  : 1,
				       exclude_host   : 1,
				       exclude_guest  : 1,
				       __reserved_1   : 43;

				union {
					Yep32u wakeup_events;
					Yep32u wakeup_watermark;
				};

				Yep32u bp_type;
				union {
					Yep64u bp_addr;
					Yep64u config1;
				};
				union {
					Yep64u bp_len;
					Yep64u config2;
				};
			};
		#endif
	#else
		#include <ucontext.h>
		#include <linux/perf_event.h>
	#endif
#elif defined(YEP_WINDOWS_OS)
	#include <windows.h>
#endif

#if defined(YEP_X86_CPU) || defined(YEP_X64_CPU)
	#if defined(YEP_GNU_COMPILER) || defined(YEP_CLANG_COMPILER)
		#if defined(YEP_X86_CPU) && defined(YEP_PIC)
			static YEP_INLINE void __cpuid(int CPUInfo[4], int InfoType) {
				CPUInfo[0] = InfoType;
				asm volatile (
					"movl %%ebx, %%edi;"
					"cpuid;"
					"xchgl %%ebx, %%edi;"
					:"+a" (CPUInfo[0]), "=D" (CPUInfo[1]), "=c" (CPUInfo[2]), "=d" (CPUInfo[3])
					:
					:
				);
			}

			static YEP_INLINE void __cpuidex(int CPUInfo[4], int InfoType, int ECXValue) {
				CPUInfo[0] = InfoType;
				CPUInfo[2] = ECXValue;
				asm volatile (
					"movl %%ebx, %%edi;"
					"cpuid;"
					"xchgl %%ebx, %%edi;"
					:"+a" (CPUInfo[0]), "=D" (CPUInfo[1]), "+c" (CPUInfo[2]), "=d" (CPUInfo[3])
					:
					:
				);
			}
		#else
			static YEP_INLINE void __cpuid(int CPUInfo[4], int InfoType) {
				CPUInfo[0] = InfoType;
				asm volatile (
					"cpuid;"
					:"+a" (CPUInfo[0]), "=b" (CPUInfo[1]), "=c" (CPUInfo[2]), "=d" (CPUInfo[3])
					:
					:
				);
			}

			static YEP_INLINE void __cpuidex(int CPUInfo[4], int InfoType, int ECXValue) {
				CPUInfo[0] = InfoType;
				CPUInfo[2] = ECXValue;
				asm volatile (
					"cpuid;"
					:"+a" (CPUInfo[0]), "=b" (CPUInfo[1]), "+c" (CPUInfo[2]), "=d" (CPUInfo[3])
					:
					:
				);
			}
		#endif
			
		static YEP_INLINE Yep64u _xgetbv(Yep32u ext_ctrl_reg) {
			Yep32u lo, hi;
			asm volatile (
				"xgetbv"
				: "=a"(lo), "=d"(hi)
				: "c"(ext_ctrl_reg)
				:
			);
			return (Yep64u(hi) << 32) | Yep64u(lo);
		}
	#elif defined(YEP_MICROSOFT_COMPILER) || defined(YEP_INTEL_COMPILER)
		#if defined(YEP_MICROSOFT_COMPILER) && _MSC_FULL_VER < 150030729
			// __cpuidex intrinsic is not suppored until Visual Studio 2008 SP1
			#pragma section(".text")
		
			#if defined(YEP_X86_CPU)
				// fastcall: first argument in ecx, second in edx, third in [esp + 4]
				
				__declspec(allocate(".text")) static const char __cpuidex_bytecode[] =
					"\x53\x56\x8B\x74\x24\x0C\x89\xD0\x0F\xA2\x89\x06\x89\x5E\x04\x89";

			#else
				// x64: first argument in ecx, second in edx, third in r8
				
				__declspec(allocate(".text")) static const char __cpuidex_bytecode[] =
					"\x53\x89\xD0\x0F\xA2\x41\x89\x00\x41\x89\x58\x04\x41\x89\x48\x08\x41\x89\x50\x0C\x5B\xC3";

			#endif
				
			typedef void(__fastcall *CpuidexPointer)(int, int, int[4]);

			static inline void __cpuidex(int CPUInfo[4], int InfoType, int ECXValue) {
				(CpuidexPointer(&__cpuidex_bytecode))(ECXValue, InfoType, CPUInfo);
			}
		
		#endif
		#if defined(YEP_MICROSOFT_COMPILER) && _MSC_FULL_VER < 160040219
			// _xgetbv intrinsic is not supported until Visual Studio 2010 SP1
			#pragma section(".text")
		
			#if defined(YEP_X86_CPU)
				// fastcall: first argument in ecx, second in edx, third in [esp + 4]
				
				__declspec(allocate(".text")) static const char _xgetbv_bytecode[] = 
					"\x0F\x01\xD0\xC3";

			#else
				// x64: first argument in ecx, second in edx, third in r8
				
				__declspec(allocate(".text")) static const char _xgetbv_bytecode[] = 
					"\x0F\x01\xD0\x48\xC1\xE2\x20\x48\x09\xD0\xC3";

			#endif
				
			typedef Yep64u(__fastcall *XgetbvPointer)(Yep32u);
			typedef void(__fastcall *CpuidexPointer)(int, int, int[4]);

			static YEP_INLINE Yep64u _xgetbv(Yep32u ext_ctrl_reg) {
				return (XgetbvPointer(&_xgetbv_bytecode))(ext_ctrl_reg);
			}
		#else
			#include <intrin.h>
			// Visual Stidio 2010 SP1: _xgetbv intrinsic is supported, but not declared
			extern "C" unsigned __int64 __cdecl _xgetbv(unsigned int ext_ctrl_reg);
			#pragma intrinsic(_xgetbv)
		#endif
	#endif
#endif

struct CacheLevelInfo {
	CacheLevelInfo() :
		cacheSize(0),
		lineSize(0),
		associativity(0),
		isUnified(false)
	{
	}
			
	Yep32u cacheSize;
	Yep16u lineSize;
	Yep8u associativity;
	Yep8u isUnified;
};

struct TraceCacheInfo {
	TraceCacheInfo() :
		microops(0),
		associativity(0)
	{
	}
	
	Yep32u microops;
	Yep32u associativity;
};

struct CacheHierarchyInfo {
	CacheHierarchyInfo() :
		prefetchLineSize(0),
		clflushLineSize(0)
	{
	}
			
	CacheLevelInfo L1ICacheInfo;
	CacheLevelInfo L1DCacheInfo;
	CacheLevelInfo L2CacheInfo;
	CacheLevelInfo L3CacheInfo;
	TraceCacheInfo traceCacheInfo;
	Yep16u prefetchLineSize;
	Yep16u clflushLineSize;
};

#if defined(YEP_X86_CPU) || defined(YEP_X64_CPU)
	struct X86ModelInfo {
		X86ModelInfo() :
			baseModel(0xFFu),
			baseFamily(0xFFu),
			stepping(0xFFu),
			extModel(0xFFu),
			extFamily(0xFFu),
			processorType(0xFFu)
		{
		}

		Yep16u model;
		Yep16u family;
		
		Yep8u baseModel;
		Yep8u baseFamily;
		Yep8u stepping;
		Yep8u extModel;
		Yep8u extFamily;
		Yep8u processorType;
	};

	static X86ModelInfo _modelInfo;
#elif defined(YEP_ARM_CPU) && defined(YEP_LINUX_OS)
	struct ARMCpuInfo {
		ARMCpuInfo() :
			featureSwp(false),
			featureHalf(false),
			featureThumb(false),
			feature26bit(false),
			featureFastmult(false),
			featureFpa(false),
			featureVfp(false),
			featureEdsp(false),
			featureJava(false),
			featureIwmmxt(false),
			featureCrunch(false),
			featureThumbee(false),
			featureNeon(false),
			featureVfpv3(false),
			featureVfpv3d16(false),
			featureTls(false),
			featureVfpv4(false),
			featureIdiva(false),
			featureIdivt(false),
			featuresValid(false),
			processors(0),
			cpuImplementer(0xFFFFFFFFu),
			cpuVariant(0xFFFFFFFFu),
			cpuPart(0xFFFFFFFFu),
			cpuRevision(0xFFFFFFFFu),
			cpuArchitectureVersion(0),
			cpuArchitectureT(false),
			cpuArchitectureE(false),
			cpuArchitectureJ(false),
			iSize(0xFFFFFFFFu),
			iAssoc(0xFFFFFFFFu),
			iLineLength(0xFFFFFFFFu),
			iSets(0xFFFFFFFFu),
			dSize(0xFFFFFFFFu),
			dAssoc(0xFFFFFFFFu),
			dLineLength(0xFFFFFFFFu),
			dSets(0xFFFFFFFFu)
		{
		}

		YepBoolean featureSwp;
		YepBoolean featureHalf;
		YepBoolean featureThumb;
		YepBoolean feature26bit;
		YepBoolean featureFastmult;
		YepBoolean featureFpa;
		YepBoolean featureVfp;
		YepBoolean featureEdsp;
		YepBoolean featureJava;
		YepBoolean featureIwmmxt;
		YepBoolean featureCrunch;
		YepBoolean featureThumbee;
		YepBoolean featureNeon;
		YepBoolean featureVfpv3;
		YepBoolean featureVfpv3d16;
		YepBoolean featureTls;
		YepBoolean featureVfpv4;
		YepBoolean featureIdiva;
		YepBoolean featureIdivt;
		YepBoolean featuresValid;

		Yep32u processors;
		Yep32u cpuImplementer;
		Yep32u cpuVariant;
		Yep32u cpuPart;
		Yep32u cpuRevision;
		Yep32u cpuArchitectureVersion;
		YepBoolean cpuArchitectureT;
		YepBoolean cpuArchitectureE;
		YepBoolean cpuArchitectureJ;
		
		Yep32u iSize;
		Yep32u iAssoc;
		Yep32u iLineLength;
		Yep32u iSets;
		Yep32u dSize;
		Yep32u dAssoc;
		Yep32u dLineLength;
		Yep32u dSets;
	};
	
	static ARMCpuInfo _cpuInfo;
#elif defined(YEP_MIPS_CPU) && defined(YEP_LINUX_OS)
	struct MIPSCpuInfo {
		MIPSCpuInfo() :
			aseMips16(false),
			aseMdmx(false),
			aseMips3d(false),
			aseSmartmips(false),
			aseDsp(false),
			aseDsp2(false),
			aseMt(false),
			aseMxu(false),
			hasFpu(false),
			hasIngenicMediaExtensions(false),
			hasWaitInstruction(false),
			tlbEntries(0),
			processors(0)
		{
		}

		YepBoolean aseMips16;
		YepBoolean aseMdmx;
		YepBoolean aseMips3d;
		YepBoolean aseSmartmips;
		YepBoolean aseDsp;
		YepBoolean aseDsp2;
		YepBoolean aseMt;
		YepBoolean aseMxu;
		
		YepBoolean hasFpu;
		YepBoolean hasIngenicMediaExtensions;
		YepBoolean hasWaitInstruction;

		Yep32u tlbEntries;
		Yep32u processors;
	};
	
	static MIPSCpuInfo _cpuInfo;
#endif


Yep64u _isaFeatures = YepIsaFeaturesDefault;
Yep64u _simdFeatures = YepSimdFeaturesDefault;
Yep64u _systemFeatures = YepSystemFeaturesDefault;
const YepCpuMicroarchitecture* _dispatchList = YEP_NULL_POINTER;
static Yep32u _logicalCoresCount = 1;
static YepCpuVendor _vendor = YepCpuVendorUnknown;
static YepCpuMicroarchitecture _microarchitecture = YepCpuMicroarchitectureUnknown;
static CacheHierarchyInfo _cache;

YepStatus yepLibrary_GetCpuVendor(YepCpuVendor *vendorPointer) {
	if YEP_UNLIKELY(vendorPointer == YEP_NULL_POINTER) {
		return YepStatusNullPointer;
	}
	*vendorPointer = _vendor;
	return YepStatusOk;
}

YepStatus yepLibrary_GetCpuArchitecture(YepCpuArchitecture *architecturePointer) {
	if YEP_UNLIKELY(architecturePointer == YEP_NULL_POINTER) {
		return YepStatusNullPointer;
	}
	#if defined(YEP_X86_CPU) || defined(YEP_X64_CPU)
		*architecturePointer = YepCpuArchitectureX86;
	#elif defined(YEP_ARM_CPU)
		*architecturePointer = YepCpuArchitectureARM;
	#elif defined(YEP_MIPS_CPU)
		*architecturePointer = YepCpuArchitectureMIPS;
	#elif defined(YEP_POWERPC_CPU)
		*architecturePointer = YepCpuArchitecturePowerPC;
	#elif defined(YEP_IA64_CPU)
		*architecturePointer = YepCpuArchitectureIA64;
	#elif defined(YEP_SPARC_CPU)
		*architecturePointer = YepCpuArchitectureSPARC;
	#else
		#error "Unsupported processor architecture"
	#endif
	return YepStatusOk;
}

YepStatus yepLibrary_GetCpuMicroarchitecture(YepCpuMicroarchitecture *microarchitecturePointer) {
	if YEP_UNLIKELY(microarchitecturePointer == YEP_NULL_POINTER) {
		return YepStatusNullPointer;
	}
	*microarchitecturePointer = _microarchitecture;
	return YepStatusOk;
}

YepStatus yepLibrary_GetLogicalCoresCount(Yep32u *logicalCoresCount) {
	if YEP_UNLIKELY(logicalCoresCount == YEP_NULL_POINTER) {
		return YepStatusNullPointer;
	}
	*logicalCoresCount = _logicalCoresCount;
	return YepStatusOk;
}

YepStatus yepLibrary_GetCpuIsaFeatures(Yep64u *isaFeaturesPointer) {
	if YEP_UNLIKELY(isaFeaturesPointer == YEP_NULL_POINTER) {
		return YepStatusNullPointer;
	}
	*isaFeaturesPointer = _isaFeatures;
	return YepStatusOk;
}

YepStatus yepLibrary_GetCpuSimdFeatures(Yep64u *simdFeaturesPointer) {
	if YEP_UNLIKELY(simdFeaturesPointer == YEP_NULL_POINTER) {
		return YepStatusNullPointer;
	}
	*simdFeaturesPointer = _simdFeatures;
	return YepStatusOk;
}

YepStatus yepLibrary_GetCpuSystemFeatures(Yep64u *systemFeaturesPointer) {
	if YEP_UNLIKELY(systemFeaturesPointer == YEP_NULL_POINTER) {
		return YepStatusNullPointer;
	}
	*systemFeaturesPointer = _systemFeatures;
	return YepStatusOk;
}

YepStatus yepLibrary_GetCpuDataCacheSize(Yep32u level, Yep32u *cacheSizePointer) {
	if YEP_UNLIKELY(cacheSizePointer == YEP_NULL_POINTER) {
		return YepStatusNullPointer;
	}
	if YEP_UNLIKELY(yepBuiltin_GetPointerMisalignment(cacheSizePointer, sizeof(Yep32u))) {
		return YepStatusMisalignedPointer;
	}
	if YEP_UNLIKELY(level == 0) {
		return YepStatusInvalidArgument;
	}
	switch (level) {
		case 1:
			*cacheSizePointer = _cache.L1DCacheInfo.cacheSize;
			return YepStatusOk;
		case 2:
			*cacheSizePointer = _cache.L2CacheInfo.cacheSize;
			return YepStatusOk;
		case 3:
			*cacheSizePointer = _cache.L3CacheInfo.cacheSize;
			return YepStatusOk;
		default:
			*cacheSizePointer = 0;
			return YepStatusOk;
	}
}

YepStatus yepLibrary_GetCpuInstructionCacheSize(Yep32u level, Yep32u *cacheSizePointer) {
	if YEP_UNLIKELY(cacheSizePointer == YEP_NULL_POINTER) {
		return YepStatusNullPointer;
	}
	if YEP_UNLIKELY(yepBuiltin_GetPointerMisalignment(cacheSizePointer, sizeof(Yep32u))) {
		return YepStatusMisalignedPointer;
	}
	if YEP_UNLIKELY(level == 0) {
		return YepStatusInvalidArgument;
	}
	switch (level) {
		case 1:
			*cacheSizePointer = _cache.L1ICacheInfo.cacheSize;
			return YepStatusOk;
		case 2:
			*cacheSizePointer = _cache.L2CacheInfo.cacheSize;
			return YepStatusOk;
		case 3:
			*cacheSizePointer = _cache.L3CacheInfo.cacheSize;
			return YepStatusOk;
		default:
			*cacheSizePointer = 0;
			return YepStatusOk;
	}
}

static const YepCpuMicroarchitecture *_yepLibrary_GetMicroarchitectureDispatchList(YepCpuMicroarchitecture microarchitecture) {
#if defined(YEP_X86_CPU)
	static const YepCpuMicroarchitecture dispatchUnknown[] = { YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchP5[] = { YepCpuMicroarchitectureP5, YepCpuMicroarchitectureK5, YepCpuMicroarchitectureKnightsFerry, YepCpuMicroarchitectureKnightsCorner, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchP6[] = { YepCpuMicroarchitectureP6, YepCpuMicroarchitectureDothan, YepCpuMicroarchitectureK7, YepCpuMicroarchitectureK6, YepCpuMicroarchitectureYonah, YepCpuMicroarchitectureConroe, YepCpuMicroarchitecturePenryn, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchWillamette[] = { YepCpuMicroarchitectureWillamette, YepCpuMicroarchitecturePrescott, YepCpuMicroarchitectureYonah, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchPrescott[] = { YepCpuMicroarchitecturePrescott, YepCpuMicroarchitectureWillamette, YepCpuMicroarchitectureYonah, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchDothan[] = { YepCpuMicroarchitectureDothan, YepCpuMicroarchitectureP6, YepCpuMicroarchitectureYonah, YepCpuMicroarchitectureConroe, YepCpuMicroarchitecturePenryn, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchYonah[] = { YepCpuMicroarchitectureYonah, YepCpuMicroarchitectureConroe, YepCpuMicroarchitecturePenryn, YepCpuMicroarchitectureDothan, YepCpuMicroarchitectureP6, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchConroe[] = { YepCpuMicroarchitectureConroe, YepCpuMicroarchitecturePenryn, YepCpuMicroarchitectureNehalem, YepCpuMicroarchitectureSandyBridge, YepCpuMicroarchitectureIvyBridge, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchPenryn[] = { YepCpuMicroarchitecturePenryn, YepCpuMicroarchitectureConroe, YepCpuMicroarchitectureNehalem, YepCpuMicroarchitectureSandyBridge, YepCpuMicroarchitectureIvyBridge, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchBonnell[] = { YepCpuMicroarchitectureBonnell, YepCpuMicroarchitectureSaltwell, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchNehalem[] = { YepCpuMicroarchitectureNehalem, YepCpuMicroarchitecturePenryn, YepCpuMicroarchitectureConroe, YepCpuMicroarchitectureSandyBridge, YepCpuMicroarchitectureIvyBridge, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchSandyBridge[] = { YepCpuMicroarchitectureSandyBridge, YepCpuMicroarchitectureNehalem, YepCpuMicroarchitecturePenryn, YepCpuMicroarchitectureConroe, YepCpuMicroarchitectureIvyBridge, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchSaltwell[] = { YepCpuMicroarchitectureSaltwell, YepCpuMicroarchitectureBonnell, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchIvyBridge[] = { YepCpuMicroarchitectureIvyBridge, YepCpuMicroarchitectureSandyBridge, YepCpuMicroarchitectureNehalem, YepCpuMicroarchitecturePenryn, YepCpuMicroarchitectureConroe, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchHaswell[] = { YepCpuMicroarchitectureHaswell, YepCpuMicroarchitectureIvyBridge, YepCpuMicroarchitectureSandyBridge, YepCpuMicroarchitectureNehalem, YepCpuMicroarchitecturePenryn, YepCpuMicroarchitectureConroe, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchKnightsFerry[] = { YepCpuMicroarchitectureKnightsFerry, YepCpuMicroarchitectureKnightsCorner, YepCpuMicroarchitectureP5, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchKnightsCorner[] = { YepCpuMicroarchitectureKnightsCorner, YepCpuMicroarchitectureKnightsFerry, YepCpuMicroarchitectureP5, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchK5[] = { YepCpuMicroarchitectureK5, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchK6[] = { YepCpuMicroarchitectureK6, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchGeode[] = { YepCpuMicroarchitectureGeode, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchK7[] = { YepCpuMicroarchitectureK7, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchK8[] = { YepCpuMicroarchitectureK8, YepCpuMicroarchitectureNehalem, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchK10[] = { YepCpuMicroarchitectureK10, YepCpuMicroarchitectureNehalem, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchBobcat[] = { YepCpuMicroarchitectureBobcat, YepCpuMicroarchitectureNehalem, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchBulldozer[] = { YepCpuMicroarchitectureBulldozer, YepCpuMicroarchitectureNehalem, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchPiledriver[] = { YepCpuMicroarchitecturePiledriver, YepCpuMicroarchitectureNehalem, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchJaguar[] = { YepCpuMicroarchitectureJaguar, YepCpuMicroarchitectureNehalem, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchSteamroller[] = { YepCpuMicroarchitectureSteamroller, YepCpuMicroarchitectureNehalem, YepCpuMicroarchitectureUnknown };
	
	switch (microarchitecture) {
		case YepCpuMicroarchitectureP5:
			return dispatchP5;
		case YepCpuMicroarchitectureP6:
			return dispatchP6;
		case YepCpuMicroarchitectureWillamette:
			return dispatchWillamette;
		case YepCpuMicroarchitecturePrescott:
			return dispatchPrescott;
		case YepCpuMicroarchitectureDothan:
			return dispatchDothan;
		case YepCpuMicroarchitectureYonah:
			return dispatchYonah;
		case YepCpuMicroarchitectureConroe:
			return dispatchConroe;
		case YepCpuMicroarchitecturePenryn:
			return dispatchPenryn;
		case YepCpuMicroarchitectureBonnell:
			return dispatchBonnell;
		case YepCpuMicroarchitectureNehalem:
			return dispatchNehalem;
		case YepCpuMicroarchitectureSandyBridge:
			return dispatchSandyBridge;
		case YepCpuMicroarchitectureSaltwell:
			return dispatchSaltwell;
		case YepCpuMicroarchitectureIvyBridge:
			return dispatchIvyBridge;
		case YepCpuMicroarchitectureHaswell:
			return dispatchHaswell;
		case YepCpuMicroarchitectureKnightsFerry:
			return dispatchKnightsFerry;
		case YepCpuMicroarchitectureKnightsCorner:
			return dispatchKnightsCorner;
		case YepCpuMicroarchitectureK5:
			return dispatchK5;
		case YepCpuMicroarchitectureK6:
			return dispatchK6;
		case YepCpuMicroarchitectureGeode:
			return dispatchGeode;
		case YepCpuMicroarchitectureK7:
			return dispatchK7;
		case YepCpuMicroarchitectureK8:
			return dispatchK8;
		case YepCpuMicroarchitectureK10:
			return dispatchK10;
		case YepCpuMicroarchitectureBobcat:
			return dispatchBobcat;
		case YepCpuMicroarchitectureBulldozer:
			return dispatchBulldozer;
		case YepCpuMicroarchitecturePiledriver:
			return dispatchPiledriver;
		case YepCpuMicroarchitectureJaguar:
			return dispatchJaguar;
		case YepCpuMicroarchitectureSteamroller:
			return dispatchSteamroller;
		default:
			return dispatchUnknown;
	}
#elif defined(YEP_X64_CPU)
	static const YepCpuMicroarchitecture dispatchUnknown[] = { YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchPrescott[] = { YepCpuMicroarchitecturePrescott, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchConroe[] = { YepCpuMicroarchitectureConroe, YepCpuMicroarchitecturePenryn, YepCpuMicroarchitectureNehalem, YepCpuMicroarchitectureSandyBridge, YepCpuMicroarchitectureIvyBridge, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchPenryn[] = { YepCpuMicroarchitecturePenryn, YepCpuMicroarchitectureConroe, YepCpuMicroarchitectureNehalem, YepCpuMicroarchitectureSandyBridge, YepCpuMicroarchitectureIvyBridge, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchBonnell[] = { YepCpuMicroarchitectureBonnell, YepCpuMicroarchitectureSaltwell, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchNehalem[] = { YepCpuMicroarchitectureNehalem, YepCpuMicroarchitecturePenryn, YepCpuMicroarchitectureConroe, YepCpuMicroarchitectureSandyBridge, YepCpuMicroarchitectureIvyBridge, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchSandyBridge[] = { YepCpuMicroarchitectureSandyBridge, YepCpuMicroarchitectureNehalem, YepCpuMicroarchitecturePenryn, YepCpuMicroarchitectureConroe, YepCpuMicroarchitectureIvyBridge, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchSaltwell[] = { YepCpuMicroarchitectureSaltwell, YepCpuMicroarchitectureBonnell, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchIvyBridge[] = { YepCpuMicroarchitectureIvyBridge, YepCpuMicroarchitectureSandyBridge, YepCpuMicroarchitectureNehalem, YepCpuMicroarchitecturePenryn, YepCpuMicroarchitectureConroe, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchHaswell[] = { YepCpuMicroarchitectureHaswell, YepCpuMicroarchitectureIvyBridge, YepCpuMicroarchitectureSandyBridge, YepCpuMicroarchitectureNehalem, YepCpuMicroarchitecturePenryn, YepCpuMicroarchitectureConroe, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchKnightsFerry[] = { YepCpuMicroarchitectureKnightsFerry, YepCpuMicroarchitectureKnightsCorner, YepCpuMicroarchitectureP5, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchKnightsCorner[] = { YepCpuMicroarchitectureKnightsCorner, YepCpuMicroarchitectureKnightsFerry, YepCpuMicroarchitectureP5, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchK8[] = { YepCpuMicroarchitectureK8, YepCpuMicroarchitectureNehalem, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchK10[] = { YepCpuMicroarchitectureK10, YepCpuMicroarchitectureNehalem, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchBobcat[] = { YepCpuMicroarchitectureBobcat, YepCpuMicroarchitectureJaguar, YepCpuMicroarchitectureNehalem, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchBulldozer[] = { YepCpuMicroarchitectureBulldozer, YepCpuMicroarchitecturePiledriver, YepCpuMicroarchitectureSteamroller, YepCpuMicroarchitectureIvyBridge, YepCpuMicroarchitectureSandyBridge, YepCpuMicroarchitectureNehalem, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchPiledriver[] = { YepCpuMicroarchitecturePiledriver, YepCpuMicroarchitectureSteamroller, YepCpuMicroarchitectureBulldozer, YepCpuMicroarchitectureIvyBridge, YepCpuMicroarchitectureSandyBridge, YepCpuMicroarchitectureNehalem, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchJaguar[] = { YepCpuMicroarchitectureJaguar, YepCpuMicroarchitectureBobcat, YepCpuMicroarchitectureNehalem, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchSteamroller[] = { YepCpuMicroarchitectureSteamroller, YepCpuMicroarchitecturePiledriver, YepCpuMicroarchitectureBulldozer, YepCpuMicroarchitectureIvyBridge, YepCpuMicroarchitectureSandyBridge, YepCpuMicroarchitectureNehalem, YepCpuMicroarchitectureUnknown };

	switch (microarchitecture) {
		case YepCpuMicroarchitecturePrescott:
			return dispatchPrescott;
		case YepCpuMicroarchitectureConroe:
			return dispatchConroe;
		case YepCpuMicroarchitecturePenryn:
			return dispatchPenryn;
		case YepCpuMicroarchitectureBonnell:
			return dispatchBonnell;
		case YepCpuMicroarchitectureNehalem:
			return dispatchNehalem;
		case YepCpuMicroarchitectureSandyBridge:
			return dispatchSandyBridge;
		case YepCpuMicroarchitectureSaltwell:
			return dispatchSaltwell;
		case YepCpuMicroarchitectureIvyBridge:
			return dispatchIvyBridge;
		case YepCpuMicroarchitectureHaswell:
			return dispatchHaswell;
		case YepCpuMicroarchitectureKnightsFerry:
			return dispatchKnightsFerry;
		case YepCpuMicroarchitectureKnightsCorner:
			return dispatchKnightsCorner;
		case YepCpuMicroarchitectureK8:
			return dispatchK8;
		case YepCpuMicroarchitectureK10:
			return dispatchK10;
		case YepCpuMicroarchitectureBobcat:
			return dispatchBobcat;
		case YepCpuMicroarchitectureBulldozer:
			return dispatchBulldozer;
		case YepCpuMicroarchitecturePiledriver:
			return dispatchPiledriver;
		case YepCpuMicroarchitectureJaguar:
			return dispatchJaguar;
		case YepCpuMicroarchitectureSteamroller: 
			return dispatchSteamroller;
		default:
			return dispatchUnknown;
	}
#elif defined(YEP_ARM_CPU)
	static const YepCpuMicroarchitecture dispatchUnknown[] = { YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchStrongARM[] = { YepCpuMicroarchitectureStrongARM, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchXScale[] = { YepCpuMicroarchitectureXScale, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchARM7[] = { YepCpuMicroarchitectureARM7, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchARM9[] = { YepCpuMicroarchitectureARM9, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchARM11[] = { YepCpuMicroarchitectureARM11, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchCortexA5[] = { YepCpuMicroarchitectureCortexA5, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchCortexA7[] = { YepCpuMicroarchitectureCortexA7, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchCortexA8[] = { YepCpuMicroarchitectureCortexA8, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchCortexA9[] = { YepCpuMicroarchitectureCortexA9, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchCortexA15[] = { YepCpuMicroarchitectureCortexA15, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchScorpion[] = { YepCpuMicroarchitectureScorpion, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchKrait[] = { YepCpuMicroarchitectureKrait, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchPJ1[] = { YepCpuMicroarchitecturePJ1, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchPJ4[] = { YepCpuMicroarchitecturePJ4, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchSwift[] = { YepCpuMicroarchitectureSwift, YepCpuMicroarchitectureUnknown };
	
	switch (microarchitecture) {
		case YepCpuMicroarchitectureStrongARM:
			return dispatchStrongARM;
		case YepCpuMicroarchitectureXScale:
			return dispatchXScale;
		case YepCpuMicroarchitectureARM7:
			return dispatchARM7;
		case YepCpuMicroarchitectureARM9:
			return dispatchARM9;
		case YepCpuMicroarchitectureARM11:
			return dispatchARM11;
		case YepCpuMicroarchitectureCortexA5:
			return dispatchCortexA5;
		case YepCpuMicroarchitectureCortexA7:
			return dispatchCortexA7;
		case YepCpuMicroarchitectureCortexA8:
			return dispatchCortexA8;
		case YepCpuMicroarchitectureCortexA9:
			return dispatchCortexA9;
		case YepCpuMicroarchitectureCortexA15:
			return dispatchCortexA15;
		case YepCpuMicroarchitectureScorpion:
			return dispatchScorpion;
		case YepCpuMicroarchitectureKrait:
			return dispatchKrait;
		case YepCpuMicroarchitecturePJ1:
			return dispatchPJ1;
		case YepCpuMicroarchitecturePJ4:
			return dispatchPJ4;
		case YepCpuMicroarchitectureSwift:
			return dispatchSwift;
		default:
			return dispatchUnknown;
	}
#elif defined(YEP_MIPS_CPU)
	static const YepCpuMicroarchitecture dispatchUnknown[] = { YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchMIPS24K[] = { YepCpuMicroarchitectureMIPS24K, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchMIPS34K[] = { YepCpuMicroarchitectureMIPS34K, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchMIPS74K[] = { YepCpuMicroarchitectureMIPS74K, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchXBurst[] = { YepCpuMicroarchitectureXBurst, YepCpuMicroarchitectureUnknown };
	static const YepCpuMicroarchitecture dispatchXBurst2[] = { YepCpuMicroarchitectureXBurst2, YepCpuMicroarchitectureUnknown };
	
	switch (microarchitecture) {
		case YepCpuMicroarchitectureMIPS24K:
			return dispatchMIPS24K;
		case YepCpuMicroarchitectureMIPS34K:
			return dispatchMIPS34K;
		case YepCpuMicroarchitectureMIPS74K:
			return dispatchMIPS74K;
		case YepCpuMicroarchitectureXBurst:
			return dispatchXBurst;
		case YepCpuMicroarchitectureXBurst2:
			return dispatchXBurst2;
		default:
			return dispatchUnknown;
	}
#else
	static const YepCpuMicroarchitecture dispatchUnknown[] = { YepCpuMicroarchitectureUnknown };
	
	return dispatchUnknown;
#endif
}

#if defined(YEP_LINUX_OS)
	static YepStatus _yepLibrary_InitLinuxLogicalCoresCount();
#elif defined(YEP_WINDOWS_OS)
	static YepStatus _yepLibrary_InitWindowsLogicalCoresCount();
#endif

#if defined(YEP_X86_CPU) || defined(YEP_X64_CPU)
	static void _yepLibrary_InitX86CpuVendor();
	static void _yepLibrary_InitX86CpuModelInfo();
	static void _yepLibrary_InitX86CpuMicroarchitectureInfo();
	static void _yepLibrary_InitX86CpuIsaInfo();
	static void _yepLibrary_InitX86CpuCacheInfo();
	static void _yepLibrary_DecodeX86CpuCacheDescriptor(Yep8u descriptor);
	static void _yepLibrary_DecodeX86CpuDeterministicCacheParameters(Yep32u cacheType, Yep32u cacheLevel, Yep32u cacheSize, Yep32u lineSize, Yep32u associativity);
#elif defined(YEP_ARM_CPU) && defined(YEP_WINDOWS_OS)
	static void _yepLibrary_InitWindowsARMCpuIsaInfo();
	static void _yepLibrary_InitWindowsARMMicroarchitectureInfo();
	static YepStatus _yepLibrary_InitWindowsARMCacheInfo();
#elif defined(YEP_ARM_CPU) && defined(YEP_LINUX_OS)
	static void _yepLibrary_DecodeLinuxARMCpuPart(const char* infoValueStart, const char* infoValueEnd);
	static void _yepLibrary_DecodeLinuxARMCpuFeatures(const char* infoValueStart, const char* infoValueEnd);
	static void _yepLibrary_DecodeLinuxARMProcessorNumber(const char* infoValueStart, const char* infoValueEnd);
	static void _yepLibrary_DecodeLinuxARMCpuVariant(const char* infoValueStart, const char* infoValueEnd);
	static void _yepLibrary_DecodeLinuxARMCpuRevision(const char* infoValueStart, const char* infoValueEnd);
	static void _yepLibrary_DecodeLinuxARMCpuImplementer(const char* infoValueStart, const char* infoValueEnd);
	static void _yepLibrary_DecodeLinuxARMCpuArchitecture(const char* infoValueStart, const char* infoValueEnd);
	static void _yepLibrary_DecodeLinuxARMCacheNumber(const char* infoValueStart, const char* infoValueEnd, Yep32u& cacheNumber);
	static void _yepLibrary_DecodeLinuxARMCpuInfoLine(const char* lineStart, const char* lineEnd);
	static void _yepLibrary_DecodeLinuxARMCpuInfo();
	static YepStatus _yepLibrary_DetectLinuxARMCycleCounterAccess();

	static void _yepLibrary_ProbeSignalHandler(int, siginfo_t *, void* ptr) {
		ucontext_t* ctx = (ucontext_t*)ptr;
		ctx->uc_mcontext.arm_pc += 4; // All probed instructions are four bytes long both in ARM and Thumb-2 mode
		ctx->uc_mcontext.arm_r0 = 1; // To indicate that the signal handler was called
	}

	static void _yepLibrary_ReadCoprocessorSignalHandler(int, siginfo_t *, void* ptr) {
		ucontext_t* ctx = (ucontext_t*)ptr;
		ctx->uc_mcontext.arm_pc += 4; // All read coprocessor instructions are four bytes long both in ARM and Thumb-2 mode
		ctx->uc_mcontext.arm_r0 = 0; // Emulate reading all zeroes from coprocessor register
	}
#elif defined(YEP_MIPS_CPU) && defined(YEP_LINUX_OS)
	static void _yepLibrary_DecodeLinuxMIPSSystemType(const char* infoValueStart, const char* infoValueEnd);
	static void _yepLibrary_DecodeLinuxMIPSProcessorNumber(const char* infoValueStart, const char* infoValueEnd);
	static void _yepLibrary_DecodeLinuxMIPSCpuModel(const char* infoValueStart, const char* infoValueEnd);
	static void _yepLibrary_DecodeLinuxMIPSWaitInstruction(const char* infoValueStart, const char* infoValueEnd);
	static void _yepLibrary_DecodeLinuxMIPSTlbEntries(const char* infoValueStart, const char* infoValueEnd);
	static void _yepLibrary_DecodeLinuxMIPSAsesImplemented(const char* infoValueStart, const char* infoValueEnd);
	static void _yepLibrary_DecodeLinuxMIPSCpuInfoLine(const char* lineStart, const char* lineEnd);
	static void _yepLibrary_DecodeLinuxMIPSCpuInfo();

	static void _yepLibrary_ProbeSignalHandler(int, siginfo_t *, void* ptr) {
		ucontext_t* ctx = (ucontext_t*)ptr;
		ctx->uc_mcontext.sc_pc += 4; // Skip unsupported instruction
		ctx->uc_mcontext.sc_regs[2] = 1; // Change $2 = $v0 to 1 to indicate that the signal handler was called
	}
#elif defined(YEP_IA64_CPU)
#endif

#if (defined(YEP_ARM_CPU) || defined(YEP_MIPS_CPU)) && defined(YEP_LINUX_OS)
	static YepStatus _yepLibrary_DetectLinuxPerfEventSupport();
	static YepStatus _yepLibrary_InitLinuxCpuInfo();

	static bool _yepLibrary_strncmp(const char* stringX, const char* stringY, size_t length);
	static void _yepLibrary_memcpy(void* destination, const void* source, size_t length);
	static void _yepLibrary_memset(void* destination, int value, size_t length);

	static Yep32u _yepLibrary_ProbeInstruction(Yep32u (*ProbeFunction)()) {
		struct kernel_sigaction oldSigillAction;
		struct kernel_sigaction probeSigillAction;
		_yepLibrary_memset(&probeSigillAction, 0, sizeof(probeSigillAction));
		probeSigillAction.sa_sigaction = &_yepLibrary_ProbeSignalHandler;
		// Needs Linux >= 2.2
		probeSigillAction.sa_flags = SA_ONSTACK | SA_RESTART | SA_SIGINFO;
		int sigactionResult = yepSyscall_rt_sigaction(SIGILL, &probeSigillAction, &oldSigillAction, KERNEL_NSIG / 8);
		if (sigactionResult == 0) {
			const Yep32u probeResult = ProbeFunction();
			yepSyscall_rt_sigaction(SIGILL, &oldSigillAction, NULL, KERNEL_NSIG / 8);
			return probeResult;
		} else {
			return 2;
		}
	}

	#if defined(YEP_ARM_CPU)
		static Yep64u _yepLibrary_ReadCoprocessor(Yep64u (*ReadFunction)()) {
			struct kernel_sigaction oldSigillAction;
			struct kernel_sigaction readSigillAction;
			_yepLibrary_memset(&readSigillAction, 0, sizeof(readSigillAction));
			readSigillAction.sa_sigaction = &_yepLibrary_ReadCoprocessorSignalHandler;
			// Needs Linux >= 2.2
			readSigillAction.sa_flags = SA_ONSTACK | SA_RESTART | SA_SIGINFO;
			int sigactionResult = yepSyscall_rt_sigaction(SIGILL, &readSigillAction, &oldSigillAction, KERNEL_NSIG / 8);
			if (sigactionResult == 0) {
				const Yep64u probeResult = ReadFunction();
				yepSyscall_rt_sigaction(SIGILL, &oldSigillAction, NULL, KERNEL_NSIG / 8);
				return probeResult;
			} else {
				return 0ull;
			}
		}
	#endif
#endif

const Yep32u LinuxPerfEventCpuCounterStateMagic = 0xCAB06128u;
const Yep32u ArmCP15CpuCounterStateMagic = 0xF2C5E274u;

YepStatus yepLibrary_GetCpuCyclesAcquire(Yep64u *statePointer) {
	if YEP_UNLIKELY(statePointer == YEP_NULL_POINTER) {
		return YepStatusNullPointer;
	}
#if defined(YEP_X86_CPU) || defined(YEP_X64_CPU)
	#if defined(YEP_X86_CPU)
		if YEP_UNLIKELY((_systemFeatures & YepSystemFeatureCycleCounter) == 0) {
			return YepStatusUnsupportedHardware;
		}
	#endif
	#if defined(YEP_MICROSOFT_COMPILER)
		Yep64u state;
		do {
			int registers[4];
			__cpuid(registers, 0);
			_ReadWriteBarrier(); // Prevent compiler from reordering CPUID and RDTSC
			state = __rdtsc();
		} while (state == 0);
		*statePointer = state;
		return YepStatusOk;
	#else
		Yep64u state;
		do {
			Yep32u lo, hi;
			#if defined(YEP_X64_CPU)
				asm volatile (
					"xorl %%eax,%%eax;"
					"cpuid;"
					"rdtsc;"
					: "=a" (lo), "=d" (hi)
					:
					: "%rbx", "%rcx"
				);
			#else
				#if defined(YEP_PIC)
					asm volatile (
						"xorl %%eax,%%eax;"
						"movl %%ebx, %%esi;"
						"cpuid;"
						"rdtsc;"
						"movl %%esi, %%ebx;"
						: "=a" (lo), "=d" (hi)
						:
						: "%esi", "%ecx"
					);
				#else
					asm volatile (
						"xorl %%eax,%%eax;"
						"cpuid;"
						"rdtsc;"
						: "=a" (lo), "=d" (hi)
						:
						: "%ebx", "%ecx"
					);
				#endif
			#endif
			state = yepBuiltin_CombineParts_32u32u_64u(hi, lo);
		} while (state == 0);
		*statePointer = state;
		return YepStatusOk;
	#endif
#elif defined(YEP_IA64_CPU)
	#if defined(YEP_MICROSOFT_COMPILER)
		Yep64u state;
		do {
			state = __getReg(__REG_IA64_ApITC);
		} while (state == 0);
		*statePointer = state;
		return YepStatusOk;
	#else
		#error "Not yet implemented for this combination of compiler and architecture (IA64)"
	#endif
#elif defined(YEP_ARM_CPU) || defined(YEP_MIPS_CPU)
	#if defined(YEP_LINUX_OS)
		const Yep64u cycleCounterMask = YepSystemFeatureCycleCounter | YepSystemFeatureCycleCounter64Bit;
		const Yep64u cycleCounterPerfEvent = YepSystemFeatureCycleCounter | YepSystemFeatureCycleCounter64Bit;
		const Yep64u cycleCounterCP15 = YepSystemFeatureCycleCounter;
		if YEP_LIKELY((_systemFeatures & cycleCounterMask) == cycleCounterPerfEvent) {
			struct perf_event_attr perf_event_attributes;
			_yepLibrary_memset(&perf_event_attributes, 0, sizeof(perf_event_attributes));
			perf_event_attributes.size = sizeof(perf_event_attributes);
			perf_event_attributes.type = PERF_TYPE_HARDWARE;
			perf_event_attributes.config = PERF_COUNT_HW_CPU_CYCLES;
			perf_event_attributes.disabled = 1;

			const pid_t processId = 0; // Current process
			const int cpu = -1; // Any CPU
			const int groupFileDescriptor = -1; // New group
			const int flags = 0; // Default
			const int perfCounter = yepSyscall_perf_event_open(&perf_event_attributes, processId, cpu, groupFileDescriptor, flags);

			if YEP_LIKELY(perfCounter >= 0) {
				const int ioctlResetSyscallResult = yepSyscall_ioctl(perfCounter, PERF_EVENT_IOC_RESET);
				if YEP_LIKELY(ioctlResetSyscallResult == 0) {
					const int ioctlEnableSyscallResult = yepSyscall_ioctl(perfCounter, PERF_EVENT_IOC_ENABLE);
					if YEP_LIKELY(ioctlEnableSyscallResult == 0) {
						const Yep64u state = yepBuiltin_CombineParts_32u32u_64u(LinuxPerfEventCpuCounterStateMagic, Yep32u(perfCounter));
						*statePointer = state;
						return YepStatusOk;
					}
				}

				yepSyscall_close(perfCounter);
				return YepStatusSystemError;
			} else {
				return YepStatusSystemError;
			}
	#if defined(YEP_ARM_CPU)
		} else if YEP_UNLIKELY((_systemFeatures & cycleCounterMask) == cycleCounterCP15) {
			const Yep64u pmccntrReadResult = _yepLibrary_ReadCoprocessor(&_yepLibrary_ReadPMCCNTR);
			const Yep32u pmccntrReadSuccess = yepBuiltin_GetLowPart_64u_32u(pmccntrReadResult);
			if YEP_LIKELY(pmccntrReadSuccess) {
				const Yep32u pmccntr = yepBuiltin_GetHighPart_64u_32u(pmccntrReadResult);
				const Yep64u state = yepBuiltin_CombineParts_32u32u_64u(ArmCP15CpuCounterStateMagic, pmccntr);
				*statePointer = state;
				return YepStatusOk;
			} else {
				return YepStatusSystemError;
			}
	#endif
		} else {
			return YepStatusUnsupportedSoftware;
		}
	#elif defined(YEP_WINDOWS_OS)
		return YepStatusUnsupportedSoftware;
	#else
		#error "Not yet supported combination of OS and CPU"
	#endif
#else
	#error "Not yet implemented for this combination of compiler and architecture"
#endif
}

YepStatus yepLibrary_GetCpuCyclesRelease(Yep64u* statePointer, Yep64u* ticksPointer) {
	if YEP_UNLIKELY(statePointer == YEP_NULL_POINTER) {
		return YepStatusNullPointer;
	}
	if YEP_UNLIKELY(ticksPointer == YEP_NULL_POINTER) {
		return YepStatusNullPointer;
	}
#if defined(YEP_X86_CPU) || defined(YEP_X64_CPU)
	#if defined(YEP_MICROSOFT_COMPILER)
		if YEP_LIKELY(_isaFeatures & YepX86IsaFeatureRdtscp) {
			unsigned int aux;
			const Yep64u endTicks = __rdtscp(&aux);
			_ReadWriteBarrier(); // Prevent compiler from reordering RDTSCP and reading the state
			const Yep64u startTicks = *statePointer;
			*statePointer = 0;
			if YEP_LIKELY(startTicks != 0) {
				const Yep64u elapsedTicks = endTicks - startTicks;
				*ticksPointer = elapsedTicks;
				return YepStatusOk;
			} else {
				return YepStatusInvalidState;
			}
		} else {
			#if defined(YEP_X86_CPU)
			if YEP_LIKELY(_systemFeatures & YepSystemFeatureCycleCounter) {
			#endif
				int registers[4];
				__cpuid(registers, 0);
				_ReadWriteBarrier(); // Prevent compiler from reordering CPUID and RDTSC
				const Yep64u endTicks = __rdtsc();
				_ReadWriteBarrier(); // Prevent compiler from reordering RDTSC and reading the state
				const Yep64u startTicks = *statePointer;
				*statePointer = 0;
				if YEP_LIKELY(startTicks != 0) {
					const Yep64u elapsedTicks = endTicks - startTicks;
					*ticksPointer = elapsedTicks;
					return YepStatusOk;
				} else {
					return YepStatusInvalidState;
				}
			#if defined(YEP_X86_CPU)
			} else {
				return YepStatusUnsupportedHardware;
			}
			#endif
		}
	#else
		Yep32u lo, hi;
		if YEP_LIKELY(_isaFeatures & YepX86IsaFeatureRdtscp) {
			#if defined(YEP_X64_CPU)
				asm volatile (
					"rdtscp;"
					: "=a" (lo), "=d" (hi)
					:
					: "%rcx"
				);
			#else
				asm volatile (
					"rdtscp;"
					: "=a" (lo), "=d" (hi)
					:
					: "%ecx"
				);
			#endif
		} else {
			#if defined(YEP_X64_CPU)
				asm volatile (
					"xorl %%eax,%%eax;"
					"cpuid;"
					"rdtsc;"
					: "=a" (lo), "=d" (hi)
					:
					: "%rbx", "%rcx"
				);
			#else
				if YEP_LIKELY(_systemFeatures & YepSystemFeatureCycleCounter) {
					#if defined(YEP_PIC)
						asm volatile (
							"xorl %%eax,%%eax;"
							"movl %%ebx, %%esi;"
							"cpuid;"
							"rdtsc;"
							"movl %%esi, %%ebx;"
							: "=a" (lo), "=d" (hi)
							:
							: "%esi", "%ecx"
						);
					#else
						asm volatile (
							"xorl %%eax,%%eax;"
							"cpuid;"
							"rdtsc;"
							: "=a" (lo), "=d" (hi)
							:
							: "%ebx", "%ecx"
						);
					#endif
				} else {
					return YepStatusUnsupportedHardware;
				}
			#endif
		}
		const Yep64u endTicks = yepBuiltin_CombineParts_32u32u_64u(hi, lo);
		const Yep64u startTicks = *statePointer;
		*statePointer = 0;
		if YEP_LIKELY(startTicks != 0) {
			const Yep64u elapsedTicks = endTicks - startTicks;
			*ticksPointer = elapsedTicks;
			return YepStatusOk;
		} else {
			return YepStatusInvalidState;
		}
	#endif
#elif defined(YEP_IA64_CPU)
	#if defined(YEP_MICROSOFT_COMPILER)
		const Yep64u endTicks = __getReg(__REG_IA64_ApITC);
		const Yep64u startTicks = *statePointer;
		*statePointer = 0;
		if YEP_LIKELY(startTicks != 0) {
			const Yep64u elapsedTicks = endTicks - startTicks;
			*ticksPointer = elapsedTicks;
			return YepStatusOk;
		} else {
			return YepStatusInvalidState;
		}
	#else
		#error "Not yet implemented for this combination of compiler and architecture (IA64)"
	#endif
#elif defined(YEP_ARM_CPU) || defined(YEP_MIPS_CPU)
	#if defined(YEP_LINUX_OS)
		const Yep64u cycleCounterMask = YepSystemFeatureCycleCounter | YepSystemFeatureCycleCounter64Bit;
		const Yep64u cycleCounterPerfEvent = YepSystemFeatureCycleCounter | YepSystemFeatureCycleCounter64Bit;
		const Yep64u cycleCounterCP15 = YepSystemFeatureCycleCounter;
		if YEP_LIKELY((_systemFeatures & cycleCounterMask) == cycleCounterPerfEvent) {
			const Yep64u state = *statePointer;
			*statePointer = 0;
			if YEP_LIKELY(yepBuiltin_GetHighPart_64u_32u(state) == LinuxPerfEventCpuCounterStateMagic) {
				const int perfCounter = int(yepBuiltin_GetLowPart_64u_32u(state));

				const int syscallIoctlDisableResult = yepSyscall_ioctl(perfCounter, PERF_EVENT_IOC_DISABLE);
				if YEP_LIKELY(syscallIoctlDisableResult == 0) {
					Yep64u elapsedTicks;
					const ssize_t bytesRead = yepSyscall_read(perfCounter, &elapsedTicks, sizeof(elapsedTicks));
					if YEP_LIKELY(bytesRead == sizeof(elapsedTicks)) {
						const int syscallCloseResult = yepSyscall_close(perfCounter);
						if YEP_LIKELY(syscallCloseResult == 0) {
							*ticksPointer = elapsedTicks;
							return YepStatusOk;
						}
					}
				}

				yepSyscall_close(perfCounter);
				return YepStatusSystemError;
			} else {
				return YepStatusInvalidState;
			}
	#if defined(YEP_ARM_CPU)
		} else if YEP_UNLIKELY((_systemFeatures & cycleCounterMask) == cycleCounterCP15) {
			const Yep64u state = *statePointer;
			*statePointer = 0;
			if YEP_LIKELY(yepBuiltin_GetHighPart_64u_32u(state) == ArmCP15CpuCounterStateMagic) {
				const Yep32u pmccntrStart = yepBuiltin_GetLowPart_64u_32u(state);

				const Yep64u pmccntrReadResult = _yepLibrary_ReadCoprocessor(&_yepLibrary_ReadPMCCNTR);
				const Yep32u pmccntrReadSuccess = yepBuiltin_GetLowPart_64u_32u(pmccntrReadResult);
				if YEP_LIKELY(pmccntrReadSuccess) {
					const Yep32u pmccntrEnd = yepBuiltin_GetHighPart_64u_32u(pmccntrReadResult);
					const Yep32u ticks = pmccntrEnd - pmccntrStart;
					
					const Yep64u pmcrReadResult = _yepLibrary_ReadCoprocessor(&_yepLibrary_ReadPMCR);
					const Yep32u pmcrReadSuccess = yepBuiltin_GetLowPart_64u_32u(pmcrReadResult);
					if YEP_LIKELY(pmcrReadSuccess) {
						const Yep32u pmcr = yepBuiltin_GetHighPart_64u_32u(pmcrReadResult);
						// PMCR[bit 3] = divider (if set then every 64th cycle is counted)
						const Yep32u cycleCounterClockDivider = pmcr & 0x00000008u;
						if YEP_UNLIKELY(cycleCounterClockDivider != 0) {
							*ticksPointer = yepBuiltin_CombineParts_32u32u_64u(ticks >> (32 - 6), ticks << 6);
						} else {
							*ticksPointer = Yep64u(ticks);
						}
						return YepStatusOk;
					}
				}
				return YepStatusSystemError;
			} else {
				return YepStatusInvalidState;
			}
	#endif
		} else {
			*statePointer = 0;
			return YepStatusUnsupportedSoftware;
		}
	#elif defined(YEP_WINDOWS_OS)
		return YepStatusUnsupportedSoftware;
	#else
		#error "Not yet supported combination of OS and CPU"
	#endif
#else
	#error "Not yet implemented for this combination of compiler and architecture"
#endif
}

YepStatus _yepLibrary_InitCpuInfo() {
#if defined(YEP_LINUX_OS)
	_yepLibrary_InitLinuxLogicalCoresCount();
#elif defined(YEP_WINDOWS_OS)
	_yepLibrary_InitWindowsLogicalCoresCount();
#else
	#error "This OS is not supported yet"
#endif	

#if defined(YEP_X86_CPU) || defined(YEP_X64_CPU)
	_yepLibrary_InitX86CpuVendor();
	_yepLibrary_InitX86CpuModelInfo();
	_yepLibrary_InitX86CpuCacheInfo();
	_yepLibrary_InitX86CpuMicroarchitectureInfo();
	_yepLibrary_InitX86CpuIsaInfo();
	_dispatchList = _yepLibrary_GetMicroarchitectureDispatchList(_microarchitecture);
	return YepStatusOk;
#elif defined(YEP_ARM_CPU) && defined(YEP_WINDOWS_OS)
	_yepLibrary_InitWindowsARMCpuIsaInfo();
	_yepLibrary_InitWindowsARMMicroarchitectureInfo();
	YepStatus status = _yepLibrary_InitWindowsARMCacheInfo();
	_dispatchList = _yepLibrary_GetMicroarchitectureDispatchList(_microarchitecture);
	return status;
#elif defined(YEP_ARM_CPU) && defined(YEP_LINUX_OS)
	YepStatus status = _yepLibrary_InitLinuxCpuInfo();
	if (status == YepStatusOk) {
		_yepLibrary_DecodeLinuxARMCpuInfo();
	}
	YepStatus perfEventSupportStatus = _yepLibrary_DetectLinuxPerfEventSupport();
	if YEP_LIKELY(perfEventSupportStatus != YepStatusOk) {
		_yepLibrary_DetectLinuxARMCycleCounterAccess();
	}
	_dispatchList = _yepLibrary_GetMicroarchitectureDispatchList(_microarchitecture);
	return status;
#elif defined(YEP_MIPS_CPU) && defined(YEP_LINUX_OS)
	YepStatus status = _yepLibrary_InitLinuxCpuInfo();
	if (status == YepStatusOk) {
		_yepLibrary_DecodeLinuxMIPSCpuInfo();
	}
	_yepLibrary_DetectLinuxPerfEventSupport();
	_dispatchList = _yepLibrary_GetMicroarchitectureDispatchList(_microarchitecture);
	return status;
#else
	#error "This CPU architecture (perhaps in combination with OS) is not supported yet"
#endif
}

#if defined(YEP_LINUX_OS)

	#pragma pack(push, 1)
	/* This is an incomplete linux_direct structure definition, but it is enough for our purpose */
	struct linux_dirent {
		unsigned long d_ino;
		unsigned long d_off;
		unsigned short d_reclen;
		char d_name[];
	};
	#pragma pack(pop)

	static YepStatus _yepLibrary_InitLinuxLogicalCoresCount() {
		YepStatus status = YepStatusOk;
		Yep32u logicalCoresCount = 0;
		int rootCpuDirectory = yepSyscall_open("/sys/devices/system/cpu", O_RDONLY | O_NONBLOCK | O_DIRECTORY);
		if YEP_UNLIKELY(rootCpuDirectory < 0) {
			return YepStatusSystemError;
		}

		const YepSize bufferSize = 1024;
		Yep8u buffer[bufferSize];
readDirectoryEntries:
		int syscallResult = yepSyscall_getdents(rootCpuDirectory, reinterpret_cast<linux_dirent*>(buffer), bufferSize);
		if YEP_LIKELY(syscallResult > 0) {
			int entriesLength = syscallResult;
			
			const Yep8u* entryPointer = buffer;
			while (entriesLength > 0) {
				const char* filename = (reinterpret_cast<const linux_dirent*>(entryPointer))->d_name;
				unsigned short recordLength = reinterpret_cast<const linux_dirent*>(entryPointer)->d_reclen;
				// Check that the filename has /cpu[0-9]+/ pattern
				if (YEP_UNLIKELY(filename[0] == 'c') && YEP_LIKELY(filename[1] == 'p') && YEP_LIKELY(filename[2] == 'u')) {
					// Filename starts with "cpu"
					if (YEP_LIKELY(filename[3] >= '0') && YEP_LIKELY(filename[3] <= '9')) {
						// Filename start with /cpu[0-9]/
						for (filename = &filename[4]; *filename != 0; filename++) {
							if (YEP_UNLIKELY(*filename < '0') || YEP_UNLIKELY(*filename > '9')) {
								goto nextEntry;
							}
						}
						// Filename matches with /cpu[0-9]+/: increment the number of cores
						logicalCoresCount += 1;
					}
				}
nextEntry:
				entriesLength -= recordLength;
				entryPointer += recordLength;
			}
			goto readDirectoryEntries;
		} else if YEP_UNLIKELY(syscallResult < 0) {
			status = YepStatusSystemError;
		}

		if (logicalCoresCount == 1) {
			// Set this flag only if we know for sure.
			_systemFeatures |= YepSystemFeatureSingleThreaded;
		} else {
			#if defined(YEP_ARM_CPU)
				if (logicalCoresCount == 0) {
					logicalCoresCount = _cpuInfo.processors;
				}
			#endif
			if (logicalCoresCount == 0) {
				logicalCoresCount = 1;
			}
		}
		_logicalCoresCount = logicalCoresCount;

		int closeResult = yepSyscall_close(rootCpuDirectory);
		if (closeResult < 0) {
			status = YepStatusSystemError;
		}
		return status;
	}
#elif defined(YEP_WINDOWS_OS)
	static YepStatus _yepLibrary_InitWindowsLogicalCoresCount() {
		SYSTEM_INFO systemInfo;
		::GetSystemInfo(&systemInfo);
		if (systemInfo.dwNumberOfProcessors != 0) {
			_logicalCoresCount = systemInfo.dwNumberOfProcessors;
			if (systemInfo.dwNumberOfProcessors == 1u) {
				_systemFeatures |= YepSystemFeatureSingleThreaded;
			}
		} else {
			_logicalCoresCount = 1;
		}
		return YepStatusOk;
	}
#endif

#if defined(YEP_X86_CPU) || defined(YEP_X64_CPU)
static void _yepLibrary_InitX86CpuVendor() {
	/// Intel vendor string: "GenuineIntel"
	const Yep32u Genu = 0x756E6547U;
	const Yep32u ineI = 0x49656E69U;
	const Yep32u ntel = 0x6C65746EU;
	/// AMD vendor strings: "AuthenticAMD", "AMDisbetter!", "AMD ISBETTER"
	const Yep32u Auth = 0x68747541U;
	const Yep32u enti = 0x69746E65U;
	const Yep32u cAMD = 0x444D4163U;
	const Yep32u AMDi = 0x69444D41U;
	const Yep32u sbet = 0x74656273U;
	const Yep32u ter = 0x21726574U;
	const Yep32u AMD = 0x20444D41U;
	const Yep32u ISBE = 0x45425349U;
	const Yep32u TTER = 0x52455454U;
	/// VIA (Centaur) vendor strings: "CentaurHauls", "VIA VIA VIA "
	const Yep32u Cent = 0x746E6543U;
	const Yep32u aurH = 0x48727561U;
	const Yep32u auls = 0x736C7561U;
	const Yep32u VIA = 0x20414956U;
	/// Transmeta vendor strings: "GenuineTMx86", "TransmetaCPU"
	const Yep32u ineT = 0x54656E69U;
	const Yep32u Mx86 = 0x3638784DU;
	const Yep32u Tran = 0x6E617254U;
	const Yep32u smet = 0x74656D73U;
	const Yep32u aCPU = 0x55504361U;
	/// Cyrix vendor string: "CyrixInstead"
	const Yep32u Cyri = 0x69727943U;
	const Yep32u xIns = 0x736E4978U;
	const Yep32u tead = 0x64616574U;
	/// Rise vendor string: "RiseRiseRise"
	const Yep32u Rise = 0x65736952U;
	/// NSC vendor string: "Geode by NSC"
	const Yep32u Geod = 0x646F6547U;
	const Yep32u e_by = 0x79622065U;
	const Yep32u NSC = 0x43534E20U;
	/// SiS vendor string: "SiS SiS SiS "
	const Yep32u SiS = 0x20536953U;
	/// NexGen vendor string: "NexGenDriven"
	const Yep32u NexG = 0x4778654EU;
	const Yep32u enDr = 0x72446E65U;
	const Yep32u iven = 0x6E657669U;
	/// UMC vendor string: "UMC UMC UMC "
	const Yep32u UMC = 0x20434D55U;
	/// RDC vendor string: "Genuine  RDC"
	const Yep32u ine = 0x20656E69U;
	const Yep32u RDC = 0x43445220U;
	/// D&MP vendor string: "Vortex86 SoC"
	const Yep32u Vort = 0x74726F56U;
	const Yep32u ex86 = 0x36387865U;
	const Yep32u SoC = 0x436F5320U;
	Yep32u vendorString0, vendorString1, vendorString2;
	{
		int registers[4];
		__cpuid(registers, 0);
		vendorString0 = registers[1];
		vendorString1 = registers[3];
		vendorString2 = registers[2];
	}
	if( (vendorString0 == Genu) && (vendorString1 == ineI) && (vendorString2 == ntel) ) {
		_vendor = YepCpuVendorIntel;
	} else if( (vendorString0 == Auth) && (vendorString1 == enti) && (vendorString2 == cAMD) ) {
		_vendor = YepCpuVendorAMD;
	} else if( (vendorString0 == AMDi) && (vendorString1 == sbet) && (vendorString2 == ter) ) {
		_vendor = YepCpuVendorAMD;
	} else if( (vendorString0 == AMD) && (vendorString1 == ISBE) && (vendorString2 == TTER) ) {
		_vendor = YepCpuVendorAMD;
	} else if( (vendorString0 == Cent) && (vendorString1 == aurH) && (vendorString2 == auls) ) {
		_vendor = YepCpuVendorVIA;
	} else if( (vendorString0 == VIA) && (vendorString1 == VIA) && (vendorString2 == VIA) ) {
		_vendor = YepCpuVendorVIA;
	} else if( (vendorString0 == Genu) && (vendorString1 == ineT) && (vendorString2 == Mx86) ) {
		_vendor = YepCpuVendorTransmeta;
	} else if( (vendorString0 == Tran) && (vendorString1 == smet) && (vendorString2 == aCPU) ) {
		_vendor = YepCpuVendorTransmeta;
	} else if( (vendorString0 == Cyri) && (vendorString1 == xIns) && (vendorString2 == tead) ) {
		_vendor = YepCpuVendorCyrix;
	} else if( (vendorString0 == Rise) && (vendorString1 == Rise) && (vendorString2 == Rise) ) {
		_vendor = YepCpuVendorRise;
	} else if( (vendorString0 == Geod) && (vendorString1 == e_by) && (vendorString2 == NSC) ) {
		_vendor = YepCpuVendorNSC;
	} else if( (vendorString0 == SiS) && (vendorString1 == SiS) && (vendorString2 == SiS) ) {
		_vendor = YepCpuVendorSiS;
	} else if( (vendorString0 == NexG) && (vendorString1 == enDr) && (vendorString2 == iven) ) {
		_vendor = YepCpuVendorNexGen;
	} else if( (vendorString0 == UMC) && (vendorString1 == UMC) && (vendorString2 == UMC) ) {
		_vendor = YepCpuVendorUMC;
	} else if( (vendorString0 == Genu) && (vendorString1 == ine) && (vendorString2 == RDC) ) {
		_vendor = YepCpuVendorRDC;
	} else if( (vendorString0 == Vort) && (vendorString1 == ex86) && (vendorString2 == SoC) ) {
		_vendor = YepCpuVendorDMP;
	} else {
		_vendor = YepCpuVendorUnknown;
	}
}

static void _yepLibrary_InitX86CpuModelInfo() {
	Yep32u maxBaseCpuidIndex = 0;
	{
		int registers[4] = { 0, 0, 0, 0 };
		__cpuid(registers, 0);
		maxBaseCpuidIndex = registers[0];
	}
	if (maxBaseCpuidIndex >= 1u) {
		int registers[4] = { 0, 0, 0, 0 };
		__cpuid( registers, 1 );
		_modelInfo.stepping = registers[0] & 0xF;
		_modelInfo.baseModel = (registers[0] >> 4) & 0xF;
		_modelInfo.baseFamily = (registers[0] >> 8) & 0xF;
		_modelInfo.processorType = (registers[0] >> 12) & 0x3;
		_modelInfo.extModel = (registers[0] >> 16) & 0xF;
		_modelInfo.extFamily = (registers[0] >> 20) & 0xFF;
		
		_modelInfo.family = _modelInfo.baseFamily + _modelInfo.extFamily;
		_modelInfo.model = _modelInfo.baseModel + _modelInfo.extModel * 16;
	}
}

static void _yepLibrary_InitX86CpuMicroarchitectureInfo() {
	switch (_vendor) {
		case YepCpuVendorIntel:
			switch (_modelInfo.family) {
				case 0x05:
					_microarchitecture = YepCpuMicroarchitectureP5;
					break;
				case 0x06:
					switch (_modelInfo.model) {
						case 0x01: // Pentium Pro
						case 0x03: // Pentium II (Klamath) and Pentium II Overdrive
						case 0x05: // Pentium II (Deschutes, Tonga), Pentium II Celeron (Covington), Pentium II Xeon (Drake)
						case 0x06: // Pentium II (Dixon), Pentium II Celeron (Mendocino)
						case 0x07: // Pentium III (Katmai), Pentium III Xeon (Tanner)
						case 0x08: // Pentium III (Coppermine), Pentium II Celeron (Coppermine-128), Pentium III Xeon (Cascades)
						case 0x0A: // Pentium III Xeon (Cascades-2MB)
						case 0x0B: // Pentium III (Tualatin), Pentium III Celeron (Tualatin-256)
							_microarchitecture = YepCpuMicroarchitectureP6;
							break;
						case 0x09: // Pentium M (Banias), Pentium M Celeron (Banias-0, Banias-512)
						case 0x0D: // Pentium M (Dothan), Pentium M Celeron (Dothan-512, Dothan-1024)
						case 0x15: // Intel 80579 (Tolapai)
							_microarchitecture = YepCpuMicroarchitectureDothan;
							break;
						case 0x0E: // Core Solo/Duo (Yonah), Pentium Dual-Core T2xxx (Yonah), Celeron M (Yonah-512, Yonah-1024), Dual-Core Xeon (Sossaman)
							_microarchitecture = YepCpuMicroarchitectureYonah;
							break;
						case 0x0F: // Core 2 Duo (Conroe, Conroe-2M, Merom), Core 2 Quad (Tigerton), Xeon (Woodcrest, Clovertown, Kentsfield)
						case 0x16: // Celeron (Conroe-L, Merom-L), Core 2 Duo (Merom)
							_microarchitecture = YepCpuMicroarchitectureConroe;
							break;
						case 0x17: // Core 2 Duo (Penryn-3M), Core 2 Quad (Yorkfield), Core 2 Extreme (Yorkfield), Xeon (Harpertown), Pentium Dual-Core (Penryn)
						case 0x1D: // Xeon (Dunnington)
							_microarchitecture = YepCpuMicroarchitecturePenryn;
							break;
						case 0x1C: // Diamondville, Silverthorne, Pineview
						case 0x26: // Tunnel Creek
							_microarchitecture = YepCpuMicroarchitectureBonnell;
							break;
						case 0x27: // Medfield
						case 0x35: // Cloverview
						case 0x36: // Cedarview, Centerton 
							_microarchitecture = YepCpuMicroarchitectureSaltwell;
							break;
						case 0x1A: // Core iX (Bloomfield), Xeon (Gainestown)
						case 0x1E: // Core iX (Lynnfield, Clarksfield)
						case 0x1F: // Core iX (Havendale)
						case 0x2E: // Xeon (Beckton)
						case 0x25: // Core iX (Clarkdale)
						case 0x2C: // Core iX (Gulftown), Xeon (Gulftown)
						case 0x2F: // Xeon (Eagleton)
							_microarchitecture = YepCpuMicroarchitectureNehalem;
							break;
						case 0x2A: // Core iX (Sandy Bridge)
						case 0x2D: // Core iX (Sandy Bridge-E), Xeon (Sandy Bridge EP/EX)
							_microarchitecture = YepCpuMicroarchitectureSandyBridge;
							break;
						case 0x3A: // Core iX (Ivy Bridge)
							_microarchitecture = YepCpuMicroarchitectureIvyBridge;
							break;
						case 0x3C:
							_microarchitecture = YepCpuMicroarchitectureHaswell;
							break;
					}
					break;
				case 0x0B:
					switch (_modelInfo.model) {
						case 0x00:
							_microarchitecture = YepCpuMicroarchitectureKnightsCorner;
							break;
					}
					break;
				case 0x0F:
					switch (_modelInfo.model) {
						case 0x00: // Pentium 4 Xeon (Foster)
						case 0x01: // Pentium 4 Celeron (Willamette-128), Pentium 4 Xeon (Foster, Foster MP)
						case 0x02: // Pentium 4 (Northwood), Pentium 4 EE (Gallatin), Pentium 4 Celeron (Northwood-128, Northwood-256), Pentium 4 Xeon (Gallatin DP, Prestonia)
							_microarchitecture = YepCpuMicroarchitectureWillamette;
							break;
						case 0x03: // Pentium 4 (Prescott), Pentium 4 Xeon (Nocona)
						case 0x04: // Pentium 4 (Prescott-2M), Pentium 4 EE (Prescott-2M), Pentium D (Smithfield), Celeron D (Prescott-256), Pentium 4 Xeon (Cranford, Irwindale, Paxville)
						case 0x06: // Pentium 4 (Cedar Mill), Pentium D EE (Presler), Celeron D (Cedar Mill), Pentium 4 Xeon (Dempsey, Tulsa)
							_microarchitecture = YepCpuMicroarchitecturePrescott;
							break;
					}
					break;
			}
			break;
		case YepCpuVendorAMD:
			switch (_modelInfo.family) {
				case 0x5:
					switch (_modelInfo.model) {
						case 0x00:
						case 0x01:
						case 0x02:
							_microarchitecture = YepCpuMicroarchitectureK5;
							break;
						case 0x06:
						case 0x07:
						case 0x08:
						case 0x0D:
							_microarchitecture = YepCpuMicroarchitectureK6;
							break;
						case 0x0A:
							_microarchitecture = YepCpuMicroarchitectureGeode;
							break;
					}
					break;
				case 0x6:
					_microarchitecture = YepCpuMicroarchitectureK7;
					break;
				case 0xF: // Opteron, Athlon 64, Sempron
				case (0xF + 0x02): // Turion
					_microarchitecture = YepCpuMicroarchitectureK8;
					break;
				case (0xF + 0x01): // Opteron, Phenom, Athlon, Sempron
				case (0xF + 0x03): // Llano APU
					_microarchitecture = YepCpuMicroarchitectureK10;
					break;
				case (0xF + 0x05):
					_microarchitecture = YepCpuMicroarchitectureBobcat;
					break;
				case (0xF + 0x06):
					switch (_modelInfo.model) {
						case 0x0:
							_microarchitecture = YepCpuMicroarchitectureBulldozer;
							break;
						case 0x1: // With L3 cache
						case 0x2: // No L3 cache
							_microarchitecture = YepCpuMicroarchitecturePiledriver;
							break;
						case 0x3: // With L3 cache
						case 0x4: // No L3 cache
							_microarchitecture = YepCpuMicroarchitectureSteamroller;
							break;
					}
					break;
				case (0xF + 0x07):
					_microarchitecture = YepCpuMicroarchitectureJaguar;
					break;
			}
			break;
	}
}

static void _yepLibrary_InitX86CpuIsaInfo() {
	Yep64u isaFeatures = YepIsaFeaturesDefault;
	Yep64u simdFeatures = YepSimdFeaturesDefault;
	Yep64u systemFeatures = YepSystemFeaturesDefault;

	// Processors without CPUID instruction are not supported
	isaFeatures |= YepX86IsaFeatureCpuid;
	// All x86 and x86-64 processors support misaligned memory access
	systemFeatures |= YepSystemFeatureMisalignedAccess;
#if defined(YEP_X64_CPU)
	systemFeatures |= YepSystemFeatureAddressSpace64Bit;
	systemFeatures |= YepSystemFeatureGPRegisters64Bit;
#endif

	Yep32u maxBaseCpuidIndex = 0;
	Yep32u maxExtendedCpuidIndex = 0;
	{
		int registers[4] = { 0, 0, 0, 0 };
		__cpuid(registers, 0);
		maxBaseCpuidIndex = registers[0];
		__cpuid(registers, 0x80000000);
		maxExtendedCpuidIndex = registers[0];
	}

	int basicInfo[4] = { 0, 0, 0, 0 };
	if YEP_LIKELY(maxBaseCpuidIndex >= 1u) {
		__cpuid(basicInfo, 1);
	}
	int structuredFeatureInfo[4] = { 0, 0, 0, 0 };
	if YEP_LIKELY(maxBaseCpuidIndex >= 7u) {
		__cpuidex(structuredFeatureInfo, 7, 0);
		// Intel, AMD: ebx[bit 3] in structured feature info (ecx = 0).
		if YEP_LIKELY(structuredFeatureInfo[1] & 0x00000008u) {
			isaFeatures |= YepX86IsaFeatureBMI;
		}
		// Intel: ebx[bit 8] in structured feature info (ecx = 0).
		if YEP_LIKELY(structuredFeatureInfo[1] & 0x00000100u) {
			isaFeatures |= YepX86IsaFeatureBMI2;
		}
		// Intel: ebx[bit 18] in structured feature info (ecx = 0).
		if YEP_LIKELY(structuredFeatureInfo[1] & 0x00040000u) {
			isaFeatures |= YepX86IsaFeatureRdseed;
		}
		// Intel: ebx[bit 19] in structured feature info (ecx = 0).
		if YEP_LIKELY(structuredFeatureInfo[1] & 0x00080000u) {
			isaFeatures |= YepX86IsaFeatureADX;
		}
	}
	int extendedInfo[4] = { 0, 0, 0, 0 };
	if YEP_LIKELY(maxExtendedCpuidIndex >= 0x80000001u) {
		__cpuid(extendedInfo, 0x80000001);
	}

	// Check for RDTSC instruction.
#if defined(YEP_X86_CPU)
	// Intel, AMD: edx[bit 4] in basic info.
	// AMD: edx [bit 4] in extended info (reserved bit on Intel CPUs).
	if YEP_LIKELY((basicInfo[3] | extendedInfo[3]) & 0x00000010u) {
		isaFeatures |= YepX86IsaFeatureRdtsc;
		systemFeatures |= YepSystemFeatureCycleCounter;
		systemFeatures |= YepSystemFeatureCycleCounter64Bit;
	}
#else
	isaFeatures |= YepX86IsaFeatureRdtsc;
	systemFeatures |= YepSystemFeatureCycleCounter;
	systemFeatures |= YepSystemFeatureCycleCounter64Bit;
#endif

	// Intel, AMD: edx[bit 11] in basic info.
	if YEP_LIKELY(basicInfo[3] & 0x00000800u) {
		isaFeatures |= YepX86IsaFeatureSYSENTER;
	}
	// Intel, AMD: edx[bit 11] in extended info.
	if YEP_LIKELY(extendedInfo[3] & 0x00000800u) {
		isaFeatures |= YepX86IsaFeatureSYSCALL;
	}
	// Intel, AMD: edx[bit 5] in basic info.
	// AMD: edx[bit 5] in extended info (reserved bit on Intel CPUs).
	if YEP_LIKELY((basicInfo[3] | extendedInfo[3]) & 0x00000020u) {
		isaFeatures |= YepX86IsaFeatureMSR;
	}
	// Intel, AMD: edx[bit 19] in basic info.
	if YEP_LIKELY(basicInfo[3] & 0x00080000u) {
		isaFeatures |= YepX86IsaFeatureClflush;
	}
	// Intel, AMD: ecx[bit 3] in basic info.
	if YEP_LIKELY(basicInfo[2] & 0x00000008u) {
		isaFeatures |= YepX86IsaFeatureMONITOR;
	}
	// Intel, AMD: edx[bit 24] in basic info.
	// AMD: edx[bit 24] in extended info (zero bit on Intel CPUs, EMMX bit on Cyrix CPUs).
	if YEP_LIKELY(basicInfo[3] & 0x01000000u) {
		isaFeatures |= YepX86IsaFeatureFXSAVE;
	}
	if YEP_LIKELY(extendedInfo[3] & 0x01000000u) {
#if defined(YEP_X86_CPU)
		if YEP_UNLIKELY(_vendor == YepCpuVendorCyrix) {
			simdFeatures |= YepX86SimdFeatureEMMX;
		} else {
			isaFeatures |= YepX86IsaFeatureFXSAVE;
		}
#else
		isaFeatures |= YepX86IsaFeatureFXSAVE;
#endif
	}
	// Intel, AMD: ecx[bit 26] in basic info.
	if YEP_LIKELY(basicInfo[2] & 0x04000000u) {
		isaFeatures |= YepX86IsaFeatureXSAVE;
	}
	// Intel, AMD: edx[bit 0] in basic info.
	// AMD: edx[bit 0] in extended info (reserved bit on Intel CPUs).
	if YEP_LIKELY((basicInfo[3] | extendedInfo[3]) & 0x00000001u) {
		isaFeatures |= YepX86IsaFeatureFPU;
	}
	
	// Intel, AMD: edx[bit 23] in basic info.
	// AMD: edx[bit 23] in extended info (zero bit on Intel CPUs).
	if YEP_LIKELY((basicInfo[3] | extendedInfo[3]) & 0x00800000u) {
		simdFeatures |= YepX86SimdFeatureMMX;
	}
	// Intel, AMD: edx[bit 25] in basic info (SSE feature flag).
	// Pre-SSE AMD: edx[bit 22] in extended info (zero bit on Intel CPUs).
	if YEP_LIKELY((basicInfo[3] & 0x02000000u) | (extendedInfo[3] & 0x00400000u)) {
		simdFeatures |= YepX86SimdFeatureMMXPlus;
	}

	// AMD: edx[bit 31] of extended info (zero bit on Intel CPUs).
	if YEP_LIKELY(extendedInfo[3] & 0x80000000u) {
		simdFeatures |= YepX86SimdFeature3dnow;
	}
	// AMD: edx[bit 30] of extended info (zero bit on Intel CPUs).
	if YEP_LIKELY(extendedInfo[3] & 0x40000000u) {
		simdFeatures |= YepX86SimdFeature3dnowPlus;
		if YEP_UNLIKELY(_microarchitecture == YepCpuMicroarchitectureGeode) {
			simdFeatures |= YepX86SimdFeature3dnowGeode;
		}
	}
	// AMD, Intel: ecx[bit 8] of extended info (called PREFETCHW in Intel manuals).
	// AMD: edx[bit 31] of extended info (same as for 3dnow!)
	// AMD: edx[bit 30] of extended info (same as for 3dnow!+)
	if YEP_LIKELY((extendedInfo[2] & 0x00000100u) | (extendedInfo[3] & 0xC0000000u)) {
		simdFeatures |= YepX86SimdFeature3dnowPrefetch;
	}
#if defined(YEP_X64_CPU)
	isaFeatures |= YepX86IsaFeatureX64;
	simdFeatures |= YepX86SimdFeatureSSE;
	simdFeatures |= YepX86SimdFeatureSSE2;
#else
	// Intel, AMD: edx[bit 29] in extended info.
	if YEP_LIKELY(basicInfo[3] & 0x20000000u) {
		isaFeatures |= YepX86IsaFeatureX64;
	}
	// Intel, AMD: edx[bit 25] in basic info.
	if YEP_LIKELY(basicInfo[3] & 0x02000000u) {
		simdFeatures |= YepX86SimdFeatureSSE;
	}
	// Intel, AMD: edx[bit 26] in basic info.
	if YEP_LIKELY(basicInfo[3] & 0x04000000u) {
		simdFeatures |= YepX86SimdFeatureSSE2;
	}
#endif
	// Intel, AMD: ecx[bit 0] in basic info.
	if YEP_LIKELY(basicInfo[2] & 0x00000001u) {
		simdFeatures |= YepX86SimdFeatureSSE3;
	}
	// Intel, AMD: ecx[bit 9] in basic info.
	if YEP_LIKELY(basicInfo[2] & 0x0000200u) {
		simdFeatures |= YepX86SimdFeatureSSSE3;
	}
	// Intel, AMD: ecx[bit 19] in basic info.
	if YEP_LIKELY(basicInfo[2] & 0x00080000u) {
		simdFeatures |= YepX86SimdFeatureSSE4_1;
	}
	// Intel: ecx[bit 20] in basic info (reserved bit on AMD CPUs).
	if YEP_LIKELY(basicInfo[2] & 0x00100000u) {
		simdFeatures |= YepX86SimdFeatureSSE4_2;
	}
	// AMD: ecx[bit 6] in extended info (reserved bit on Intel CPUs).
	if YEP_LIKELY(extendedInfo[2] & 0x00000040u) {
		simdFeatures |= YepX86SimdFeatureSSE4A;
	}
	// AMD: ecx[bit 7] in extended info (reserved bit on Intel CPUs).
	if YEP_LIKELY(extendedInfo[2] & 0x00000080u) {
		systemFeatures |= YepX86SystemFeatureMisalignedSSE;
	}

	// Intel, AMD: ecx[bit 28] in basic info.
	if YEP_LIKELY(basicInfo[2] & 0x10000000u) {
		simdFeatures |= YepX86SimdFeatureAVX;
	}
	// Intel: ebx[bit 5] in structured feature into.
	if YEP_LIKELY(structuredFeatureInfo[1] & 0x00000020u) {
		simdFeatures |= YepX86SimdFeatureAVX2;
	}

	// Intel: ecx[bit 12] in basic info (reserved bit on AMD CPUs).
	if YEP_LIKELY(basicInfo[2] & 0x00001000u) {
		simdFeatures |= YepX86SimdFeatureFMA3;
	}
	// AMD: ecx[bit 16] in extended info (reserved bit on Intel CPUs).
	if YEP_LIKELY(extendedInfo[2] & 0x00010000u) {
		simdFeatures |= YepX86SimdFeatureFMA4;
	}

	// AMD: ecx[bit 11] in extended info (reserved bit on Intel CPUs).
	if YEP_LIKELY(extendedInfo[2] & 0x00000800u) {
		simdFeatures |= YepX86SimdFeatureXOP;
	}
	// Intel, AMD: ecx[bit 29] in basic info.
	if YEP_LIKELY(basicInfo[2] & 0x20000000u) {
		simdFeatures |= YepX86SimdFeatureF16C;
	}

	// Intel, AMD: ecx[bit 26] in basic info = XSAVE.
	// Intel, AMD: ecx[bit 27] in basic info = OSXSAVE.
	if YEP_LIKELY((basicInfo[2] & 0x0C000000u) == 0x0C000000u) {
		Yep64u xcr0ValidBits = 0ull;
		if YEP_LIKELY(maxBaseCpuidIndex >= 0x0000000Du) {
			int extendedStateEnumerationMainParameters[4];
			__cpuidex(extendedStateEnumerationMainParameters, 0x0000000Du, 0x00000000u);
			xcr0ValidBits = yepBuiltin_CombineParts_32u32u_64u(extendedStateEnumerationMainParameters[3], extendedStateEnumerationMainParameters[0]);
		}
		
		Yep64u xfeatureEnabledMask = _xgetbv(0);
		// Intel, AMD: XFEATURE_ENABLED_MASK[bit 0] for FPU XSAVE
		if YEP_LIKELY(isaFeatures & YepX86IsaFeatureFPU) {
			if YEP_LIKELY(xcr0ValidBits & 0x0000000000000001ull) {
				if YEP_LIKELY(xfeatureEnabledMask & 0x0000000000000001ull) {
					systemFeatures |= YepX86SystemFeatureFPU;
				}
			} else {
				systemFeatures |= YepX86SystemFeatureFPU;
			}
		}
		// Intel, AMD: XFEATURE_ENABLED_MASK[bit 1] for SSE XSAVE
		if YEP_LIKELY(simdFeatures & YepX86SimdFeatureSSE) {
			if YEP_LIKELY(xcr0ValidBits & 0x0000000000000002ull) {
				if YEP_LIKELY(xfeatureEnabledMask & 0x0000000000000002ull) {
					systemFeatures |= YepX86SystemFeatureSSE;
				}
			} else {
				systemFeatures |= YepX86SystemFeatureSSE;
			}
		}
		// Intel, AMD: XFEATURE_ENABLED_MASK[bit 2] for AVX XSAVE
		if YEP_LIKELY(simdFeatures & YepX86SimdFeatureAVX) {
			if YEP_LIKELY((xcr0ValidBits & xfeatureEnabledMask) & 0x0000000000000004ull) {
				systemFeatures |= YepX86SystemFeatureAVX;
			}
		}
	} else {
		// If the OSXSAVE feature is not supported, speculate that OS supports FPU & SSE if the CPU does.
		if YEP_LIKELY(isaFeatures & YepX86IsaFeatureFPU) {
			systemFeatures |= YepX86SystemFeatureFPU;
		}
		if YEP_LIKELY(simdFeatures & YepX86SimdFeatureSSE) {
			systemFeatures |= YepX86SystemFeatureSSE;
		}
	}
	
#if defined(YEP_X64_CPU)
	isaFeatures |= YepX86IsaFeatureCMOV;
#else
	// Intel, AMD: edx[bit 15] in basic info.
	// AMD: edx[bit 15] in extended info (zero bit on Intel CPUs).
	if YEP_LIKELY((basicInfo[3] | extendedInfo[3]) & 0x00008000u) {
		isaFeatures |= YepX86IsaFeatureCMOV;
	}
#endif

#if defined(YEP_X64_CPU)
	isaFeatures |= YepX86IsaFeatureCmpxchg8b;
#else
	// Intel, AMD: edx[bit 8] in basic info.
	// AMD: edx[bit 8] in extended info (reserved bit on Intel CPUs).
	if YEP_LIKELY((basicInfo[3] | extendedInfo[3]) & 0x00000100u) {
		isaFeatures |= YepX86IsaFeatureCmpxchg8b;
	}
#endif

	// Intel, AMD: ecx[bit 13] in basic info.
	if YEP_LIKELY(basicInfo[2] & 0x00002000u) {
		isaFeatures |= YepX86IsaFeatureCmpxchg16b;
	}

	// Intel: ecx[bit 22] in basic info.
	if YEP_LIKELY(basicInfo[2] & 0x00400000u) {
		isaFeatures |= YepX86IsaFeatureMovbe;
	}

	// Some early x86-64 CPUs lack LAHF & SAHF instructions.
	// A special CPU feature bit must be checked to ensure their availability.
	// Intel, AMD: ecx[bit 0] in extended info.
	if YEP_LIKELY(extendedInfo[2] & 0x00000001u) {
		isaFeatures |= YepX86IsaFeatureLahfSahf64;
	}
	// Intel: ebx[bit 0] in structured feature info.
	if YEP_LIKELY(structuredFeatureInfo[1] & 0x00000001U) {
		isaFeatures |= YepX86IsaFeatureFsGsBase;
	}

	// Intel, AMD: ecx[bit 5] in extended info.
	if YEP_LIKELY(extendedInfo[2] & 0x00000020u) {
		isaFeatures |= YepX86IsaFeatureLzcnt;
	}
	// Intel, AMD: ecx[bit 23] in basic info.
	if YEP_LIKELY(basicInfo[2] & 0x00800000u) {
		isaFeatures |= YepX86IsaFeaturePopcnt;
	}
	// AMD: ecx[bit 21] in extended info (reserved bit on Intel CPUs).
	if YEP_LIKELY(extendedInfo[2] & 0x00200000u) {
		isaFeatures |= YepX86IsaFeatureTBM;
	}

	// Intel, AMD: ebx[bit 3] in structured feature info.
	if YEP_LIKELY(structuredFeatureInfo[1] & 0x00000008u) {
		isaFeatures |= YepX86IsaFeatureBMI;
	}
	// Intel: ebx[bit 8] in structured feature info.
	if YEP_LIKELY(structuredFeatureInfo[1] & 0x00000100u) {
		isaFeatures |= YepX86IsaFeatureBMI2;
	}

	// Intel: ecx[bit 25] in basic info (reserved bit on AMD CPUs).
	if YEP_LIKELY(basicInfo[2] & 0x02000000u) {
		isaFeatures |= YepX86IsaFeatureAES;
	}
	// Intel: ecx[bit 1] in basic info (reserved bit on AMD CPUs).
	if YEP_LIKELY(basicInfo[2] & 0x00000002u) {
		isaFeatures |= YepX86IsaFeaturePclmulqdq;
	}
	// Intel: ecx[bit 30] in basic info (reserved bit on AMD CPUs).
	if YEP_LIKELY(basicInfo[2] & 0x40000000u) {
		isaFeatures |= YepX86IsaFeatureRdrand;
	}

	int padlockInfo[4] = { 0, 0, 0, 0 };
	__cpuid(padlockInfo, 0xC0000000u);
	Yep32u maxPadlockCpuidIndex = padlockInfo[0];
	if YEP_UNLIKELY(maxPadlockCpuidIndex >= 0xC0000001u) {
		__cpuid(padlockInfo, 0xC0000001);
		
		// VIA: edx[bit 2] in padlock info = RNG exists on chip flag.
		if YEP_LIKELY(padlockInfo[3] & 0x00000004u) {
			isaFeatures |= YepX86IsaFeatureRNG;
			// VIA: edx[bit 3] in padlock info = RNG enabled by OS.
			if YEP_LIKELY(padlockInfo[3] & 0x00000008u) {
				systemFeatures |= YepX86SystemFeatureRNG;
			}
		}
		// VIA: edx[bit 6] in padlock info = ACE exists on chip flag.
		if YEP_LIKELY(padlockInfo[3] & 0x00000040u) {
			isaFeatures |= YepX86IsaFeatureACE;
			// VIA: edx[bit 7] in padlock info = ACE enabled by OS.
			if YEP_LIKELY(padlockInfo[3] & 0x00000080u) {
				systemFeatures |= YepX86SystemFeatureACE;
			}
		}
		// VIA: edx[bit 8] in padlock info = ACE2 exists on chip flag.
		if YEP_LIKELY(padlockInfo[3] & 0x00000100u) {
			isaFeatures |= YepX86IsaFeatureACE2;
			// VIA: edx[bit 9] in padlock info = ACE 2 enabled by OS.
			if YEP_LIKELY(padlockInfo[3] & 0x00000200u) {
				systemFeatures |= YepX86SystemFeatureACE2;
			}
		}
		// VIA: edx[bit 10] in padlock info = PHE exists on chip flag.
		if YEP_LIKELY(padlockInfo[3] & 0x00000400u) {
			isaFeatures |= YepX86IsaFeaturePHE;
			// VIA: edx[bit 11] in padlock info = PHE enabled by OS.
			if YEP_LIKELY(padlockInfo[3] & 0x00000400u) {
				systemFeatures |= YepX86SystemFeaturePHE;
			}
		}
		// VIA: edx[bit 12] in padlock info = PMM exists on chip flag.
		if YEP_LIKELY(padlockInfo[3] & 0x00001000u) {
			isaFeatures |= YepX86IsaFeaturePMM;
			// VIA: edx[bit 13] in padlock info = PMM enabled by OS.
			if YEP_LIKELY(padlockInfo[3] & 0x00002000u) {
				systemFeatures |= YepX86SystemFeaturePMM;
			}
		}
	}
	
	// AMD: ecx[bit 15] in extended info (reserved bit on Intel CPUs).
	if YEP_LIKELY(extendedInfo[2] & 0x00008000u) {
		isaFeatures |= YepX86IsaFeatureLWP;
	}

	// Intel, AMD: edx[bit 27] in extended info.
	if YEP_LIKELY(extendedInfo[3] & 0x08000000u) {
		isaFeatures |= YepX86IsaFeatureRdtscp;
	}
	
	// Intel: ebx[bit 4] in structured feature info
	if YEP_LIKELY(structuredFeatureInfo[1] & 0x00000010u) {
		isaFeatures |= YepX86IsaFeatureHLE;
	}
	// Intel: ebx[bit 11] in structured feature info
	if YEP_LIKELY(structuredFeatureInfo[1] & 0x00000800u) {
		isaFeatures |= YepX86IsaFeatureRTM;
	}
	// Intel: either HLE or RTM is supported
	if YEP_LIKELY(structuredFeatureInfo[1] & 0x00000810u) {
		isaFeatures |= YepX86IsaFeatureXtest;
	}

	_isaFeatures = isaFeatures;
	_simdFeatures = simdFeatures;
	_systemFeatures = systemFeatures;
}

static void _yepLibrary_InitX86CpuCacheInfo() {
	Yep32u maxBaseCpuidIndex = 0;
	Yep32u maxExtendedCpuidIndex = 0;
	{
		int registers[4];
		__cpuid(registers, 0);
		maxBaseCpuidIndex = registers[0];
		__cpuid(registers, 0x80000000);
		maxExtendedCpuidIndex = registers[0];
	}
	if YEP_LIKELY(maxBaseCpuidIndex >= 1u) {
		int registers[4];
		__cpuid(registers, 1);
		const Yep32u clflushLineSizeInQwords = static_cast<Yep8u>(registers[1] >> 8);
		_cache.clflushLineSize = static_cast<Yep16u>( clflushLineSizeInQwords * 8 );
	}
	if YEP_LIKELY(maxBaseCpuidIndex >= 2u) {
		int cacheDescriptors[4];
		__cpuid(cacheDescriptors, 2);
		Yep32u numberOfCpuidIterations = cacheDescriptors[0] & 0xFF;
		cacheDescriptors[0] &= 0xFFFFFF00;
		while (numberOfCpuidIterations != 0) {
			for (Yep32u registerIndex = 0; registerIndex < 4; registerIndex++) {
				if YEP_UNLIKELY(cacheDescriptors[registerIndex] < 0) {
					continue;
				} else {
					for (Yep32u descriptorIndex = 0; descriptorIndex < sizeof(cacheDescriptors[registerIndex]); descriptorIndex++) {
						_yepLibrary_DecodeX86CpuCacheDescriptor(static_cast<Yep8u>(cacheDescriptors[registerIndex] >> (descriptorIndex * 8)));
					}
				}
			}
			if (--numberOfCpuidIterations != 0) {
				__cpuid(cacheDescriptors, 2);
			}
		} ;
	}
	if YEP_LIKELY(maxBaseCpuidIndex >= 4u) {
		for (Yep32s index = 0; index >= 0; index++) {
			int deterministicCacheParameters[4];
			__cpuidex(deterministicCacheParameters, 4, index);
			const Yep32u cacheType = deterministicCacheParameters[0] & 0x1F;
			if YEP_LIKELY(cacheType == 0) {
				break;
			} else {
				const Yep32u cacheLevel = (deterministicCacheParameters[0] >> 5) & 0x7;
				const Yep32u isFullyAssociative = YepBoolean(deterministicCacheParameters[0] & 0x200);
				const Yep32u numberOfSets = 1 + deterministicCacheParameters[2];
				const Yep32u waysOfAssociativity = 1 + (Yep32u(deterministicCacheParameters[1]) >> 22);
				const Yep32u physicalLinePartitions = 1 + ((deterministicCacheParameters[1] >> 12) & 0x3FF);
				const Yep32u lineSize = 1 + (deterministicCacheParameters[1] & 0x0FFF);
				const Yep32u cacheSize = waysOfAssociativity * physicalLinePartitions * lineSize * numberOfSets;
				_yepLibrary_DecodeX86CpuDeterministicCacheParameters(cacheType, cacheLevel, cacheSize, lineSize, isFullyAssociative ? 0xFFFFFFFFu : waysOfAssociativity);
			}
		}
	}
	if (YEP_LIKELY(maxExtendedCpuidIndex >= 0x80000005u) && YEP_UNLIKELY(_vendor == YepCpuVendorAMD)) {
		int cacheParameters[4];
		__cpuid(cacheParameters, 0x80000005u);
		
		// AMD: ecx[bits 24:31] = L1 data cache size in KB
		_cache.L1DCacheInfo.cacheSize = (static_cast<Yep32u>(cacheParameters[2]) >> 24) * 1024;
		// AMD: ecx[bits 16:23] = L1 data cache associativity
		_cache.L1DCacheInfo.associativity = static_cast<Yep16s>(static_cast<Yep8s>(cacheParameters[2] >> 16));
		// AMD: ecx[bits 0:7] = L1 data cache line size in bytes
		_cache.L1DCacheInfo.lineSize = cacheParameters[2] & 0xFF;
		
		// AMD: edx[bits 24:31] = L1 instruction cache size in KB
		_cache.L1ICacheInfo.cacheSize = (static_cast<Yep32u>(cacheParameters[3]) >> 24) * 1024;
		// AMD: edx[bits 16:23] = L1 instruction cache associativity
		_cache.L1ICacheInfo.associativity = static_cast<Yep16s>(static_cast<Yep8s>(cacheParameters[3] >> 16));
		// AMD: edx[bits 0:7] = L1 instruction cache line size in bytes
		_cache.L1ICacheInfo.lineSize = cacheParameters[3] & 0xFFu;
	}
	if YEP_LIKELY(maxExtendedCpuidIndex >= 0x80000006u) {
		int cacheParameters[4];
		__cpuid(cacheParameters, 0x80000006u);
		
		static const Yep8u associativityMap[16] = {
			 0,  1,   2,  0,
			 4,  0,   8,  0,
			16,  0,  32, 48,
			64, 96, 128, 0xFF
		};
		
		// AMD, Intel: ecx[bits 16:31] = L2 cache size in KB
		_cache.L2CacheInfo.cacheSize = ((cacheParameters[2] >> 16) & 0xFFFFu) * 1024;
		// AMD, Intel: ecx[bits 12:15] = L2 cache associativity
		_cache.L2CacheInfo.associativity = associativityMap[(cacheParameters[2] >> 12) & 0xFU];
		// AMD, Intel: ecx[bits 0:7] = L2 cache line size in bytes
		_cache.L2CacheInfo.lineSize = cacheParameters[2] & 0xFFU;
		
		if YEP_UNLIKELY(_vendor == YepCpuVendorAMD) {
			// AMD: edx[bits 18:31] = L3 cache size in KB
			_cache.L3CacheInfo.cacheSize = (Yep32u(cacheParameters[3]) >> 18) * 524288;
			// AMD: edx[bits 12:15] = L3 cache associativity
			_cache.L3CacheInfo.associativity = associativityMap[(cacheParameters[3] >> 12) & 0xFu];
			// AMD: edx[bits 0:7] = L3 cache line size in bytes
			_cache.L3CacheInfo.lineSize = cacheParameters[3] & 0xFFu;
		}
	}
	if (YEP_LIKELY(maxExtendedCpuidIndex >= 0x8000001Du) && YEP_UNLIKELY(_vendor == YepCpuVendorAMD)) {
		int extendedInfo[4];
		__cpuid(extendedInfo, 0x80000001u);
		
		// AMD: ecx[bit 22] in extended info = Topology Extensions
		if YEP_UNLIKELY(extendedInfo[2] & 0x00400000u) {
			int cacheParameters[4];
			
			int parameterIndex = 0;
			__cpuidex(cacheParameters, 0x8000001Du, parameterIndex);
			while ((cacheParameters[0] & 0x1F) != 0) {
				// AMD: eax[bits 0:4] = cache type:
				// - 0 = No more caches
				// - 1 = Data cache
				// - 2 = Instruction cache
				// - 3 = Unified cache
				const Yep32u cacheType = cacheParameters[0] & 0x1F;
				// AMD: eax[bits 5:7] = cache level (1, 2, or 3)
				const Yep32u cacheLevel = (cacheParameters[0] >> 5) & 0x7;

				// AMD: ebx[bits 22:31] = associativity - 1
				const Yep32u associativity = (Yep32u(cacheParameters[1]) >> 22) + 1;
				// AMD: ebx[bits 0:11] = line size - 1
				const Yep32u lineSize = (cacheParameters[1] & 0xFFF) + 1;
				// AMD: ecx[bits 0:31] = number of sets - 1
				const Yep32u numberOfSets = cacheParameters[2];
				
				if ((cacheLevel == 1) && (cacheType == 1)) {
					_cache.L1DCacheInfo.cacheSize = associativity * lineSize * numberOfSets;
					_cache.L1DCacheInfo.associativity = associativity;
					_cache.L1DCacheInfo.lineSize = lineSize;
				} else if ((cacheLevel == 1) && (cacheType == 2)) {
					_cache.L1ICacheInfo.cacheSize = associativity * lineSize * numberOfSets;
					_cache.L1ICacheInfo.associativity = associativity;
					_cache.L1ICacheInfo.lineSize = lineSize;
				} else if ((cacheLevel == 2) && (cacheType == 3)) {
					_cache.L2CacheInfo.cacheSize = associativity * lineSize * numberOfSets;
					_cache.L2CacheInfo.associativity = associativity;
					_cache.L2CacheInfo.lineSize = lineSize;
				} else if ((cacheLevel == 3) && (cacheType == 3)) {
					_cache.L3CacheInfo.cacheSize = associativity * lineSize * numberOfSets;
					_cache.L3CacheInfo.associativity = associativity;
					_cache.L3CacheInfo.lineSize = lineSize;
				}
				__cpuidex(cacheParameters, 0x8000001Du, ++parameterIndex);
			}
		}
	}
}

static void _yepLibrary_DecodeX86CpuCacheDescriptor(Yep8u descriptor) {
	// Descriptors from
	// * Application Note 485: Intel Processor Indentification and CPUID Instruction, August 2009, Order Number 241618-036
	// * Intel 64 and IA-32 Architectures Software Developer's Manual, Volume 2A: Instruction Set Reference, A-M, March 2010, Order Number 253666-034US
	// * Cyrix CPU Detection Guide, Preliminary Revision 1.01
	switch (descriptor) {
		case 0x06:
			_cache.L1ICacheInfo.cacheSize = 8*1024;
			_cache.L1ICacheInfo.lineSize = 32;
			_cache.L1ICacheInfo.associativity = 4;
			_cache.L1ICacheInfo.isUnified = false;
			break;
		case 0x08:
			_cache.L1ICacheInfo.cacheSize = 16*1024;
			_cache.L1ICacheInfo.lineSize = 32;
			_cache.L1ICacheInfo.associativity = 4;
			_cache.L1ICacheInfo.isUnified = false;
			break;
		case 0x09:
			_cache.L1ICacheInfo.cacheSize = 32*1024;
			_cache.L1ICacheInfo.lineSize = 64;
			_cache.L1ICacheInfo.associativity = 4;
			_cache.L1ICacheInfo.isUnified = false;
			break;
		case 0x0A:
			_cache.L1DCacheInfo.cacheSize = 8*1024;
			_cache.L1DCacheInfo.lineSize = 32;
			_cache.L1DCacheInfo.associativity = 2;
			_cache.L1DCacheInfo.isUnified = false;
			break;
		case 0x0C:
			_cache.L1DCacheInfo.cacheSize = 16*1024;
			_cache.L1DCacheInfo.lineSize = 32;
			_cache.L1DCacheInfo.associativity = 4;
			_cache.L1DCacheInfo.isUnified = false;
			break;
		case 0x0D:
			_cache.L1DCacheInfo.cacheSize = 16*1024;
			_cache.L1DCacheInfo.lineSize = 64;
			_cache.L1DCacheInfo.associativity = 4;
			_cache.L1DCacheInfo.isUnified = false;
			break;
		case 0x0E:
			_cache.L1DCacheInfo.cacheSize = 24*1024;
			_cache.L1DCacheInfo.lineSize = 64;
			_cache.L1DCacheInfo.associativity = 6;
			_cache.L1DCacheInfo.isUnified = false;
			break;
		case 0x21:
			_cache.L2CacheInfo.cacheSize = 256*1024;
			_cache.L2CacheInfo.lineSize = 64;
			_cache.L2CacheInfo.associativity = 8;
			_cache.L2CacheInfo.isUnified = true;
			break;
		case 0x22:
			_cache.L3CacheInfo.cacheSize = 512*1024;
			_cache.L3CacheInfo.lineSize = 64;
			_cache.L3CacheInfo.associativity = 4;
			_cache.L3CacheInfo.isUnified = true;
			break;
		case 0x23:
			_cache.L3CacheInfo.cacheSize = 1024*1024;
			_cache.L3CacheInfo.lineSize = 64;
			_cache.L3CacheInfo.associativity = 8;
			_cache.L3CacheInfo.isUnified = true;
			break;
		case 0x25:
			_cache.L3CacheInfo.cacheSize = 2*1024*1024;
			_cache.L3CacheInfo.lineSize = 64;
			_cache.L3CacheInfo.associativity = 8;
			_cache.L3CacheInfo.isUnified = true;
			break;
		case 0x29:
			_cache.L3CacheInfo.cacheSize = 4*1024*1024;
			_cache.L3CacheInfo.lineSize = 64;
			_cache.L3CacheInfo.associativity = 8;
			_cache.L3CacheInfo.isUnified = true;
			break;
		case 0x2C:
			_cache.L1DCacheInfo.cacheSize = 32*1024;
			_cache.L1DCacheInfo.lineSize = 64;
			_cache.L1DCacheInfo.associativity = 8;
			_cache.L1DCacheInfo.isUnified = false;
			break;
		case 0x30:
			_cache.L1ICacheInfo.cacheSize = 32*1024;
			_cache.L1ICacheInfo.lineSize = 64;
			_cache.L1ICacheInfo.associativity = 8;
			_cache.L1ICacheInfo.isUnified = false;
			break;
		case 0x39:
			_cache.L2CacheInfo.cacheSize = 128*1024;
			_cache.L2CacheInfo.lineSize = 64;
			_cache.L2CacheInfo.associativity = 4;
			_cache.L2CacheInfo.isUnified = true;
			break;
		case 0x3A:
			_cache.L2CacheInfo.cacheSize = 192*1024;
			_cache.L2CacheInfo.lineSize = 64;
			_cache.L2CacheInfo.associativity = 6;
			_cache.L2CacheInfo.isUnified = true;
			break;
		case 0x3B:
			_cache.L2CacheInfo.cacheSize = 128*1024;
			_cache.L2CacheInfo.lineSize = 64;
			_cache.L2CacheInfo.associativity = 2;
			_cache.L2CacheInfo.isUnified = true;
			break;
		case 0x3C:
			_cache.L2CacheInfo.cacheSize = 256*1024;
			_cache.L2CacheInfo.lineSize = 64;
			_cache.L2CacheInfo.associativity = 4;
			_cache.L2CacheInfo.isUnified = true;
			break;
		case 0x3D:
			_cache.L2CacheInfo.cacheSize = 384*1024;
			_cache.L2CacheInfo.lineSize = 64;
			_cache.L2CacheInfo.associativity = 6;
			_cache.L2CacheInfo.isUnified = true;
			break;
		case 0x3E:
			_cache.L2CacheInfo.cacheSize = 512*1024;
			_cache.L2CacheInfo.lineSize = 64;
			_cache.L2CacheInfo.associativity = 4;
			_cache.L2CacheInfo.isUnified = true;
			break;
		case 0x41:
			_cache.L2CacheInfo.cacheSize = 128*1024;
			_cache.L2CacheInfo.lineSize = 32;
			_cache.L2CacheInfo.associativity = 4;
			_cache.L2CacheInfo.isUnified = true;
			break;
		case 0x42:
			_cache.L2CacheInfo.cacheSize = 256*1024;
			_cache.L2CacheInfo.lineSize = 32;
			_cache.L2CacheInfo.associativity = 4;
			_cache.L2CacheInfo.isUnified = true;
			break;
		case 0x43:
			_cache.L2CacheInfo.cacheSize = 512*1024;
			_cache.L2CacheInfo.lineSize = 32;
			_cache.L2CacheInfo.associativity = 4;
			_cache.L2CacheInfo.isUnified = true;
			break;
		case 0x44:
			_cache.L2CacheInfo.cacheSize = 1024*1024;
			_cache.L2CacheInfo.lineSize = 32;
			_cache.L2CacheInfo.associativity = 4;
			_cache.L2CacheInfo.isUnified = true;
			break;
		case 0x45:
			_cache.L2CacheInfo.cacheSize = 2*1024*1024;
			_cache.L2CacheInfo.lineSize = 32;
			_cache.L2CacheInfo.associativity = 4;
			_cache.L2CacheInfo.isUnified = true;
			break;
		case 0x46:
			_cache.L3CacheInfo.cacheSize = 4*1024*1024;
			_cache.L3CacheInfo.lineSize = 64;
			_cache.L3CacheInfo.associativity = 4;
			_cache.L3CacheInfo.isUnified = true;
			break;
		case 0x47:
			_cache.L3CacheInfo.cacheSize = 8*1024*1024;
			_cache.L3CacheInfo.lineSize = 64;
			_cache.L3CacheInfo.associativity = 8;
			_cache.L3CacheInfo.isUnified = true;
			break;
		case 0x48:
			_cache.L2CacheInfo.cacheSize = 3*1024*1024;
			_cache.L2CacheInfo.lineSize = 64;
			_cache.L2CacheInfo.associativity = 12;
			_cache.L3CacheInfo.isUnified = true;
			break;
		case 0x49:
			if (YEP_LIKELY(_vendor == YepCpuVendorIntel) && YEP_UNLIKELY(_modelInfo.model == 0x06) && YEP_UNLIKELY(_modelInfo.family == 0x0F)) {
				_cache.L3CacheInfo.cacheSize = 4*1024*1024;
				_cache.L3CacheInfo.lineSize = 64;
				_cache.L3CacheInfo.associativity = 16;
				_cache.L3CacheInfo.isUnified = true;
			} else {
				_cache.L2CacheInfo.cacheSize = 4*1024*1024;
				_cache.L2CacheInfo.lineSize = 64;
				_cache.L2CacheInfo.associativity = 16;
				_cache.L2CacheInfo.isUnified = true;
			}
			break;
		case 0x4A:
			_cache.L3CacheInfo.cacheSize = 6*1024*1024;
			_cache.L3CacheInfo.lineSize = 64;
			_cache.L3CacheInfo.associativity = 12;
			_cache.L3CacheInfo.isUnified = true;
			break;
		case 0x4B:
			_cache.L3CacheInfo.cacheSize = 8*1024*1024;
			_cache.L3CacheInfo.lineSize = 64;
			_cache.L3CacheInfo.associativity = 16;
			_cache.L3CacheInfo.isUnified = true;
			break;
		case 0x4C:
			_cache.L3CacheInfo.cacheSize = 12*1024*1024;
			_cache.L3CacheInfo.lineSize = 64;
			_cache.L3CacheInfo.associativity = 12;
			_cache.L3CacheInfo.isUnified = true;
			break;
		case 0x4D:
			_cache.L3CacheInfo.cacheSize = 16*1024*1024;
			_cache.L3CacheInfo.lineSize = 64;
			_cache.L3CacheInfo.associativity = 16;
			_cache.L3CacheInfo.isUnified = true;
			break;
		case 0x4E:
			_cache.L2CacheInfo.cacheSize = 6*1024*1024;
			_cache.L2CacheInfo.lineSize = 64;
			_cache.L2CacheInfo.associativity = 24;
			_cache.L2CacheInfo.isUnified = true;
			break;
		case 0x60:
			_cache.L1DCacheInfo.cacheSize = 16*1024;
			_cache.L1DCacheInfo.lineSize = 64;
			_cache.L1DCacheInfo.associativity = 8;
			_cache.L1DCacheInfo.isUnified = false;
			break;
		case 0x66:
			_cache.L1DCacheInfo.cacheSize = 8*1024;
			_cache.L1DCacheInfo.lineSize = 64;
			_cache.L1DCacheInfo.associativity = 4;
			_cache.L1DCacheInfo.isUnified = false;
			break;
		case 0x67:
			_cache.L1DCacheInfo.cacheSize = 16*1024;
			_cache.L1DCacheInfo.lineSize = 64;
			_cache.L1DCacheInfo.associativity = 4;
			_cache.L1DCacheInfo.isUnified = false;
			break;
		case 0x68:
			_cache.L1DCacheInfo.cacheSize = 32*1024;
			_cache.L1DCacheInfo.lineSize = 64;
			_cache.L1DCacheInfo.associativity = 4;
			_cache.L1DCacheInfo.isUnified = false;
			break;
		case 0x70:
			_cache.traceCacheInfo.microops = 12*1024;
			_cache.traceCacheInfo.associativity = 8;
			break;
		case 0x71:
			_cache.traceCacheInfo.microops = 16*1024;
			_cache.traceCacheInfo.associativity = 8;
			break;
		case 0x72:
			_cache.traceCacheInfo.microops = 32*1024;
			_cache.traceCacheInfo.associativity = 8;
			break;
		case 0x73:
			_cache.traceCacheInfo.microops = 64*1024;
			_cache.traceCacheInfo.associativity = 8;
			break;
		case 0x78:
			_cache.L2CacheInfo.cacheSize = 1024*1024;
			_cache.L2CacheInfo.lineSize = 64;
			_cache.L2CacheInfo.associativity = 4;
			_cache.L2CacheInfo.isUnified = true;
			break;
		case 0x79:
			_cache.L2CacheInfo.cacheSize = 128*1024;
			_cache.L2CacheInfo.lineSize = 64;
			_cache.L2CacheInfo.associativity = 8;
			_cache.L2CacheInfo.isUnified = true;
			break;
		case 0x7A:
			_cache.L2CacheInfo.cacheSize = 256*1024;
			_cache.L2CacheInfo.lineSize = 64;
			_cache.L2CacheInfo.associativity = 8;
			_cache.L2CacheInfo.isUnified = true;
			break;
		case 0x7B:
			_cache.L2CacheInfo.cacheSize = 512*1024;
			_cache.L2CacheInfo.lineSize = 64;
			_cache.L2CacheInfo.associativity = 8;
			_cache.L2CacheInfo.isUnified = true;
			break;
		case 0x7C:
			_cache.L2CacheInfo.cacheSize = 1024*1024;
			_cache.L2CacheInfo.lineSize = 64;
			_cache.L2CacheInfo.associativity = 8;
			_cache.L2CacheInfo.isUnified = true;
			break;
		case 0x7D:
			_cache.L2CacheInfo.cacheSize = 2*1024*1024;
			_cache.L2CacheInfo.lineSize = 64;
			_cache.L2CacheInfo.associativity = 8;
			_cache.L2CacheInfo.isUnified = true;
			break;
		case 0x7F:
			_cache.L2CacheInfo.cacheSize = 512*1024;
			_cache.L2CacheInfo.lineSize = 64;
			_cache.L2CacheInfo.associativity = 2;
			_cache.L2CacheInfo.isUnified = true;
			break;
		case 0x80:
			if (_vendor == YepCpuVendorCyrix) {
				_cache.L1ICacheInfo.cacheSize = 16*1024;
				_cache.L1DCacheInfo.cacheSize = 16*1024;
				_cache.L1ICacheInfo.lineSize = 16;
				_cache.L1DCacheInfo.lineSize = 16;
				_cache.L1ICacheInfo.associativity = 4;
				_cache.L1DCacheInfo.associativity = 4;
				_cache.L1ICacheInfo.isUnified = true;
				_cache.L1DCacheInfo.isUnified = true;
			} else {
				_cache.L2CacheInfo.cacheSize = 512*1024;
				_cache.L2CacheInfo.lineSize = 64;
				_cache.L2CacheInfo.associativity = 8;
				_cache.L2CacheInfo.isUnified = true;
			}
			break;
		case 0x82:
			_cache.L2CacheInfo.cacheSize = 256*1024;
			_cache.L2CacheInfo.lineSize = 32;
			_cache.L2CacheInfo.associativity = 8;
			_cache.L2CacheInfo.isUnified = true;
			break;
		case 0x83:
			_cache.L2CacheInfo.cacheSize = 512*1024;
			_cache.L2CacheInfo.lineSize = 32;
			_cache.L2CacheInfo.associativity = 8;
			_cache.L2CacheInfo.isUnified = true;
			break;
		case 0x84:
			_cache.L2CacheInfo.cacheSize = 1024*1024;
			_cache.L2CacheInfo.lineSize = 32;
			_cache.L2CacheInfo.associativity = 8;
			_cache.L2CacheInfo.isUnified = true;
			break;
		case 0x85:
			_cache.L2CacheInfo.cacheSize = 2*1024*1024;
			_cache.L2CacheInfo.lineSize = 32;
			_cache.L2CacheInfo.associativity = 8;
			_cache.L2CacheInfo.isUnified = true;
			break;
		case 0x86:
			_cache.L2CacheInfo.cacheSize = 512*1024;
			_cache.L2CacheInfo.lineSize = 64;
			_cache.L2CacheInfo.associativity = 4;
			_cache.L2CacheInfo.isUnified = true;
			break;
		case 0x87:
			_cache.L2CacheInfo.cacheSize = 1024*1024;
			_cache.L2CacheInfo.lineSize = 64;
			_cache.L2CacheInfo.associativity = 8;
			_cache.L2CacheInfo.isUnified = true;
			break;
		case 0xD0:
			_cache.L3CacheInfo.cacheSize = 512*1024;
			_cache.L3CacheInfo.lineSize = 64;
			_cache.L3CacheInfo.associativity = 4;
			_cache.L3CacheInfo.isUnified = true;
			break;
		case 0xD1:
			_cache.L3CacheInfo.cacheSize = 1024*1024;
			_cache.L3CacheInfo.lineSize = 64;
			_cache.L3CacheInfo.associativity = 4;
			_cache.L3CacheInfo.isUnified = true;
			break;
		case 0xD2:
			_cache.L3CacheInfo.cacheSize = 2*1024*1024;
			_cache.L3CacheInfo.lineSize = 64;
			_cache.L3CacheInfo.associativity = 4;
			_cache.L3CacheInfo.isUnified = true;
			break;
		case 0xD6:
			_cache.L3CacheInfo.cacheSize = 1024*1024;
			_cache.L3CacheInfo.lineSize = 64;
			_cache.L3CacheInfo.associativity = 8;
			_cache.L3CacheInfo.isUnified = true;
			break;
		case 0xD7:
			_cache.L3CacheInfo.cacheSize = 2*1024*1024;
			_cache.L3CacheInfo.lineSize = 64;
			_cache.L3CacheInfo.associativity = 8;
			_cache.L3CacheInfo.isUnified = true;
			break;
		case 0xD8:
			_cache.L3CacheInfo.cacheSize = 4*1024*1024;
			_cache.L3CacheInfo.lineSize = 64;
			_cache.L3CacheInfo.associativity = 8;
			_cache.L3CacheInfo.isUnified = true;
			break;
		case 0xDC:
			_cache.L3CacheInfo.cacheSize = 3*512*1024;
			_cache.L3CacheInfo.lineSize = 64;
			_cache.L3CacheInfo.associativity = 12;
			_cache.L3CacheInfo.isUnified = true;
			break;
		case 0xDD:
			_cache.L3CacheInfo.cacheSize = 3*1024*1024;
			_cache.L3CacheInfo.lineSize = 64;
			_cache.L3CacheInfo.associativity = 12;
			_cache.L3CacheInfo.isUnified = true;
			break;
		case 0xDE:
			_cache.L3CacheInfo.cacheSize = 6*1024*1024;
			_cache.L3CacheInfo.lineSize = 64;
			_cache.L3CacheInfo.associativity = 12;
			_cache.L3CacheInfo.isUnified = true;
			break;
		case 0xE2:
			_cache.L3CacheInfo.cacheSize = 2*1024*1024;
			_cache.L3CacheInfo.lineSize = 64;
			_cache.L3CacheInfo.associativity = 16;
			_cache.L3CacheInfo.isUnified = true;
			break;
		case 0xE3:
			_cache.L3CacheInfo.cacheSize = 4*1024*1024;
			_cache.L3CacheInfo.lineSize = 64;
			_cache.L3CacheInfo.associativity = 16;
			_cache.L3CacheInfo.isUnified = true;
			break;
		case 0xE4:
			_cache.L3CacheInfo.cacheSize = 8*1024*1024;
			_cache.L3CacheInfo.lineSize = 64;
			_cache.L3CacheInfo.associativity = 16;
			_cache.L3CacheInfo.isUnified = true;
			break;
		case 0xEA:
			_cache.L3CacheInfo.cacheSize = 12*1024*1024;
			_cache.L3CacheInfo.lineSize = 64;
			_cache.L3CacheInfo.associativity = 24;
			_cache.L3CacheInfo.isUnified = true;
			break;
		case 0xEB:
			_cache.L3CacheInfo.cacheSize = 18*1024*1024;
			_cache.L3CacheInfo.lineSize = 64;
			_cache.L3CacheInfo.associativity = 24;
			_cache.L3CacheInfo.isUnified = true;
			break;
		case 0xEC:
			_cache.L3CacheInfo.cacheSize = 24*1024*1024;
			_cache.L3CacheInfo.lineSize = 64;
			_cache.L3CacheInfo.associativity = 24;
			_cache.L3CacheInfo.isUnified = true;
			break;
		case 0xF0:
			_cache.prefetchLineSize = 64;
			break;
		case 0xF1:
			_cache.prefetchLineSize = 128;
			break;
	}
}

static void _yepLibrary_DecodeX86CpuDeterministicCacheParameters(Yep32u cacheType, Yep32u cacheLevel, Yep32u cacheSize, Yep32u lineSize, Yep32u associativity) {
	if YEP_LIKELY(cacheLevel == 1) {
		// 1st level cache
		switch (cacheType) {
			case 1: // Data cache
				_cache.L1DCacheInfo.cacheSize = cacheSize;
				_cache.L1DCacheInfo.lineSize = static_cast<Yep16u>(lineSize);
				_cache.L1DCacheInfo.associativity = static_cast<Yep16u>(associativity);
				_cache.L1DCacheInfo.isUnified = false;
				break;
			case 2: // Instruction cache
				_cache.L1ICacheInfo.cacheSize = cacheSize;
				_cache.L1ICacheInfo.lineSize = static_cast<Yep16u>(lineSize);
				_cache.L1ICacheInfo.associativity = static_cast<Yep16u>(associativity);
				_cache.L1DCacheInfo.isUnified = false;
				break;
			case 3: // Unified cache
				_cache.L1ICacheInfo.cacheSize = cacheSize;
				_cache.L1DCacheInfo.cacheSize = cacheSize;
				_cache.L1ICacheInfo.lineSize = static_cast<Yep16u>(lineSize);
				_cache.L1DCacheInfo.lineSize = static_cast<Yep16u>(lineSize);
				_cache.L1ICacheInfo.associativity = static_cast<Yep16u>(associativity);
				_cache.L1DCacheInfo.associativity = static_cast<Yep16u>(associativity);
				_cache.L1ICacheInfo.isUnified = true;
				_cache.L1DCacheInfo.isUnified = true;
				break;
		}
	} else if YEP_LIKELY((cacheLevel == 2) && (cacheType == 3)) {
		// 2nd level unified cache
		_cache.L2CacheInfo.cacheSize = cacheSize;
		_cache.L2CacheInfo.lineSize = static_cast<Yep16u>(lineSize);
		_cache.L2CacheInfo.associativity = static_cast<Yep16u>(associativity);
		_cache.L2CacheInfo.isUnified = true;
	} else if YEP_LIKELY((cacheLevel == 3) && (cacheType == 3)) {
		// 3rd level unified cache
		_cache.L3CacheInfo.cacheSize = cacheSize;
		_cache.L3CacheInfo.lineSize = static_cast<Yep16u>(lineSize);
		_cache.L3CacheInfo.associativity = static_cast<Yep16u>(associativity);
		_cache.L3CacheInfo.isUnified = true;
	}
}
#endif

#if defined(YEP_ARM_CPU) && defined(YEP_WINDOWS_OS)
void _yepLibrary_InitWindowsARMCpuIsaInfo() {
	Yep32u isaFeatures = YepARMIsaFeatureDefault;
	Yep32u simdFeatures = YepARMSimdFeatureDefault;

	isaFeatures |= YepARMFeatureV4;
	isaFeatures |= YepARMFeatureV5;
	isaFeatures |= YepARMFeatureV6;
	isaFeatures |= YepARMFeatureV7;
	isaFeatures |= YepARMFeatureThumb;
	isaFeatures |= YepARMFeatureThumb2;
	isaFeatures |= YepARMFeatureEDSP;
	isaFeatures |= YepARMFeatureVFP3;

	if YEP_LIKELY(::IsProcessorFeaturePresent(PF_ARM_VFP_32_REGISTERS_AVAILABLE) != 0) {
		isaFeatures |= YepARMIsaFeatureVFPd32;
	}
	if YEP_LIKELY(::IsProcessorFeaturePresent(PF_ARM_FMAC_INSTRUCTIONS_AVAILABLE) != 0) {
		isaFeatures |= YepARMFeatureVFP4;
	}
	if YEP_LIKELY(::IsProcessorFeaturePresent(PF_ARM_DIVIDE_INSTRUCTION_AVAILABLE) != 0) {
		isaFeatures |= YepARMFeatureDiv;
	}
	if YEP_LIKELY(::IsProcessorFeaturePresent(PF_ARM_64BIT_LOADSTORE_ATOMIC) != 0) {
		isaFeatures |= YepARMIsaFeatureV6K;
	}
	if YEP_LIKELY(::IsProcessorFeaturePresent(PF_ARM_NEON_INSTRUCTIONS_AVAILABLE) != 0) {
		simdFeatures |= YepARMSimdFeatureNEON;
		isaFeatures |= YepARMIsaFeatureVFPd32;
	}

	_isaFeatures = isaFeatures;
	_simdFeatures = simdFeatures;
}

void _yepLibrary_InitWindowsARMMicroarchitectureInfo() {
	SYSTEM_INFO systemInfo;
	::GetNativeSystemInfo(&systemInfo);
	// On WinCE SYSTEM_INFO.dwProcessorType matches the "Part number" field of Main ID register,
	// at least for those CPUs listed in WinNT.h
	// Hope this will also hold for future CPUs on Windows 8
	// Since Windows 8 will only run on ARMv7 we check only new CPUs
	switch (systemInfo.dwProcessorType) {
		case 0xC05: // ARM Cortex-A5
			_microarchitecture = YepCpuMicroarchitectureCortexA5;
			_vendor = YepCpuVendorARM;
			break;
		case 0xC07: // ARM Cortex-A7
			_microarchitecture = YepCpuMicroarchitectureCortexA7;
			_vendor = YepCpuVendorARM;
			break;
		case 0xC08: // ARM Cortex-A8
			_microarchitecture = YepCpuMicroarchitectureCortexA8;
			_vendor = YepCpuVendorARM;
			break;
		case 0xC09: // ARM Cortex-A9
			_microarchitecture = YepCpuMicroarchitectureCortexA9;
			_vendor = YepCpuVendorARM;
			break;
		case 0xC0F: // ARM Cortex-A15
			_microarchitecture = YepCpuMicroarchitectureCortexA15;
			_vendor = YepCpuVendorARM;
			break;
		case 0x00F: // Single-core Qualcomm Snapdragon
			// These are mainly Scorpions, but might be some Cortex-A5 as well
			// I know no simple way to distinguish between them, so lets assume everything is Scorpion
			// In future, cache size could be used to separate the two microarchitectures.
		case 0x02D: // Dual-core Qualcomm Scorpion
			_microarchitecture = YepCpuMicroarchitectureScorpion;
			_vendor = YepCpuVendorQualcomm;
			break;
		case 0x04D: // Dual-core Qualcomm Krait
			_microarchitecture = YepCpuMicroarchitectureKrait;
			_vendor = YepCpuVendorQualcomm;
			break;
		case 0x06F: // Quad-core Qualcomm Krait
			_microarchitecture = YepCpuMicroarchitectureKrait;
			_vendor = YepCpuVendorQualcomm;
			break;
		case 0x581: // Marvell Armada 510 (PJ4 core, ARMv7)
			_microarchitecture = YepCpuMicroarchitecturePJ4;
			_vendor = YepCpuVendorMarvell;
			break;
	}
}

YepStatus _yepLibrary_InitWindowsARMMicroarchitectureInfo() {
	DWORD infoLength = 0;
	BOOL syscallResult = ::GetLogicalProcessorInformation(NULL, &infoLength);
	if YEP_LIKELY(syscallResult == TRUE) {
		// Windows does not report CPU information
		return YepStatusUnsupportedSoftware;
	} else if (::GetLastError() != ERROR_INSUFFICIENT_BUFFER) 
		return YepStatusSystemError;
	}
	HANDLE heap = ::GetProcessHeap();
	if (heap == NULL) {
		return YepStatusSystemError;
	}
	const PSYSTEM_LOGICAL_PROCESSOR_INFORMATION infoBuffer =
		static_cast<PSYSTEM_LOGICAL_PROCESSOR_INFORMATION>(::HeapAlloc(heap, HEAP_ZERO_MEMORY, infoLength));
	if (infoBuffer == NULL) {
		return YepStatusSystemError;
	}
	
	syscallResult = ::GetLogicalProcessorInformation(buffer, &infoLength);
	
	DWORD infoChunks = bufferSize / sizeof(SYSTEM_LOGICAL_PROCESSOR_INFORMATION);
	for (DWORD infoChunk = 0; infoChunk < infoChunks; infoChunk++) {

		// We look only at caches connected to the first CPU
		if (infoBuffer[infoChunk].ProcessorMask & 0x00000001) {

			// Ignore other information (NUMA, Hyperthreading, CPU Package)
			if (infoBuffer[infoChunk].Relationship == RelationCache) {

				const BYTE level = infoBuffer[infoChunk].Cache.Level;
				const BYTE associativity = infoBuffer[infoChunk].Cache.Associativity;
				const WORD lineSize = infoBuffer[infoChunk].Cache.LineSize;
				const DWORD size = infoBuffer[infoChunk].Cache.Size;
				if (associativity == 0xFFu) {
					associativity = size / lineSize;
				}

				switch (infoBuffer[infoChunk].Cache.Type == 1) {
					case CacheUnified:
						switch (level) {
							case 1:
								_cache.L1ICacheInfo.cacheSize = size;
								_cache.L1ICacheInfo.lineSize = lineSize;
								_cache.L1ICacheInfo.associativity = associativity;
								_cache.L1ICacheInfo.isUnified = true;

								_cache.L1DCacheInfo.cacheSize = size;
								_cache.L1DCacheInfo.lineSize = lineSize;
								_cache.L1DCacheInfo.associativity = associativity;
								_cache.L1DCacheInfo.isUnified = true;
								break;
							case 2:
								_cache.L2CacheInfo.cacheSize = size;
								_cache.L2CacheInfo.lineSize = lineSize;
								_cache.L2CacheInfo.associativity = associativity;
								_cache.L2CacheInfo.isUnified = true;
								break;
							case 3:
								_cache.L3CacheInfo.cacheSize = size;
								_cache.L3CacheInfo.lineSize = lineSize;
								_cache.L3CacheInfo.associativity = associativity;
								_cache.L3CacheInfo.isUnified = true;
								break;
						}
						break;
					case CacheInstruction:
						if (level == 1) {
							_cache.L1ICacheInfo.cacheSize = size;
							_cache.L1ICacheInfo.lineSize = lineSize;
							_cache.L1ICacheInfo.associativity = associativity;
							_cache.L1ICacheInfo.isUnified = false;
						}
						break;
					case CacheData:
						if (level == 1) {
							_cache.L1DCacheInfo.cacheSize = size;
							_cache.L1DCacheInfo.lineSize = lineSize;
							_cache.L1DCacheInfo.associativity = associativity;
							_cache.L1DCacheInfo.isUnified = false;
						}
						break;
					case CacheTrace:
						if (level == 1) {
							_cache.traceCacheInfo.microops = size;
							if (infoBuffer[infoChunk].Cache.Associativity == 0xFFu) {
								_cache.traceCacheInfo.associativity = size;
							} else {
								_cache.traceCacheInfo.associativity = associativity;
							}
						}
						break;
				}
			}
		}
	}
	syscallResult = ::HeapFree(heap, 0, infoBuffer);
	if (syscallResult == 0) {
		return YepStatusSystemError;
	}
	return YepStatusOk;
}
#endif

#if (defined(YEP_ARM_CPU) || defined(YEP_MIPS_CPU)) && defined(YEP_LINUX_OS)
static YepStatus _yepLibrary_DetectLinuxPerfEventSupport() {
	struct perf_event_attr perfEventAttributes;
	_yepLibrary_memset(&perfEventAttributes, 0, sizeof(perfEventAttributes));
	perfEventAttributes.size = sizeof(perfEventAttributes);
	perfEventAttributes.type = PERF_TYPE_HARDWARE;
	perfEventAttributes.config = PERF_COUNT_HW_CPU_CYCLES;
	perfEventAttributes.disabled = 1;

	const pid_t processId = 0; // Current process
	const int cpu = -1; // Any CPU
	const int groupFileDescriptor = -1; // New group
	const int flags = 0; // Default
	const int perfCounter = yepSyscall_perf_event_open(&perfEventAttributes, processId, cpu, groupFileDescriptor, flags);
	if YEP_LIKELY(perfCounter >= 0) {
		const int syscallIoctlResetResult = yepSyscall_ioctl(perfCounter, PERF_EVENT_IOC_RESET);
		if YEP_LIKELY(syscallIoctlResetResult == 0) {
			const int syscallIoctlEnableResult = yepSyscall_ioctl(perfCounter, PERF_EVENT_IOC_ENABLE);
			if YEP_LIKELY(syscallIoctlEnableResult == 0) {
				const int syscallIoctlDisableResult = yepSyscall_ioctl(perfCounter, PERF_EVENT_IOC_DISABLE);
				if YEP_LIKELY(syscallIoctlDisableResult == 0) {
					Yep64u cpuCycles = 0;
					const ssize_t bytesRead = yepSyscall_read(perfCounter, &cpuCycles, sizeof(cpuCycles));
					if YEP_LIKELY(bytesRead == sizeof(cpuCycles)) {
						// Sanity check
						if YEP_LIKELY(cpuCycles != 0) {
							const int syscallCloseResult = yepSyscall_close(perfCounter);
							if YEP_LIKELY(syscallCloseResult == 0) {
								_systemFeatures |= YepSystemFeatureCycleCounter | YepSystemFeatureCycleCounter64Bit;
								return YepStatusOk;
							} else {
								// Avoid calling the close twice
								return YepStatusSystemError;
							}
						}
					}
				}
			}
		}
		yepSyscall_close(perfCounter);
	}
	return YepStatusSystemError;
}
#endif

#if defined(YEP_ARM_CPU) && defined(YEP_LINUX_OS)
static YepStatus _yepLibrary_DetectLinuxARMCycleCounterAccess() {
	const Yep64u pmccntrReadResult = _yepLibrary_ReadCoprocessor(&_yepLibrary_ReadPMCCNTR);
	const Yep32u pmccntrReadSuccess = yepBuiltin_GetLowPart_64u_32u(pmccntrReadResult);
	if YEP_UNLIKELY(pmccntrReadSuccess) {
		const Yep32u pmccntrStart = yepBuiltin_GetHighPart_64u_32u(pmccntrReadResult);
		const Yep64u pmccntrReReadResult = _yepLibrary_ReadCoprocessor(&_yepLibrary_ReadPMCCNTR);
		const Yep32u pmccntrReReadSuccess = yepBuiltin_GetLowPart_64u_32u(pmccntrReReadResult);
		if YEP_LIKELY(pmccntrReReadSuccess) {
			const Yep32u pmccntrEnd = yepBuiltin_GetHighPart_64u_32u(pmccntrReReadResult);
			if YEP_LIKELY(pmccntrEnd != pmccntrStart) {
				_systemFeatures |= YepSystemFeatureCycleCounter;
				return YepStatusOk;
			}
		}
	}
	return YepStatusUnsupportedSoftware;
}

static void _yepLibrary_DecodeLinuxARMProcessorNumber(const char* valueStart, const char* valueEnd) {
	for (const char* valuePointer = valueStart; valuePointer != valueEnd; valuePointer++) {
		const char digit = *valuePointer;
		if ((digit < '0') || (digit > '9'))
			return;
	}
	_cpuInfo.processors += 1;
}

/*
 *	Full list of ARM features reported in /proc/cpuinfo:
 *	
 *	* swp - support for SWP instruction (deprecated in ARMv7, can be removed in future)
 *	* half - support for half-word loads and stores. These instruction are part of ARMv4,
 *	         so no need to check it on supported CPUs.
 *	* thumb - support for 16-bit Thumb instruction set. Note that BX instruction is detected
 *	          by ARMv4T architecture, not by this flag.
 *	* 26bit - old CPUs merged 26-bit PC and program status register (flags) into 32-bit PC
 *	          and had special instructions for working with packed PC. Now it is all deprecated.
 *	* fastmult - most old ARM CPUs could only compute 2 bits of multiplication result per clock
 *	             cycle, but CPUs with M suffix (e.g. ARM7TDMI) could compute 4 bits per cycle.
 *	             Of course, now it makes no sense.
 *	* fpa - floating point accelerator available. On original ARM ABI all floating-point operations
 *	        generated FPA instructions. If FPA was not available, these instructions generated
 *	        "illegal operation" interrupts, and the OS processed them by emulating the FPA instructions.
 *	        Debian used this ABI before it switched to EABI. Now FPA is deprecated.
 *	* vfp - vector floating point instructions. Available on most modern CPUs (as part of VFPv3).
 *	        Required by Android ARMv7A ABI and by Ubuntu on ARM.
 *              Note: there is no flag for VFPv2.
 *	* edsp - V5E instructions: saturating add/sub and 16-bit x 16-bit -> 32/64-bit multiplications.
 *	         Required on Android, supported by all CPUs in production.
 *	* java - Jazelle extension. Supported on most CPUs.
 *	* iwmmxt - Intel/Marvell Wireless MMX instructions. 64-bit integer SIMD.
 *	           Supported on XScale (Since PXA270) and Sheeva (PJ1, PJ4) architectures.
 *	           Note that there is no flag for WMMX2 instructions.
 *	* crunch - Maverick Crunch instructions. Junk.
 *	* thumbee - ThumbEE instructions. Almost no documentation is available.
 *	* neon - NEON instructions (aka Advanced SIMD). MVFR1 register gives more
 *	         fine-grained information on particular supported features, but
 *	         the Linux kernel exports only a single flag for all of them.
 *	         According to ARMv7A docs it also implies the availability of VFPv3
 *	         (with 32 double-precision registers d0-d31).
 *	* vfpv3 - VFPv3 instructions. Available on most modern CPUs. Augment VFPv2 by
 *	          conversion to/from integers and load constant instructions.
 *	          Required by Android ARMv7A ABI and by Ubuntu on ARM.
 *	* vfpv3d16 - VFPv3 instructions with only 16 double-precision registers (d0-d15).
 *	* tls - software thread ID registers.
 *	        Used by kernel (and likely libc) for efficient implementation of TLS.
 *	* vfpv4 - fused multiply-add instructions.
 *	* idiva - DIV instructions available in ARM mode.
 *	* idivt - DIV instructions available in Thumb mode.
 *
 *	/proc/cpuinfo on ARM is populated in file arch/arm/kernel/setup.c in Linux kernel
 *	Note that some devices may use patched Linux kernels with different feature names.
 *	However, the names above were checked on a large number of /proc/cpuinfo listings.
 */
static void _yepLibrary_DecodeLinuxARMCpuFeatures(const char* valueStart, const char* valueEnd) {
	const char* featureStart = valueStart;
	const char* featureEnd;
	_cpuInfo.featuresValid = true;
	
	do {
		featureEnd = featureStart + 1;
		for (; featureEnd != valueEnd; featureEnd++) {
			if (*featureEnd == ' ')
				break;
		}
		const YepSize featureLength = featureEnd - featureStart;
		
		switch (featureLength) {
			case 3:
				if (!_yepLibrary_strncmp(featureStart, "swp", 3)) {
					_cpuInfo.featureSwp = true;
				} else if (!_yepLibrary_strncmp(featureStart, "fpa", 3)) {
					_cpuInfo.featureFpa = true;
				} else if (!_yepLibrary_strncmp(featureStart, "vfp", 3)) {
					_cpuInfo.featureVfp = true;
				} else if (!_yepLibrary_strncmp(featureStart, "tls", 3)) {
					_cpuInfo.featureTls = true;
				}
				break;
			case 4:
				if (!_yepLibrary_strncmp(featureStart, "half", 4)) {
					_cpuInfo.featureHalf = true;
				} else if (!_yepLibrary_strncmp(featureStart, "edsp", 4)) {
					_cpuInfo.featureEdsp = true;
				} else if (!_yepLibrary_strncmp(featureStart, "java", 4)) {
					_cpuInfo.featureJava = true;
				} else if (!_yepLibrary_strncmp(featureStart, "neon", 4)) {
					_cpuInfo.featureNeon = true;
				}
				break;
			case 5:
				if (!_yepLibrary_strncmp(featureStart, "thumb", 5)) {
					_cpuInfo.featureThumb = true;
				} else if (!_yepLibrary_strncmp(featureStart, "26bit", 5)) {
					_cpuInfo.feature26bit = true;
				} else if (!_yepLibrary_strncmp(featureStart, "vfpv3", 5)) {
					_cpuInfo.featureVfpv3 = true;
				} else if (!_yepLibrary_strncmp(featureStart, "vfpv4", 5)) {
					_cpuInfo.featureVfpv4 = true;
				} else if (!_yepLibrary_strncmp(featureStart, "idiva", 5)) {
					_cpuInfo.featureIdiva = true;
				} else if (!_yepLibrary_strncmp(featureStart, "idivt", 5)) {
					_cpuInfo.featureIdivt = true;
				}
				break;
			case 6:
				if (!_yepLibrary_strncmp(featureStart, "iwmmxt", 6)) {
					_cpuInfo.featureIwmmxt = true;
				} else if (!_yepLibrary_strncmp(featureStart, "crunch", 6)) {
					_cpuInfo.featureCrunch = true;
				}
				break;
			case 7:
				if (!_yepLibrary_strncmp(featureStart, "thumbee", 7)) {
					_cpuInfo.featureThumbee = true;
				}
				break;
			case 8:
				if (!_yepLibrary_strncmp(featureStart, "fastmult", 8)) {
					_cpuInfo.featureFastmult = true;
				} else if (!_yepLibrary_strncmp(featureStart, "vfpv3d16", 8)) {
					_cpuInfo.featureVfpv3d16 = true;
				}
				break;
		}
		featureStart = featureEnd;
		for (; featureStart != valueEnd; featureStart++) {
			if (*featureStart != ' ')
				break;
		}
	} while (featureStart != featureEnd);
}

static void _yepLibrary_DecodeLinuxARMCpuPart(const char* valueStart, const char* valueEnd) {
	const YepSize valueLength = valueEnd - valueStart;

	// Value should contain hex prefix (0x) and one to three hex digits.
	// I have never seen less than three digits as a value of this field,
	// but I don't think it is impossible to see such values in future.
	// Value can not contain more than three hex digits since
	// Main ID Register (MIDR) assigns only a 12-bit value for CPU part.
	if ((valueLength < 3) || (valueLength > 5))
		return;

	// Skip if there is no hex prefix (0x)
	if ((valueStart[0] != '0') || (valueStart[1] != 'x'))
		return;

	// Check if the values after hex prefix are really hex digits and decode them.
	Yep32u cpuPart = 0;
	for (const char* digitPosition = valueStart + 2; digitPosition != valueEnd; digitPosition++) {
		const char digitCharacter = *digitPosition;
		Yep32u digit;
		if ((digitCharacter >= '0') && (digitCharacter <= '9')) {
			digit = digitCharacter - '0';
		} else if ((digitCharacter >= 'A') && (digitCharacter <= 'F')) {
			digit = 10 + (digitCharacter - 'A');
		} else if ((digitCharacter >= 'a') && (digitCharacter <= 'f')) {
			digit = 10 + (digitCharacter - 'a');
		} else {
			// Not a hex digit
			return;
		}
		cpuPart = cpuPart * 16 + digit;
	}

	_cpuInfo.cpuPart = cpuPart;
}

static void _yepLibrary_DecodeLinuxARMCpuImplementer(const char* valueStart, const char* valueEnd) {
	const YepSize valueLength = valueEnd - valueStart;

	// Value should contain hex prefix (0x) and one or two hex digits.
	// I have never seen single hex digit as a value of this field,
	// but I don't think it is impossible in future.
	// Value can not contain more than two hex digits since
	// Main ID Register (MIDR) assigns only an 8-bit value for CPU implementer.
	if ((valueLength != 3) && (valueLength != 4))
		return;

	// Skip if there is no hex prefix (0x)
	if ((valueStart[0] != '0') || (valueStart[1] != 'x'))
		return;

	// Check if the values after hex prefix are really hex digits and decode them.
	const char hexDigits[2] = { ((valueLength == 3) ? '0' : valueEnd[-2]), valueEnd[-1] };
	Yep32u digitValues[2];
	for (YepSize digitPosition = 0; digitPosition < 2; digitPosition++) {
		const char digit = hexDigits[digitPosition];
		if ((digit >= '0') && (digit <= '9')) {
			digitValues[digitPosition] = digit - '0';
		} else if ((digit >= 'A') && (digit <= 'F')) {
			digitValues[digitPosition] = 10 + (digit - 'A');
		} else if ((digit >= 'a') && (digit <= 'f')) {
			digitValues[digitPosition] = 10 + (digit - 'a');
		} else {
			// Not a hex digit
			return;
		}
	}

	_cpuInfo.cpuImplementer = digitValues[0] * 16 + digitValues[1];
}

static void _yepLibrary_DecodeLinuxARMCpuVariant(const char* valueStart, const char* valueEnd) {
	const YepSize valueLength = valueEnd - valueStart;

	// Value should contain hex prefix (0x) and one hex digit.
	// Value can not contain more than one hex digits since
	// Main ID Register (MIDR) assigns only a 4-bit value for CPU variant.
	if (valueLength != 3)
		return;

	// Skip if there is no hex prefix (0x)
	if ((valueStart[0] != '0') || (valueStart[1] != 'x'))
		return;

	// Check if the value after hex prefix is indeed a hex digit and decode it.
	const char digit = valueStart[2];
	if ((digit >= '0') && (digit <= '9')) {
		_cpuInfo.cpuVariant = digit - '0';
	} else if ((digit >= 'A') && (digit <= 'F')) {
		_cpuInfo.cpuVariant = 10 + (digit - 'A');
	} else if ((digit >= 'a') && (digit <= 'f')) {
		_cpuInfo.cpuVariant = 10 + (digit - 'a');
	} else {
		// Not a hex digit
		return;
	}
}

static void _yepLibrary_DecodeLinuxARMCpuArchitecture(const char* valueStart, const char* valueEnd) {
	Yep32u architecture = 0;
	const char* valuePointer = valueStart;
	for (; valuePointer != valueEnd; valuePointer++) {
		const char character = *valuePointer;
		// Verify that CPU architecture is a decimal number
		if ((character < '0') || (character > '9'))
			break;

		architecture = architecture * 10 + (character - '0');
	}
	_cpuInfo.cpuArchitectureVersion = architecture;

	for (; valuePointer != valueEnd; valuePointer++) {
		const char character = *valuePointer;
		switch (character) {
			case 'E':
				_cpuInfo.cpuArchitectureE = true;
				break;
			case 'J':
				_cpuInfo.cpuArchitectureJ = true;
				break;
			case 'T':
				_cpuInfo.cpuArchitectureT = true;
				break;
		}
	}
}

static void _yepLibrary_DecodeLinuxARMCpuRevision(const char* valueStart, const char* valueEnd) {
	Yep32u revision = 0;
	for (const char* valuePointer = valueStart; valuePointer != valueEnd; valuePointer++) {
		const char character = *valuePointer;
		// Verify that CPU revision is a decimal number
		if ((character < '0') || (character > '9'))
			return;

		revision = revision * 10 + (character - '0');
	}

	_cpuInfo.cpuRevision = revision;
}

static void _yepLibrary_DecodeLinuxARMCacheNumber(const char* valueStart, const char* valueEnd, Yep32u& cacheNumber) {
	Yep32u number = 0;
	for (const char* valuePointer = valueStart; valuePointer != valueEnd; valuePointer++) {
		const char character = *valuePointer;
		// Verify that the value is a decimal number
		if ((character < '0') || (character > '9'))
			return;

		number = number * 10 + (character - '0');
	}

	cacheNumber = number;
}

/*	Decode a single line of /proc/cpuinfo information.
 *	Lines have format <words-with-spaces>[ ]*:[ ]<space-separated words>
 *	An example of /proc/cpuinfo (from Pandaboard-ES):
 *
 *		Processor       : ARMv7 Processor rev 10 (v7l)
 *		processor       : 0
 *		BogoMIPS        : 1392.74
 *
 *		processor       : 1
 *		BogoMIPS        : 1363.33
 *
 *		Features        : swp half thumb fastmult vfp edsp thumbee neon vfpv3
 *		CPU implementer : 0x41
 *		CPU architecture: 7
 *		CPU variant     : 0x2
 *		CPU part        : 0xc09
 *		CPU revision    : 10
 *
 *		Hardware        : OMAP4 Panda board
 *		Revision        : 0020
 *		Serial          : 0000000000000000
 */
static void _yepLibrary_DecodeLinuxARMCpuInfoLine(const char* lineStart, const char* lineEnd) {
	// Empty line. Skip.
	if (lineStart == lineEnd)
		return;

	// Search for ':' on the line.
	const char* separator = lineStart;
	for (; separator != lineEnd; separator++) {
		if (*separator == ':')
			break;
	}
	// Skip line if no ':' separator was found.
	if (separator == lineEnd)
		return;

	// Skip trailing spaces in key part.
	const char* infoKeyEnd = separator;
	for (; infoKeyEnd != lineStart; infoKeyEnd--) {
		if ((infoKeyEnd[-1] != ' ') && (infoKeyEnd[-1] != '\t'))
			break;
	}
	// Skip line if key contains nothing but spaces.
	if (infoKeyEnd == lineStart)
		return;

	// Skip leading spaces in value part.
	const char* infoValueStart = separator + 1;
	for (; infoValueStart != lineEnd; infoValueStart++) {
		if (*infoValueStart != ' ')
			break;
	}
	// Value part contains nothing but spaces. Skip line.
	if (infoValueStart == lineEnd)
		return;

	// Skip trailing spaces in value part (if any)
	const char* infoValueEnd = lineEnd;
	for (; infoValueEnd != separator; infoValueEnd--) {
		if (infoValueEnd[-1] != ' ')
			break;
	}

	YepSize infoKeyLength = infoKeyEnd - lineStart;
	switch (infoKeyLength) {
		case 6:
			if (!_yepLibrary_strncmp(lineStart, "I size", 6)) {
				_yepLibrary_DecodeLinuxARMCacheNumber(infoValueStart, infoValueEnd, _cpuInfo.iSize);
			} else if (!_yepLibrary_strncmp(lineStart, "I sets", 6)) {
				_yepLibrary_DecodeLinuxARMCacheNumber(infoValueStart, infoValueEnd, _cpuInfo.iSets);
			} else if (!_yepLibrary_strncmp(lineStart, "D size", 6)) {
				_yepLibrary_DecodeLinuxARMCacheNumber(infoValueStart, infoValueEnd, _cpuInfo.dSize);
			} else if (!_yepLibrary_strncmp(lineStart, "D sets", 6)) {
				_yepLibrary_DecodeLinuxARMCacheNumber(infoValueStart, infoValueEnd, _cpuInfo.dSets);
			}
			break;
		case 7:
			if (!_yepLibrary_strncmp(lineStart, "I assoc", 7)) {
				_yepLibrary_DecodeLinuxARMCacheNumber(infoValueStart, infoValueEnd, _cpuInfo.iAssoc);
			} else if (!_yepLibrary_strncmp(lineStart, "D assoc", 7)) {
				_yepLibrary_DecodeLinuxARMCacheNumber(infoValueStart, infoValueEnd, _cpuInfo.dAssoc);
			}
			break;
		case 8:
			if (!_yepLibrary_strncmp(lineStart, "CPU part", 8)) {
				_yepLibrary_DecodeLinuxARMCpuPart(infoValueStart, infoValueEnd);
			} else if (!_yepLibrary_strncmp(lineStart, "Features", 8)) {
				_yepLibrary_DecodeLinuxARMCpuFeatures(infoValueStart, infoValueEnd);
			}
			break;
		case 9:
			if (!_yepLibrary_strncmp(lineStart, "processor", 9)) {
				_yepLibrary_DecodeLinuxARMProcessorNumber(infoValueStart, infoValueEnd);
			}
			break;
		case 11:
			if (!_yepLibrary_strncmp(lineStart, "CPU variant", 11)) {
				_yepLibrary_DecodeLinuxARMCpuVariant(infoValueStart, infoValueEnd);
			}
			break;
		case 12:
			if (!_yepLibrary_strncmp(lineStart, "CPU revision", 12)) {
				_yepLibrary_DecodeLinuxARMCpuRevision(infoValueStart, infoValueEnd);
			}
			break;
		case 13:
			if (!_yepLibrary_strncmp(lineStart, "I line length", 13)) {
				_yepLibrary_DecodeLinuxARMCacheNumber(infoValueStart, infoValueEnd, _cpuInfo.iLineLength);
			} else if (!_yepLibrary_strncmp(lineStart, "D line length", 13)) {
				_yepLibrary_DecodeLinuxARMCacheNumber(infoValueStart, infoValueEnd, _cpuInfo.dLineLength);
			}
			break;
		case 15:
			if (!_yepLibrary_strncmp(lineStart, "CPU implementer", 15)) {
				_yepLibrary_DecodeLinuxARMCpuImplementer(infoValueStart, infoValueEnd);
			} else if (!_yepLibrary_strncmp(lineStart, "CPU implementor", 15)) {
				_yepLibrary_DecodeLinuxARMCpuImplementer(infoValueStart, infoValueEnd);
			}
			break;
		case 16:
			if (!_yepLibrary_strncmp(lineStart, "CPU architecture", 16)) {
				_yepLibrary_DecodeLinuxARMCpuArchitecture(infoValueStart, infoValueEnd);
			}
	}
}

static void _yepLibrary_DecodeLinuxARMCpuInfo() {
	Yep64u isaFeatures = YepIsaFeaturesDefault;
	Yep64u simdFeatures = YepSimdFeaturesDefault;
	Yep64u systemFeatures = _systemFeatures;
	switch (_cpuInfo.cpuImplementer) {
		case 'A':
			_vendor = YepCpuVendorARM;
			switch (_cpuInfo.cpuPart) {
				case 0xB56: // ARM1156
					isaFeatures |= YepARMIsaFeatureThumb2;
				case 0xB02: // ARM11 MPCore
				case 0xB36: // ARM1136
				case 0xB76: // ARM1176
					// WTF? Raspberry Pi shows Architecture: 7 in /proc/cpuinfo
					if (_cpuInfo.cpuArchitectureVersion > 6) {
						_cpuInfo.cpuArchitectureVersion = 6;
					}
					_microarchitecture = YepCpuMicroarchitectureARM11;
					break;
				case 0xC05:
					_microarchitecture = YepCpuMicroarchitectureCortexA5;
					break;
				case 0xC07:
					_microarchitecture = YepCpuMicroarchitectureCortexA7;
					break;
				case 0xC08:
					_microarchitecture = YepCpuMicroarchitectureCortexA8;
					break;
				case 0xC09:
					_microarchitecture = YepCpuMicroarchitectureCortexA9;
					break;
				case 0xC0F:
					_microarchitecture = YepCpuMicroarchitectureCortexA15;
					break;
				default:
					if ((_cpuInfo.cpuPart & 0xF00) == 0x700) {
						_microarchitecture = YepCpuMicroarchitectureARM7;
					} else if ((_cpuInfo.cpuPart & 0xF00) == 0x900) {
						_microarchitecture = YepCpuMicroarchitectureARM9;
					}
					break;
			}
			break;
		case 'D':
			_vendor = YepCpuVendorDEC;
			_microarchitecture = YepCpuMicroarchitectureStrongARM;
			break;
		case 'M':
			_vendor = YepCpuVendorMotorola;
			break;
		case 'T':
			_vendor = YepCpuVendorTI;
			if (_cpuInfo.cpuPart == 0x925) {
				_vendor = YepCpuVendorARM;
				_microarchitecture = YepCpuMicroarchitectureARM9;
			}
			break;
		case 'Q':
			_vendor = YepCpuVendorQualcomm;
			switch (_cpuInfo.cpuPart) {
				case 0x00F: // Mostly Scorpions, but some Cortex A5 may report this value as well
					if (_cpuInfo.featureVfpv4) {
						// Unlike Scorpion, Cortex-A5 comes with VFPv4
						_vendor = YepCpuVendorARM;
						_microarchitecture = YepCpuMicroarchitectureCortexA5;
					} else {
						_microarchitecture = YepCpuMicroarchitectureScorpion;
					}
					break;
				case 0x02D: // Dual-core Scorpions
					_microarchitecture = YepCpuMicroarchitectureScorpion;
					break;
				case 0x04D: // Dual-core Krait
				case 0x06F: // Quad-core Krait
					_microarchitecture = YepCpuMicroarchitectureKrait;
					break;
			}
			break;
		case 'V':
			// I don't understand the logic behind Marvell's CPU parts,
			// so only verified CPU parts are listed here
			switch (_cpuInfo.cpuPart) {
				case 0x693: // PXA 935
					_microarchitecture = YepCpuMicroarchitecturePJ1;
					break;
				case 0x131: // Feroceon 88FR131
				case 0x301: // Feroceon 88FR301 ?
				case 0x331: // Feroceon 88FR331 ?
				case 0x531: // Feroceon 88FR531
				case 0x571: // Feroceon 88FR571 ?
					_microarchitecture = YepCpuMicroarchitecturePJ1;
					break;
				case 0x581: // Armada 510
					_microarchitecture = YepCpuMicroarchitecturePJ4;
					break;
			}
			_vendor = YepCpuVendorMarvell;
			break;
		case 'i':
			if (_cpuInfo.cpuPart == 0xB11) {
				_microarchitecture = YepCpuMicroarchitectureStrongARM;
			} else if ((_cpuInfo.cpuPart & 0xF00) == 0x200) {
				// PXA 210/25X/26X
				_microarchitecture = YepCpuMicroarchitectureXScale;
			} else if ((_cpuInfo.cpuPart & 0xF00) == 0x400) {
				// PXA 27X
				_microarchitecture = YepCpuMicroarchitectureXScale;
			} else if ((_cpuInfo.cpuPart & 0xF00) == 0x600) {
				// PXA 3XX
				_microarchitecture = YepCpuMicroarchitectureXScale;
			}
			_vendor = YepCpuVendorIntel;
			break;
	}

	if (_cpuInfo.cpuArchitectureT) {
		isaFeatures |= YepARMIsaFeatureThumb;
	}
	if (_cpuInfo.cpuArchitectureE && (_cpuInfo.cpuArchitectureVersion == 5)) {
		isaFeatures |= YepARMIsaFeatureV5E;
	}
	if (_cpuInfo.cpuArchitectureJ) {
		isaFeatures |= YepARMIsaFeatureJazelle;
	}

	if (_cpuInfo.cpuArchitectureVersion >= 4) {
		isaFeatures |= YepARMIsaFeatureV4;
	}
	if (_cpuInfo.cpuArchitectureVersion >= 5) {
		isaFeatures |= YepARMIsaFeatureV5;

		const Yep32u probeXScaleResult = _yepLibrary_ProbeInstruction(&_yepLibrary_ProbeXScale);
		if (probeXScaleResult == 0) {
			simdFeatures |= YepARMSimdFeatureXScale;
		}
	}
	if (_cpuInfo.cpuArchitectureVersion >= 6) {
		isaFeatures |= YepARMIsaFeatureV6;
		
		if (_cpuInfo.cpuArchitectureVersion == 6) {
			const Yep32u probeV6KResult = _yepLibrary_ProbeInstruction(&_yepLibrary_ProbeV6K);
			if (probeV6KResult == 0) {
				isaFeatures |= YepARMIsaFeatureV6K;
			}
		}
	}
	if (_cpuInfo.cpuArchitectureVersion >= 7) {
		isaFeatures |= YepARMIsaFeatureV5E;
		isaFeatures |= YepARMIsaFeatureV6K;
		isaFeatures |= YepARMIsaFeatureV7;
		isaFeatures |= YepARMIsaFeatureThumb2;

		const Yep32u probeV7MPResult = _yepLibrary_ProbeInstruction(&_yepLibrary_ProbeV7MP);
		if (probeV7MPResult == 0) {
			isaFeatures |= YepARMIsaFeatureV7MP;
		}
	}

	if (_cpuInfo.featuresValid) {
		if (_cpuInfo.featureThumb) {
			isaFeatures |= YepARMIsaFeatureThumb;
		}
		if (_cpuInfo.featureFpa) {
			isaFeatures |= YepARMIsaFeatureFPA;
		}
		if (_cpuInfo.featureVfp) {
			isaFeatures |= YepARMIsaFeatureVFP;
		}
		if (_cpuInfo.featureEdsp) {
			isaFeatures |= YepARMIsaFeatureV5E;
		}
		if (_cpuInfo.featureJava) {
			isaFeatures |= YepARMIsaFeatureJazelle;
		}
		if (_cpuInfo.featureIwmmxt) {
			const Yep64u wcidReadResult = _yepLibrary_ReadCoprocessor(&_yepLibrary_ReadWCID);
			const Yep32u wcidReadSuccess = yepBuiltin_GetLowPart_64u_32u(wcidReadResult);
			if YEP_LIKELY(wcidReadSuccess) {
				const Yep32u wcid = yepBuiltin_GetHighPart_64u_32u(wcidReadResult);
				const Yep32u coprocessorType = (wcid >> 8) & 0xFF;
				if YEP_LIKELY(coprocessorType >= 0x20) {
					simdFeatures |= YepARMSimdFeatureWMMX2;
					simdFeatures |= YepARMSimdFeatureWMMX;
				} else if YEP_LIKELY(coprocessorType >= 0x10) {
					simdFeatures |= YepARMSimdFeatureWMMX;
				}
			}
		}
		if (_cpuInfo.featureThumbee) {
			isaFeatures |= YepARMIsaFeatureThumbEE;
		}
		if (_cpuInfo.featureNeon) {
			simdFeatures |= YepARMSimdFeatureNEON;
			isaFeatures |= YepARMIsaFeatureVFP;
			isaFeatures |= YepARMIsaFeatureVFP3;
			/* NEON mandates support for VFPv3-D32 */
			isaFeatures |= YepARMIsaFeatureVFPd32;
		}
		if (_cpuInfo.featureVfpv3) {
			isaFeatures |= YepARMIsaFeatureVFP;
			isaFeatures |= YepARMIsaFeatureVFP3;
			// VFPv3 could mean either VFPv3-D16 or VFPv3-D32
			const Yep32u probeVFPd32Result = _yepLibrary_ProbeInstruction(&_yepLibrary_ProbeVFPd32);
			if (probeVFPd32Result == 0) {
				isaFeatures |= YepARMIsaFeatureVFPd32;
			}
		}
		if (_cpuInfo.featureVfpv3d16) {
			isaFeatures |= YepARMIsaFeatureVFP;
			isaFeatures |= YepARMIsaFeatureVFP3;
		}
		if (_cpuInfo.featureVfpv4) {
			isaFeatures |= YepARMIsaFeatureVFP;
			isaFeatures |= YepARMIsaFeatureVFP3;
			isaFeatures |= YepARMIsaFeatureVFPd32;
			isaFeatures |= YepARMIsaFeatureVFP4;
		}
		if (_cpuInfo.featureIdiva) {
			isaFeatures |= YepARMIsaFeatureDiv;
		}
	}
	// Old kernels might not report DIV instructions,
	// so they are not indicated by kernel, probe them
	if (!(isaFeatures & YepARMIsaFeatureDiv)) {
		const Yep32u probeDivResult = _yepLibrary_ProbeInstruction(&_yepLibrary_ProbeDiv);
		if (probeDivResult == 0) {
			isaFeatures |= YepARMIsaFeatureDiv;
		}
	}
	// If CPU supports VFP, but doesn't support VFPv3, it seems strange...
	if ((isaFeatures & YepARMIsaFeatureVFP) && !(isaFeatures & YepARMIsaFeatureVFP3)) {
		const Yep32u probeVFP3Result = _yepLibrary_ProbeInstruction(&_yepLibrary_ProbeVFP3);
		if YEP_LIKELY(probeVFP3Result == 0) {
			isaFeatures |= YepARMIsaFeatureVFP3;
			const Yep32u probeVFPd32Result = _yepLibrary_ProbeInstruction(&_yepLibrary_ProbeVFPd32);
			if (probeVFPd32Result == 0) {
				isaFeatures |= YepARMIsaFeatureVFPd32;
			}
		}
	}
	// If CPU supports VFPv3, perhaps it supports half-precision extension
	if (isaFeatures & YepARMIsaFeatureVFP3) {
		const Yep32u probeVFP3HPResult = _yepLibrary_ProbeInstruction(&_yepLibrary_ProbeVFP3HP);
		if YEP_LIKELY(probeVFP3HPResult == 0) {
			isaFeatures |= YepARMIsaFeatureVFP3HP;
		}
	}
	// If CPU supports VFPv3, but doesn't support VFPv4, it seems strange...
	if ((isaFeatures & YepARMIsaFeatureVFP3) && !(isaFeatures & YepARMIsaFeatureVFP4)) {
		const Yep32u probeVFP4Result = _yepLibrary_ProbeInstruction(&_yepLibrary_ProbeVFP4);
		if YEP_LIKELY(probeVFP4Result == 0) {
			isaFeatures |= YepARMIsaFeatureVFP4;
		}
	}
	// If CPU supports NEON, perhaps it supports half-precision extension
	if (simdFeatures & YepARMSimdFeatureNEON) {
		const Yep32u probeNeonHpResult = _yepLibrary_ProbeInstruction(&_yepLibrary_ProbeNeonHp);
		if YEP_LIKELY(probeNeonHpResult == 0) {
			simdFeatures |= YepARMSimdFeatureNEONHP;
		}
	}
	// If CPU supports NEON and VFPv4, perhaps it supports NEONv2
	if ((isaFeatures & YepARMIsaFeatureVFP4) && (simdFeatures & YepARMSimdFeatureNEON)) {
		const Yep32u probeNeonResult = _yepLibrary_ProbeInstruction(&_yepLibrary_ProbeNeon2);
		if YEP_LIKELY(probeNeonResult == 0) {
			simdFeatures |= YepARMSimdFeatureNEON2;
		}
	}
	// On Marvell CPUs check Marvell Armada extensions.
	// I doubt that any third-party processors will ever support them.
	if YEP_UNLIKELY(_vendor == YepCpuVendorMarvell) {
		const Yep32u probeArmadaResult = _yepLibrary_ProbeInstruction(&_yepLibrary_ProbeCnt32);
		if YEP_LIKELY(probeArmadaResult == 0) {
			isaFeatures |= YepARMIsaFeatureArmada;
		}
	}
	
	if YEP_LIKELY(isaFeatures & YepARMIsaFeatureV6) {
		systemFeatures |= YepSystemFeatureMisalignedAccess;
	}
	if YEP_LIKELY(isaFeatures & YepARMIsaFeatureVFP3) {
		isaFeatures |= YepARMIsaFeatureVFP2;
	}
	
	// VFPv2 is not permitted on ARMv7.
	if YEP_UNLIKELY(!(isaFeatures & YepARMIsaFeatureV7)) {
		if YEP_UNLIKELY(!(isaFeatures & YepARMIsaFeatureVFP2)) {
			const Yep64u fpsidReadResult = _yepLibrary_ReadCoprocessor(&_yepLibrary_ReadFPSID);
			const Yep32u fpsidReadSuccess = yepBuiltin_GetLowPart_64u_32u(fpsidReadResult);
			if YEP_LIKELY(fpsidReadSuccess) {
				const Yep32u fpsid = yepBuiltin_GetHighPart_64u_32u(fpsidReadResult);
				const Yep32u subarchitecture = (fpsid >> 16) & 0x7F;
				if YEP_LIKELY(subarchitecture >= 0x01) {
					isaFeatures |= YepARMIsaFeatureVFP2;
				}
			}
		}
		const Yep64u mvfr0ReadResult = _yepLibrary_ReadCoprocessor(&_yepLibrary_ReadMVFR0);
		const Yep32u mvfr0ReadSuccess = yepBuiltin_GetLowPart_64u_32u(mvfr0ReadResult);
		if YEP_LIKELY(mvfr0ReadSuccess) {
			const Yep32u mvfr0 = yepBuiltin_GetHighPart_64u_32u(mvfr0ReadResult);
			const Yep32u shortVectorsSupport = (mvfr0 >> 24) & 0xF;
			if YEP_UNLIKELY(shortVectorsSupport == 0x1) {
				systemFeatures |= YepARMSystemFeatureVFPVectorMode;
			}
		}
	}

	_isaFeatures = isaFeatures;
	_simdFeatures = simdFeatures;
	_systemFeatures = systemFeatures;

	if YEP_UNLIKELY((_cpuInfo.iSize != 0xFFFFFFFFu) && (_cpuInfo.dSize != 0xFFFFFFFFu)) {
		// For old architectures (pre-v7) the kernel exported the cache information in /proc/cpuinfo
		// If there is information about the cache, use it

		_cache.L1ICacheInfo.cacheSize = _cpuInfo.iSize;
		_cache.L1ICacheInfo.lineSize = _cpuInfo.iLineLength;
		_cache.L1ICacheInfo.associativity = _cpuInfo.iAssoc;
		_cache.L1ICacheInfo.isUnified = false;
		
		_cache.L1DCacheInfo.cacheSize = _cpuInfo.dSize;
		_cache.L1DCacheInfo.lineSize = _cpuInfo.dLineLength;
		_cache.L1DCacheInfo.associativity = _cpuInfo.dAssoc;
		_cache.L1DCacheInfo.isUnified = false;
	} else {
		// There is no way to determine the cache size in user mode in Linux
		// So I make a best guess from the architecture.
		// Guess about L1 cache size is almost always accurate,
		// but L2 might be different on an actual platform.
		switch (_microarchitecture) {
			case YepCpuMicroarchitectureCortexA5:
				_cache.L1ICacheInfo.cacheSize = 32*1024;
				_cache.L1ICacheInfo.isUnified = false;
				_cache.L1DCacheInfo.cacheSize = 32*1024;
				_cache.L1DCacheInfo.isUnified = false;
				_cache.L2CacheInfo.cacheSize = 256*1024;
				_cache.L2CacheInfo.isUnified = true;
				break;
			case YepCpuMicroarchitectureCortexA7:
				_cache.L1ICacheInfo.cacheSize = 32*1024;
				_cache.L1ICacheInfo.isUnified = false;
				_cache.L1DCacheInfo.cacheSize = 32*1024;
				_cache.L1DCacheInfo.isUnified = false;
				_cache.L2CacheInfo.cacheSize = 512*1024;
				_cache.L2CacheInfo.isUnified = true;
				break;
			case YepCpuMicroarchitectureCortexA8:
				_cache.L1ICacheInfo.cacheSize = 32*1024;
				_cache.L1ICacheInfo.lineSize = 64;
				_cache.L1ICacheInfo.isUnified = false;
				_cache.L1DCacheInfo.cacheSize = 32*1024;
				_cache.L1DCacheInfo.lineSize = 64;
				_cache.L1DCacheInfo.isUnified = false;
				_cache.L2CacheInfo.cacheSize = 256*1024;
				_cache.L2CacheInfo.lineSize = 64;
				_cache.L2CacheInfo.isUnified = true;
				break;
			case YepCpuMicroarchitectureCortexA9:
				_cache.L1ICacheInfo.cacheSize = 32*1024;
				_cache.L1ICacheInfo.lineSize = 32;
				_cache.L1ICacheInfo.isUnified = false;
				_cache.L1DCacheInfo.cacheSize = 32*1024;
				_cache.L1DCacheInfo.lineSize = 32;
				_cache.L1DCacheInfo.isUnified = false;
				if (_cpuInfo.processors >= 2) {
					_cache.L2CacheInfo.cacheSize = 1024*1024;
					_cache.L2CacheInfo.associativity = 16;
				} else {
					_cache.L2CacheInfo.cacheSize = 512*1024;
					_cache.L2CacheInfo.associativity = 8; /* Speculation */
				}
				_cache.L2CacheInfo.lineSize = 32;
				_cache.L2CacheInfo.isUnified = true;
				break;
			case YepCpuMicroarchitectureCortexA15:
				_cache.L1ICacheInfo.cacheSize = 64*1024;
				_cache.L1ICacheInfo.lineSize = 64;
				_cache.L1ICacheInfo.isUnified = false;
				_cache.L1DCacheInfo.cacheSize = 64*1024;
				_cache.L1DCacheInfo.lineSize = 64;
				_cache.L1DCacheInfo.isUnified = false;
				_cache.L2CacheInfo.cacheSize = 1024*1024;
				_cache.L2CacheInfo.lineSize = 64;
				_cache.L2CacheInfo.isUnified = true;
				break;
			case YepCpuMicroarchitectureARM11:
				_cache.L1ICacheInfo.cacheSize = 16*1024;
				_cache.L1ICacheInfo.lineSize = 32;
				_cache.L1ICacheInfo.isUnified = false;
				_cache.L1DCacheInfo.cacheSize = 16*1024;
				_cache.L1ICacheInfo.lineSize = 32;
				_cache.L1DCacheInfo.isUnified = false;
				_cache.L2CacheInfo.cacheSize = 128*1024;
				_cache.L1ICacheInfo.lineSize = 32;
				_cache.L2CacheInfo.isUnified = true;
				break;
			case YepCpuMicroarchitectureARM9:
				_cache.L1ICacheInfo.cacheSize = 16*1024;
				_cache.L1ICacheInfo.lineSize = 32;
				_cache.L1ICacheInfo.isUnified = false;
				_cache.L1DCacheInfo.cacheSize = 16*1024;
				_cache.L1ICacheInfo.lineSize = 32;
				_cache.L1DCacheInfo.isUnified = false;
				break;
			case YepCpuMicroarchitectureScorpion:
				_cache.L1ICacheInfo.cacheSize = 32*1024;
				_cache.L1ICacheInfo.lineSize = 32;
				_cache.L1ICacheInfo.isUnified = false;
				_cache.L1DCacheInfo.cacheSize = 32*1024;
				_cache.L1DCacheInfo.lineSize = 32;
				_cache.L1DCacheInfo.isUnified = false;
				if (_cpuInfo.processors >= 2) {
					_cache.L2CacheInfo.cacheSize = 512*1024;
				} else {
					_cache.L2CacheInfo.cacheSize = 256*1024;
				}
				_cache.L2CacheInfo.lineSize = 32;
				_cache.L2CacheInfo.isUnified = true;
				break;
			case YepCpuMicroarchitectureKrait:
				/*
				_cache.L0ICacheInfo.cacheSize = 4*1024;
				_cache.L0ICacheInfo.isUnified = false;
				_cache.L0ICacheInfo.associativity = 1;
				_cache.L0DCacheInfo.cacheSize = 4*1024;
				_cache.L0DCacheInfo.isUnified = false;
				_cache.L0DCacheInfo.associativity = 1;
				*/
			
				_cache.L1ICacheInfo.cacheSize = 16*1024;
				_cache.L1ICacheInfo.isUnified = false;
				_cache.L1ICacheInfo.associativity = 4;
				_cache.L1DCacheInfo.cacheSize = 16*1024;
				_cache.L1DCacheInfo.isUnified = false;
				_cache.L1DCacheInfo.associativity = 4;
				if (_cpuInfo.processors >= 4) {
					_cache.L2CacheInfo.cacheSize = 2*1024*1024;
					_cache.L2CacheInfo.associativity = 16; /* Speculation */
				} else {
					_cache.L2CacheInfo.cacheSize = 1024*1024;
					_cache.L2CacheInfo.associativity = 8;
				}
				_cache.L2CacheInfo.isUnified = true;
				break;
			case YepCpuMicroarchitecturePJ1:
				// Marvell Kirkwood 88F6281 characteristics
				_cache.L1ICacheInfo.cacheSize = 16*1024;
				_cache.L1ICacheInfo.isUnified = false;
				_cache.L1DCacheInfo.cacheSize = 16*1024;
				_cache.L1DCacheInfo.isUnified = false;
				_cache.L2CacheInfo.cacheSize = 256*1024;
				_cache.L2CacheInfo.isUnified = true;
				break;
			case YepCpuMicroarchitecturePJ4:
				// Marvell Armada 510 characteristics
				_cache.L1ICacheInfo.cacheSize = 32*1024;
				_cache.L1ICacheInfo.isUnified = false;
				_cache.L1DCacheInfo.cacheSize = 32*1024;
				_cache.L1DCacheInfo.isUnified = false;
				_cache.L2CacheInfo.cacheSize = 512*1024;
				_cache.L2CacheInfo.isUnified = true;
				break;
			case YepCpuMicroarchitectureXScale:
				if ((_cpuInfo.cpuPart & 0xF00) == 0x200) {
					_cache.L1ICacheInfo.cacheSize = 16*1024;
					_cache.L1ICacheInfo.isUnified = false;
					_cache.L1DCacheInfo.cacheSize = 16*1024;
					_cache.L1DCacheInfo.isUnified = false;
				} else {
					_cache.L1ICacheInfo.cacheSize = 32*1024;
					_cache.L1ICacheInfo.isUnified = false;
					_cache.L1DCacheInfo.cacheSize = 32*1024;
					_cache.L1DCacheInfo.isUnified = false;
				}
				if ((_cpuInfo.cpuPart & 0xF00) == 0x600) {
					// Third-generation XScale CPUs also have L2 cache
					_cache.L2CacheInfo.cacheSize = 256*1024;
					_cache.L2CacheInfo.isUnified = true;
				}
		}
	}
}
#endif
#if defined(YEP_MIPS_CPU) && defined(YEP_LINUX_OS)
static void _yepLibrary_DecodeLinuxMIPSCpuInfo();

static void _yepLibrary_DecodeLinuxMIPSProcessorNumber(const char* valueStart, const char* valueEnd) {
	for (const char* valuePointer = valueStart; valuePointer != valueEnd; valuePointer++) {
		const char digit = *valuePointer;
		if ((digit < '0') || (digit > '9'))
			return;
	}
	_cpuInfo.processors += 1;
}

/*
 *	Full list of MIPS application-specific extensions reported in /proc/cpuinfo:
 *	
 *	* mips16 - support for MIPS16 mode (16-bit encoding for MIPS instructions).
 *	* mdmx - integer SIMD extension (seems to be deprecated).
 *	* mips3d - floating-point SIMD extension (seems to be deprecated).
 *	* smartmips - Cryptography extension (targets mostly smart cards).
 *	* dsp - integer and fixed-point SIMD extension.
 *	* dsp2 - integer and fixed-point SIMD extension (release 2).
 *	* mt - instructions for better exploitation of multithreaded architectures.
 *	* mxu - Ingenic Media Extension 32-bit SIMD instructions (not in official kernel sources).
 *
 *	/proc/cpuinfo on MIPS is populated in file arch/mips/kernel/proc.c in Linux kernel
 *	Note that some devices may use patched Linux kernels with different ASE names.
 *	However, the names above were checked on a large number of /proc/cpuinfo listings.
 *	Also note that Loongson and Ingenic extensions are not reported in ASE list,
 *	so we use "system type" and "cpu model" info to detect them.
 */
static void _yepLibrary_DecodeLinuxMIPSAsesImplemented(const char* valueStart, const char* valueEnd) {
	const char* aseStart = valueStart;
	const char* aseEnd;

	do {
		aseEnd = aseStart + 1;
		for (; aseEnd != valueEnd; aseEnd++) {
			if (*aseEnd == ' ')
				break;
		}
		const YepSize aseLength = aseEnd - aseStart;
		
		switch (aseLength) {
			case 2:
				if (!_yepLibrary_strncmp(aseStart, "mt", 2)) {
					_cpuInfo.aseMt = true;
				}
			case 3:
				if (!_yepLibrary_strncmp(aseStart, "dsp", 3)) {
					_cpuInfo.aseDsp = true;
				} else if (!_yepLibrary_strncmp(aseStart, "mxu", 3)) {
					_cpuInfo.aseMxu = true;
				}
				break;
			case 4:
				if (!_yepLibrary_strncmp(aseStart, "mdmx", 4)) {
					_cpuInfo.aseMdmx = true;
				} else if (!_yepLibrary_strncmp(aseStart, "dsp2", 4)) {
					_cpuInfo.aseDsp2 = true;
				}
				break;
			case 6:
				if (!_yepLibrary_strncmp(aseStart, "mips3d", 6)) {
					_cpuInfo.aseMips3d = true;
				} else if (!_yepLibrary_strncmp(aseStart, "mips16", 6)) {
					_cpuInfo.aseMips16 = true;
				}
				break;
			case 9:
				if (!_yepLibrary_strncmp(aseStart, "smartmips", 8)) {
					_cpuInfo.aseSmartmips = true;
				}
				break;
		}
		aseStart = aseEnd;
		for (; aseStart != valueEnd; aseStart++) {
			if (*aseStart != ' ')
				break;
		}
	} while (aseStart != aseEnd);
}

static void _yepLibrary_DecodeLinuxMIPSWaitInstruction(const char* valueStart, const char* valueEnd) {
	const YepSize valueLength = valueEnd - valueStart;

	// Value should contain either "yes" or "no".
	// Since MIPSCpuInfo.hasWaitInstruction is initialized to false
	// we only check for value "yes".
	switch (valueLength) {
		case 3:
			if (!_yepLibrary_strncmp(valueStart, "yes", 3)) {
				_cpuInfo.hasWaitInstruction = true;
			}
			break;
	}
}

static void _yepLibrary_DecodeLinuxMIPSSystemType(const char* valueStart, const char* valueEnd) {
	const YepSize valueLength = valueEnd - valueStart;
	
	// Usually this field contains SoC manufacturer name
	// which is of no interest. However, for Ingenic CPUs
	// we use this name to detect the availability of
	// Ingenic Media Extensions Instruction Set.
	// Ingenic CPUs (http://en.wikipedia.org/wiki/Ingenic):
	//	* JZ4720
	//		- 240 MHz
	//		- SIMD extension
	//		- 16 KB L1I
	//		- 16 KB L1D
	//		- 32-entry dual-pages joint-TLB
	//		- 4-entry ITLB
	//		- 4-entry DTLB
	//	* JZ4725B
	//		- 360 MHz
	//		- SIMD extension
	//		- 16 KB L1I
	//		- 16 KB L1D
	//		- 32-entry dual-pages joint-TLB
	//		- 4-entry ITLB
	//		- 4-entry DTLB
	//	* JZ4730
	//		- 336 MHz
	//		- MIPS32 R1
	//		- 16 KB L1I
	//		- 16 KB L1D
	//		- 32-entry dual-pages joint-TLB
	//		- 4-entry ITLB
	//		- 4-entry DTLB
	//	* JZ4740/JZ4732
	//		- 360 MHz
	//		- MIPS32 R1
	//		- SIMD extension
	//		- 16 KB L1I
	//		- 16 KB L1D
	//		- 32-entry dual-pages joint-TLB
	//		- 4-entry ITLB
	//		- 4-entry DTLB
	//	* JZ4750
	//		- 360 MHz
	//		- MIPS32 R1
	//		- SIMD extension (SIMD2)
	//		- 16 KB L1I
	//		- 16 KB L1D
	//		- 32-entry dual-pages joint-TLB
	//		- 4-entry ITLB
	//		- 4-entry DTLB
	//	* JZ4755
	//		- 360 MHz
	//		- MIPS32 R1
	//		- SIMD extension (SIMD2)
	//		- 16 KB L1I
	//		- 16 KB L1D
	//		- 32-entry dual-pages joint-TLB
	//		- 4-entry ITLB
	//		- 4-entry DTLB
	//	* JZ4760
	//		- 600 MHz
	//		- MIPS32 R1
	//		- FPU
	//		- SIMD extension (SIMD2)
	//		- 16 KB L1I
	//		- 16 KB L1D
	//		- 32-entry dual-pages joint-TLB
	//		- 4-entry ITLB
	//		- 4-entry DTLB
	//	* JZ4770
	//		- 1000 MHz
	//		- MIPS32 R2
	//		- FPU
	//		- SIMD extension (SIMD2)
	//		- 16 KB L1I
	//		- 16 KB L1D
	//		- 256 KB L2
	//		- 32-entry dual-pages joint-TLB
	//		- 4-entry ITLB
	//		- 4-entry DTLB
	if (valueLength == 6) {
		if (!_yepLibrary_strncmp(valueStart, "JZ4720", 6)) {
			_vendor = YepCpuVendorIngenic;
			_microarchitecture = YepCpuMicroarchitectureXBurst;
			_cpuInfo.hasIngenicMediaExtensions = true;
			_cache.L1DCacheInfo.cacheSize = 16*1024;
			_cache.L1DCacheInfo.isUnified = false;
			_cache.L1ICacheInfo.cacheSize = 16*1024;
			_cache.L1ICacheInfo.isUnified = false;
		} else if (!_yepLibrary_strncmp(valueStart, "JZ4730", 6)) {
			_vendor = YepCpuVendorIngenic;
			_microarchitecture = YepCpuMicroarchitectureXBurst;
			_cpuInfo.hasIngenicMediaExtensions = false;
			_cache.L1DCacheInfo.cacheSize = 16*1024;
			_cache.L1DCacheInfo.isUnified = false;
			_cache.L1ICacheInfo.cacheSize = 16*1024;
			_cache.L1ICacheInfo.isUnified = false;
		} else if (!_yepLibrary_strncmp(valueStart, "JZ4732", 6)) {
			_vendor = YepCpuVendorIngenic;
			_microarchitecture = YepCpuMicroarchitectureXBurst;
			_cpuInfo.hasIngenicMediaExtensions = true;
			_cache.L1DCacheInfo.cacheSize = 16*1024;
			_cache.L1DCacheInfo.isUnified = false;
			_cache.L1ICacheInfo.cacheSize = 16*1024;
			_cache.L1ICacheInfo.isUnified = false;
		} else if (!_yepLibrary_strncmp(valueStart, "JZ4740", 6)) {
			_vendor = YepCpuVendorIngenic;
			_microarchitecture = YepCpuMicroarchitectureXBurst;
			_cpuInfo.hasIngenicMediaExtensions = true;
			_cache.L1DCacheInfo.cacheSize = 16*1024;
			_cache.L1DCacheInfo.isUnified = false;
			_cache.L1ICacheInfo.cacheSize = 16*1024;
			_cache.L1ICacheInfo.isUnified = false;
		} else if (!_yepLibrary_strncmp(valueStart, "JZ4750", 6)) {
			_vendor = YepCpuVendorIngenic;
			_microarchitecture = YepCpuMicroarchitectureXBurst;
			_cpuInfo.hasIngenicMediaExtensions = true;
			_cache.L1DCacheInfo.cacheSize = 16*1024;
			_cache.L1DCacheInfo.isUnified = false;
			_cache.L1ICacheInfo.cacheSize = 16*1024;
			_cache.L1ICacheInfo.isUnified = false;
		} else if (!_yepLibrary_strncmp(valueStart, "JZ4755", 6)) {
			_vendor = YepCpuVendorIngenic;
			_microarchitecture = YepCpuMicroarchitectureXBurst;
			_cpuInfo.hasIngenicMediaExtensions = true;
			_cache.L1DCacheInfo.cacheSize = 16*1024;
			_cache.L1DCacheInfo.isUnified = false;
			_cache.L1ICacheInfo.cacheSize = 16*1024;
			_cache.L1ICacheInfo.isUnified = false;
		} else if (!_yepLibrary_strncmp(valueStart, "JZ4760", 6)) {
			_vendor = YepCpuVendorIngenic;
			_microarchitecture = YepCpuMicroarchitectureXBurst;
			_cpuInfo.hasIngenicMediaExtensions = true;
			_cpuInfo.hasFpu = true;
			_cache.L1DCacheInfo.cacheSize = 16*1024;
			_cache.L1DCacheInfo.isUnified = false;
			_cache.L1ICacheInfo.cacheSize = 16*1024;
			_cache.L1ICacheInfo.isUnified = false;
		} else if (!_yepLibrary_strncmp(valueStart, "JZ4770", 6)) {
			_vendor = YepCpuVendorIngenic;
			_microarchitecture = YepCpuMicroarchitectureXBurst;
			_cpuInfo.hasIngenicMediaExtensions = true;
			_cpuInfo.hasFpu = true;
			_cache.L1DCacheInfo.cacheSize = 16*1024;
			_cache.L1DCacheInfo.isUnified = false;
			_cache.L1ICacheInfo.cacheSize = 16*1024;
			_cache.L1ICacheInfo.isUnified = false;
			_cache.L2CacheInfo.cacheSize = 256*1024;
			_cache.L2CacheInfo.isUnified = true;
		}
	} else if (valueLength == 7) {
		if (!_yepLibrary_strncmp(valueStart, "JZ4725B", 7)) {
			_vendor = YepCpuVendorIngenic;
			_microarchitecture = YepCpuMicroarchitectureXBurst;
			_cpuInfo.hasIngenicMediaExtensions = true;
			_cache.L1DCacheInfo.cacheSize = 16*1024;
			_cache.L1DCacheInfo.isUnified = false;
			_cache.L1ICacheInfo.cacheSize = 16*1024;
			_cache.L1ICacheInfo.isUnified = false;
		}
	}
}

static void _yepLibrary_DecodeLinuxMIPSCpuModel(const char* valueStart, const char* valueEnd) {
	const YepSize valueLength = valueEnd - valueStart;
	
	if (valueLength == 14) {
		if (!_yepLibrary_strncmp(valueStart, "Ingenic JZRISC", 14)) {
			_vendor = YepCpuVendorIngenic;
			_microarchitecture = YepCpuMicroarchitectureXBurst;
			return;
		}
	} else if (valueLength > 14) {
		if (!_yepLibrary_strncmp(valueStart, "Ingenic JZRISC ", 15)) {
			_vendor = YepCpuVendorIngenic;
			_microarchitecture = YepCpuMicroarchitectureXBurst;
			return;
		}
	}
	if (valueLength == 8) {
		if (!_yepLibrary_strncmp(valueStart, "MIPS 24K", 8)) {
			_vendor = YepCpuVendorMIPS;
			_microarchitecture = YepCpuMicroarchitectureMIPS24K;
			return;
		} else if (!_yepLibrary_strncmp(valueStart, "MIPS 34K", 8)) {
			_vendor = YepCpuVendorMIPS;
			_microarchitecture = YepCpuMicroarchitectureMIPS34K;
			return;
		} else if (!_yepLibrary_strncmp(valueStart, "MIPS 74K", 8)) {
			_vendor = YepCpuVendorMIPS;
			_microarchitecture = YepCpuMicroarchitectureMIPS74K;
			return;
		}
	} else if (valueLength > 8) {
		if (!_yepLibrary_strncmp(valueStart, "MIPS 24K ", 9)) {
			_vendor = YepCpuVendorMIPS;
			_microarchitecture = YepCpuMicroarchitectureMIPS24K;
			return;
		} else if (!_yepLibrary_strncmp(valueStart, "MIPS 34K ", 9)) {
			_vendor = YepCpuVendorMIPS;
			_microarchitecture = YepCpuMicroarchitectureMIPS34K;
			return;
		} else if (!_yepLibrary_strncmp(valueStart, "MIPS 74K ", 9)) {
			_vendor = YepCpuVendorMIPS;
			_microarchitecture = YepCpuMicroarchitectureMIPS74K;
			return;
		}
	}
	if (valueLength == 9) {
		if (!_yepLibrary_strncmp(valueStart, "MIPS 24Kc", 9)) {
			_vendor = YepCpuVendorMIPS;
			_microarchitecture = YepCpuMicroarchitectureMIPS24K;
			return;
		}
	} else if (valueLength > 9) {
		if (!_yepLibrary_strncmp(valueStart, "MIPS 24Kc ", 10)) {
			_vendor = YepCpuVendorMIPS;
			_microarchitecture = YepCpuMicroarchitectureMIPS24K;
			return;
		}
	}
}

static void _yepLibrary_DecodeLinuxMIPSTlbEntries(const char* valueStart, const char* valueEnd) {
	Yep32u tlbEntries = 0;
	for (const char* valuePointer = valueStart; valuePointer != valueEnd; valuePointer++) {
		const char character = *valuePointer;
		// Verify that tlb_entries is a decimal number
		if ((character < '0') || (character > '9'))
			return;

		tlbEntries = tlbEntries * 10 + (character - '0');
	}

	_cpuInfo.tlbEntries = tlbEntries;
}

/*	Decode a single line of /proc/cpuinfo information.
 *	Lines have format <words-with-spaces>[ ]*:[ ]<space-separated words>
 *	An example of /proc/cpuinfo:
 *
 *		system type             : JZ4770
 *		processor               : 0
 *		cpu model               : Ingenic JZRISC V4.15  FPU V0.0
 *		BogoMIPS                : 814.28
 *		wait instruction        : yes
 *		microsecond timers      : no
 *		tlb_entries             : 32
 *		extra interrupt vector  : yes
 *		hardware watchpoint     : yes, count: 1, address/irw mask: [0x0fff]
 *		ASEs implemented        :
 *		shadow register sets    : 1
 *		core                    : 0
 *		VCED exceptions         : not available
 *		VCEI exceptions         : not available
 */
static void _yepLibrary_DecodeLinuxMIPSCpuInfoLine(const char* lineStart, const char* lineEnd) {
	// Empty line. Skip.
	if (lineStart == lineEnd)
		return;

	// Search for ':' on the line.
	const char* separator = lineStart;
	for (; separator != lineEnd; separator++) {
		if (*separator == ':')
			break;
	}
	// Skip line if no ':' separator was found.
	if (separator == lineEnd)
		return;

	// Skip trailing spaces in key part.
	const char* infoKeyEnd = separator;
	for (; infoKeyEnd != lineStart; infoKeyEnd--) {
		if ((infoKeyEnd[-1] != ' ') && (infoKeyEnd[-1] != '\t'))
			break;
	}
	// Skip line if key contains nothing but spaces.
	if (infoKeyEnd == lineStart)
		return;

	// Skip leading spaces in value part.
	const char* infoValueStart = separator + 1;
	for (; infoValueStart != lineEnd; infoValueStart++) {
		if (*infoValueStart != ' ')
			break;
	}
	// Value part contains nothing but spaces. Skip line.
	if (infoValueStart == lineEnd)
		return;

	// Skip trailing spaces in value part (if any)
	const char* infoValueEnd = lineEnd;
	for (; infoValueEnd != separator; infoValueEnd--) {
		if (infoValueEnd[-1] != ' ')
			break;
	}

	YepSize infoKeyLength = infoKeyEnd - lineStart;
	switch (infoKeyLength) {
		case 9:
			if (!_yepLibrary_strncmp(lineStart, "processor", 9)) {
				_yepLibrary_DecodeLinuxMIPSProcessorNumber(infoValueStart, infoValueEnd);
			} else if (!_yepLibrary_strncmp(lineStart, "cpu model", 9)) {
				_yepLibrary_DecodeLinuxMIPSCpuModel(infoValueStart, infoValueEnd);
			}
			break;
		case 11:
			if (!_yepLibrary_strncmp(lineStart, "system type", 11)) {
				_yepLibrary_DecodeLinuxMIPSSystemType(infoValueStart, infoValueEnd);
			} else if (!_yepLibrary_strncmp(lineStart, "tlb_entries", 11)) {
				_yepLibrary_DecodeLinuxMIPSTlbEntries(infoValueStart, infoValueEnd);
			}
			break;
		case 16:
			if (!_yepLibrary_strncmp(lineStart, "wait instruction", 16)) {
				_yepLibrary_DecodeLinuxMIPSWaitInstruction(infoValueStart, infoValueEnd);
			} else if (!_yepLibrary_strncmp(lineStart, "ASEs implemented", 16)) {
				_yepLibrary_DecodeLinuxMIPSAsesImplemented(infoValueStart, infoValueEnd);
			}
			break;
	}
}

static void _yepLibrary_DecodeLinuxMIPSCpuInfo() {
	Yep64u isaFeatures = YepIsaFeaturesDefault;
	Yep64u simdFeatures = YepSimdFeaturesDefault;
	
	#ifdef __ANDROID__
		isaFeatures |= YepMIPSIsaFeatureFPU;
	#endif
	if (_cpuInfo.aseMt) {
		isaFeatures |= YepMIPSIsaFeatureMT;
	}
	if (_cpuInfo.aseMips16) {
		isaFeatures |= YepMIPSIsaFeatureMIPS16;
	}
	if (_cpuInfo.aseSmartmips) {
		isaFeatures |= YepMIPSIsaFeatureSmartMIPS;
	}
	if (_cpuInfo.aseMdmx) {
		simdFeatures |= YepMIPSSimdFeatureMDMX;
	}
	if (_cpuInfo.aseMips3d) {
		simdFeatures |= YepMIPSSimdFeatureMIPS3D;
	}
	if (_cpuInfo.aseDsp) {
		simdFeatures |= YepMIPSSimdFeatureDSP;
	}
	if (_cpuInfo.aseDsp2) {
		simdFeatures |= YepMIPSSimdFeatureDSP2;
	}
	if (_cpuInfo.aseMxu || _cpuInfo.hasIngenicMediaExtensions) {
		simdFeatures |= YepMIPSSimdFeatureIMX;
	}
	const Yep32u probeR2Result = _yepLibrary_ProbeInstruction(&_yepLibrary_ProbeR2);
	if (probeR2Result == 0) {
		isaFeatures |= YepMIPSIsaFeatureR2;
	}
	const Yep32u probePairedSingleResult = _yepLibrary_ProbeInstruction(&_yepLibrary_ProbePairedSingle);
	if (probePairedSingleResult == 0) {
		simdFeatures |= YepMIPSSimdFeaturePairedSingle;
	}
	switch (_microarchitecture) {
		case YepCpuMicroarchitectureMIPS24K:
		case YepCpuMicroarchitectureMIPS34K:
		case YepCpuMicroarchitectureMIPS74K:
			_cache.L1DCacheInfo.cacheSize = 64*1024;
			_cache.L1DCacheInfo.isUnified = false;
			_cache.L1DCacheInfo.lineSize = 32;
			_cache.L1DCacheInfo.associativity = 4;
			_cache.L1ICacheInfo.cacheSize = 64*1024;
			_cache.L1ICacheInfo.isUnified = false;
			_cache.L1ICacheInfo.lineSize = 32;
			_cache.L1ICacheInfo.associativity = 4;
			break;
	}

	_isaFeatures = isaFeatures;
	_simdFeatures = simdFeatures;
}
#endif

#if defined(YEP_LINUX_OS) && (defined(YEP_ARM_CPU) || defined(YEP_MIPS_CPU))
/**
 * @brief	Replacement for LibC strncmp.
 *
 * @details	Compares two strings of specified length for equality.
 *
 * @param stringX	Pointer to the first string to be compared.
 *
 * @param stringY	Pointer to the second string to be compared.
 *
 * @param length	Length of the strings to be compared.
 *
 * @retval true	If the strings are not equal.
 * @retval false If the strings are equal.
 */
static YepBoolean _yepLibrary_strncmp(const char* stringX, const char* stringY, size_t length) {
	for (; length != 0; length -= 1) {
		const char charX = *stringX++;
		const char charY = *stringY++;
		if (charX != charY) {
			return true;
		}
	}
	return false;
}

static void _yepLibrary_memcpy(void* destination, const void* source, size_t length) {
	Yep8u* destinationPointer = static_cast<Yep8u*>(destination);
	const Yep8u* sourcePointer = static_cast<const Yep8u*>(source);
	for (; length != 0; length -= 1) {
		const Yep8u byte = *sourcePointer++;
		*destinationPointer++ = byte;
	}
}

static void _yepLibrary_memset(void* destination, int value, size_t length) {
	Yep8u* destinationPointer = static_cast<Yep8u*>(destination);
	Yep8u byteValue = static_cast<Yep8u>(value);
	for (; length != 0; length -= 1) {
		*destinationPointer++ = byteValue;
	}
}

static YepStatus _yepLibrary_InitLinuxCpuInfo() {
	int file = yepSyscall_open("/proc/cpuinfo", O_RDONLY);
	if YEP_UNLIKELY(file < 0) {
		return YepStatusSystemError;
	}
	YepStatus status = YepStatusOk;
	
	const YepSize bufferSize = 1024;
	char buffer[bufferSize];
	char* bufferStart = buffer;

	ssize_t readResult;
	do {
		readResult = yepSyscall_read(file, bufferStart, bufferSize - (bufferStart - buffer));
		if YEP_UNLIKELY(readResult < 0) {
			status = YepStatusSystemError;
			goto cleanup;
		}
		const char* bufferEnd = bufferStart + readResult;
		const char* lineStart = buffer;
		const char* lineEnd;

		if YEP_UNLIKELY(readResult == 0) {
			// Process the remaining text in the buffer as a single line
			#if defined(YEP_ARM_CPU)
				_yepLibrary_DecodeLinuxARMCpuInfoLine(lineStart, bufferEnd);
			#else
				_yepLibrary_DecodeLinuxMIPSCpuInfoLine(lineStart, bufferEnd);
			#endif
		} else {
			do {
				// Find the end of line
				for (lineEnd = lineStart; lineEnd != bufferEnd; lineEnd++) {
					if (*lineEnd == '\n')
						break;
				}

				// If we found the end-of-line
				if (lineEnd != bufferEnd) {
					#if defined(YEP_ARM_CPU)
						_yepLibrary_DecodeLinuxARMCpuInfoLine(lineStart, lineEnd);
					#else
						_yepLibrary_DecodeLinuxMIPSCpuInfoLine(lineStart, lineEnd);
					#endif
					lineStart = lineEnd + 1;
				}
			} while (lineEnd != bufferEnd);
			
			const YepSize lineLength = lineEnd - lineStart;
			_yepLibrary_memcpy(buffer, lineStart, lineLength);
			bufferStart = buffer + lineLength;
		}
	} while (readResult != 0);

cleanup:
	int closeResult = yepSyscall_close(file);
	if YEP_UNLIKELY(closeResult < 0) {
		return YepStatusSystemError;
	}
	return status;
}
#endif

#if defined(YEP_IA64_CPU)
YepStatus _yepLibrary_InitIA64CpuVendor() {
	_vendor = YepCpuVendorIntel;
}
#endif
